# Copyright (C) 2009,2010  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


package Xyne::Arch::AUR;
use strict;
use warnings;

use File::Basename qw/basename/;
use LWP::Simple qw/get $ua/;
$ua->timeout(5);
use URI::Escape qw/uri_escape/;

use Xyne::Arch::Pacman qw/compare_versions check_version_inequality/;

use Xyne::Common::Functions qw/str/;
use Xyne::Common::JSON qw/json_to_obj/;
use Xyne::Common::Threads qw/parallelize parallel_mirror_urls/;

our $VERSION = '0.10';

my $AUR_URL = 'http://aur.archlinux.org';
my $AUR_RPC_URL = $AUR_URL . '/rpc.php';
my $AUR_SEARCH_FORM_URL = $AUR_URL . '/packages.php?detail=1';
my $AUR_PAGE_URL = $AUR_URL . '/packages.php?ID=';



sub new {
  my ($proto,$args) = @_;
  my $class = ref($proto) || $proto;
  my $self  = {};
  bless ($self, $class);

  $self->{RPC} = {};

  if (defined($args->{Pacman}))
  {
    $self->{Pacman} = $args->{Pacman};
  }
  else
  {
    $self->{Pacman} = Xyne::Arch::Pacman->new({autoload => 1});
  }
  $self->{CALLBACK}->{QUIET} = 0;

  return $self;
}






sub load_categories
{
  my $self = $_[0];
  my $html = get($AUR_SEARCH_FORM_URL);
  if ($html =~ m/<select name='C'>(.+?)<\/select>/s)
  {
    my $options = $1;
    while ($options =~ m/<option value="(\d+)">(.+?)<\/option>/g)
    {
      $self->{CATEGORIES}->{$1} = $2;
    }
  }
}



########## RPC ##########

sub get_cached_json
{
  my ($self,$type,@args) = @_;
  my @uncached = grep {not defined($self->{RPC}->{$type}->{$_})} @args;
  foreach my $response (&parallel_get_json($type, @uncached))
  {
    my $arg = $response->[0];
    my $obj = $response->[1];
    if (not ref($obj) eq 'HASH' or not defined($obj->{'type'}))
    {
      $self->maybe_display_error("failed to retrieve JSON data from aur.archlinux.org (type=$type&arg=$arg)\n");
      next;
    }
    elsif ($obj->{'type'} eq $type)
    {
      my @results;
      if ( ref($obj->{'results'}) eq 'ARRAY')
      {
        @results = @{ $obj->{'results'} };
      }
      else
      {
        @results = ($obj->{'results'});
      }
      foreach my $result (@results)
      {
        if (ref($result) ne 'HASH')
        {
          $self->maybe_display_error("failed to parse JSON results for $arg\n");
          next;
        }
        # Convert the relative AUR URL path to a full URL
        # This should work but some older packages have incorrect entries in the database.
        #$result->{'URLPath'} = $AUR_URL . $result->{'URLPath'} if defined($result->{'URLPath'});
        my $pkgname = $result->{'Name'};
        $result->{'URLPath'} = $AUR_URL . '/packages/'.$pkgname.'/'.$pkgname.'.tar.gz';
        $result->{'Repository'} = 'AUR';
        $result->{'AURPage'} = $AUR_PAGE_URL . $result->{'ID'};
        if (defined($result->{'CategoryID'}))
        {
          my $catID = $result->{'CategoryID'};
          my $cat = defined($self->{CATEGORIES}->{$catID}) ? $self->{CATEGORIES}->{$catID} : $catID;
          $result->{'Category'} = $cat if $cat ne $catID;
        }
        if ($type ne 'info' and defined($result->{'Name'}))
        {
          $self->{RPC}->{'info'}->{$result->{'Name'}} = [$result]
        }
      }
      $self->{RPC}->{$type}->{$arg} = [@results] if @results;
    }
    elsif ($obj->{'type'} eq 'error')
    {
      $self->maybe_display_warning("JSON error (type=$type&arg=$arg): $obj->{'results'}\n");
    }
    else
    {
      $self->maybe_display_warning("unknown JSON error (type=$type&arg=$arg)\n");
    }
  }
  return map {@{$self->{RPC}->{$type}->{$_}}} grep {exists($self->{RPC}->{$type}->{$_})} @args;
}



sub get_json_key
{
  my ($self, $k, @args) = @_;
  return map {$_->{$k}} grep {defined($_->{$k})} $self->quietly(sub{$self->get_cached_json(@_)}, 'info', @args);
}



sub get_urlpaths
{
  my ($self,@args) = @_;
  return $self->get_json_key('URLPath',@args);
}



sub get_filenames
{
  my ($self,@args) = @_;
  return map {basename($_)} $self->get_json_key('URLPath',@args);
}



sub get_versions
{
  my ($self,@args) = @_;
  return $self->get_json_key('Version',@args);
}



sub get_maintainers
{
  my ($self,@args) = @_;
  return $self->get_json_key('Maintainer',@args);
}



sub get_packages_by_maintainers
{
  my ($self,@maintainers) = @_;
  return sort map {$_->{'Name'}} $self->get_cached_json('msearch',@maintainers);
}



sub get_aur_and_foreign_pkgnames
{
  my ($self,@args) = @_;
  return ([],[]) if not @args;
  my %args;
  $args{$_}=1 foreach @args;
  my @aur_pkgnames = map {$_->{'Name'}} $self->quietly(sub{$self->get_cached_json(@_)}, 'info', @args);
  delete ($args{$_}) foreach @aur_pkgnames;
  my @foreign_pkgnames = keys(%args);
  return (\@aur_pkgnames,\@foreign_pkgnames);
}






sub is_effectively_in_aur
{
  my ($self,$pkgname,$pkgver,$ineq) = @_;
  my @pkgs = $self->quietly(sub{$self->get_cached_json(@_)}, 'info', $pkgname);
  if (@pkgs)
  {
    return 1 if not defined($pkgver);
    my $v = $pkgs[0]->{'Version'};
    return &check_version_inequality($pkgver,$v,$ineq);
  }
  else
  {
    return 0;
  }
}





########## OPERATIONS ##########

sub get_upgradable
{
  my ($self,$ignore,@pkgnames) = @_;
  my $pacman = $self->{Pacman};
  @pkgnames = $pacman->get_foreign_pkgnames() if not @pkgnames;
  @pkgnames = grep {not $pacman->should_ignore($_)} @pkgnames if $ignore;
  # Suppress the warning callback for this as not all foreign packages can
  # be expected to be in the AUR.
  my @pkgs = $self->quietly(sub{$self->get_cached_json(@_)},'info',@pkgnames);
  my @upgradable = ();
  foreach my $pkg (@pkgs)
  {
    my $pkgname = $pkg->{'Name'};
    my $pkgver = $pkg->{'Version'};
    my $cmpver = &compare_versions($pkgver, $pacman->get_local_db_pkg_ver($pkgname));
    push(@upgradable,$pkgname) if $cmpver > 0;
  }
  return @upgradable;
}









########## FILE DOWNLOADING ##########

sub save_taurballs
{
  my ($self,$dir,@args) = @_;
  return if not $dir or not @args;
  my @urls = $self->get_urlpaths(@args);
  return &parallel_mirror_urls($dir,@urls);
}










########## MESSAGES ##########

sub maybe_display_warning
{
  my $self = shift;
  return if $self->{CALLBACK}->{QUIET} or not defined($self->{CALLBACK}->{WARNING});
  foreach my $arg (@_)
  {
    $self->{CALLBACK}->{WARNING}->($arg);
  }
}

sub maybe_display_error
{
  my $self = shift;
  return if not defined($self->{CALLBACK}->{ERROR});
  foreach my $arg (@_)
  {
    $self->{CALLBACK}->{ERROR}->($arg);
  }
}

sub quietly
{
  my ($self,$subref,@args) = @_;
  my $q = $self->{CALLBACK}->{QUIET};
  $self->{CALLBACK}->{QUIET} = 1 if not $q;
  my @output = $subref->(@args);
  $self->{CALLBACK}->{QUIET} = $q;
  return @output;
}




################################
########## CLASS SUBS ##########
################################

sub get_json
{
  my ($type, $arg) = @_;
  $type = 'info' if not $type;
  my $json = get($AUR_RPC_URL . '?type=' . $type . '&arg=' . &uri_escape($arg));
  my $obj = &json_to_obj($json);
  return [$arg, $obj];
}



# with threading
sub parallel_get_json
{
  my $type = shift;
  return () if not @_;
  my @results = ();
  return &parallelize(\&get_json, map {[$type,$_]} @_);
}


1;

__END__

=head1 NAME

Xyne::Arch::AUR - Subroutines for interfacing with the AUR.

=head1 SYNOPSIS

  use Xyne::Arch::AUR;

=head1 DESCRIPTION

This module provides several subroutines for interfacing with the AUR.

=head2 EXPORT

None by default.

=head1 AUTHOR

Xyne <lt>xyne at archlinux dot ca<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009,2010  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
