# Copyright (C) 2009,2010  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


package Xyne::Arch::Bauerbill;
use strict;
use warnings;

use Xyne::Arch::AUR;
use Xyne::Arch::BauerbillConf;
use Xyne::Arch::CPAN;
use Xyne::Arch::Hackage;
use Xyne::Arch::Pacman qw/get_name_and_ver_spec get_name_and_version check_version_inequality check_name_and_ver_spec get_name_and_repo compare_versions/;
use Xyne::Arch::PKGBUILD qw/parse_pkgbuild/;
use Xyne::Arch::Powerpill;

use Xyne::Common::Functions qw/get_term_width join_paths str lstrip rstrip wrap_words wrap_line format_bytes checksum_checker which get_first_existing_parent mkdir_p which/;
use Xyne::Common::OptionHandler;
use Xyne::Common::Threads qw/parallelize parallelize_subs parallel_mirror_urls/;

use File::Copy;
use File::Find;
no warnings 'File::Find';
use File::Path qw/make_path/;
use File::Basename qw/basename dirname/;
use LWP::Simple qw/head/;
use POSIX qw/uname/;
$LWP::Simple::ua->timeout(5);

our $VERSION = '0.75';
our @ISA = ("Xyne::Arch::Powerpill");
our @INFO_KEYS = qw/Repository ID Name Version Description Maintainer URL AURPage URLPath License/;
our $ARG_MAPS = $Xyne::Arch::Powerpill::ARG_MAPS;
our $REFRESH_INTERVAL = 300;
our $HEAD_RETRIES = 3;
our @PKGBUILD_CHECKSUMS = qw/sha512sums sha284sums sha256sums sha1sums md5sums/; #order is deliberate

our @MAKEPKG_OPTIONS = (
  {
    'name' => 'PKGDEST',
    'type' => 'str'
  },
  {
    'name' => 'SRCDEST',
    'type' => 'str'
  },
  {
    'name' => 'PKGEXT',
    'type' => 'str',
    'default' => '.pkg.tar.gz'
  }
);


# This is just for completeness.
our $IS_BSD = -f '/vmunix';
our $CBREAK_ON_CMD = 'stty cbreak '.($IS_BSD ? "/dev/tty 2>&1" : "eol \001");
our $CBREAK_OFF_CMD = 'stty -cbreak '.($IS_BSD ? "/dev/tty 2>&1" : "eol ^@");

our $CABAL2ARCH = &which('/usr/bin/cabal2arch');

########## CONSTRUCTOR ##########

sub new {
  my ($proto,$args) = @_;
  my $class = ref($proto) || $proto;
  my $self  = {};
  bless ($self, $class);


  $self->reset_db;

  if ($args->{'autoload'})
  {
    $args->{'pacman_autoload'} = 1;
    $args->{'bauerbill_autoload'} = 1;
  }


  if (defined($args->{PacmanConf}))
    { $self->{PacmanConf} = $args->{PacmanConf}; }
  else
    { $self->{PacmanConf} = Xyne::Arch::PacmanConf->new(); }
  $self->{PacmanConf}->load_config($args->{'pacman_config_path'}) if $args->{'pacman_autoload'};


  if (defined($args->{PowerpillConf}))
    { $self->{PowerpillConf} = $args->{PowerpillConf}; }
  else
  {
    $self->{PowerpillConf} = Xyne::Arch::BauerbillConf->new({
      'PacmanConf' => $self->{PacmanConf}
    });
  }
  $self->{PowerpillConf}->load_config({
    powerpill_config => $args->{'powerpill_config_path'},
    config => $args->{'bauerbill_config_path'}
  }) if $args->{'bauerbill_autoload'};



  $self->{AUR} = Xyne::Arch::AUR->new({
    'Pacman' => $self
  });
  $self->{AUR}->{CALLBACK}->{WARNING} = sub{$self->pretty_warning(@_)};
  $self->{AUR}->{CALLBACK}->{ERROR} = sub{$self->pretty_error(@_)};

  $self->{PKGBUILD_OBJS} = {};
  $self->{ABS} = {
    'archives' => {}
  };
  $self->{ARCH} = [uname()]->[4];


  $self->{CPAN} = Xyne::Arch::CPAN->new({
    'Pacman' => $self
  });
  $self->{CPAN}->{CALLBACK}->{WARNING} = sub{$self->pretty_warning(@_)};
  $self->{CPAN}->{CALLBACK}->{ERROR} = sub{$self->pretty_error(@_)};



  $self->{Hackage} = Xyne::Arch::Hackage->new({
    'Pacman' => $self
  });
  $self->{Hackage}->{CALLBACK}->{WARNING} = sub{$self->pretty_warning(@_)};
  $self->{Hackage}->{CALLBACK}->{ERROR} = sub{$self->pretty_error(@_)};

  return $self;
}



sub DESTROY
{
  my $self = shift;
  if ($self->{CBREAK})
  {
    $self->cbreak(0);
  }
}



# Keep this here
sub cbreak
{
  my ($self,$on) = @_;
  return if not $self->{'USECBREAK'};
  if ($on and not $self->{CBREAK})
  {
    system $CBREAK_ON_CMD;
    $self->{CBREAK} = 1;
  }
  elsif (not $on and $self->{CBREAK})
  {
    system $CBREAK_OFF_CMD;
    $self->{CBREAK} = 0;
  }
}



sub abort
{
  my ($self,$sig) = @_;
  print "\naborting...\n";
  if ($self->{DB_LOCKED})
  {
    print "removing database lock\n";
    $self->unlock_db;
  }
  exit(1);
}



sub pretty_confirm
{
  my $self = shift;
  $self->cbreak(0);
  return &Xyne::Arch::Powerpill::pretty_confirm($self,@_);
}





########## MAIN ##########
sub main
{
  binmode(STDIN,":utf8");
  binmode(STDOUT,":utf8");
  binmode(STDERR,":utf8");
  my ($self,@args) = @_;
  $SIG{INT} = sub{
    $self->abort(@_);
  };
  my ($clopts,$popts,$clargs) = $self->parse_cli_input(@args);


  if (not @args or $self->get_value('--help'))
  {
    print $self->get_help_message;
    exit;
  }

  elsif ($self->get_value('--man'))
  {
    print $self->get_man_page;
    exit;
  }

  elsif ($self->get_value('--man-nroff'))
  {
    print $self->get_man_page_nroff;
    exit;
  }

  elsif ($self->get_value('--get-config-template'))
  {
    print $self->get_config_template;
    exit;
  }

  elsif ($self->get_value('--pacman-help'))
  {
    my $cmd = $self->get_value('PacmanBin');
    $cmd .= ' --help';
    my @args = $self->{PacmanConf}->get_cli_input();
    $cmd .= ' '.join(' ',@args) if @args;
    exec $cmd;
    exit;
  }


  # Load the configuration files and reload the command-line options to preserve
  # correct overriding order.
  $self->{PacmanConf}->unset_all_values();
  $self->{PacmanConf}->load_config($self->get_value('PacmanConf'));
  $self->{PacmanConf}->reload_cli();

  $self->{PowerpillConf}->load_config({
    powerpill_config => $self->get_value('PowerpillConf'),
    config => $self->get_value('--config')
  });

  $self->{MEMOIZE} = $self->get_value('Memoize');
  $self->{USECBREAK} = $self->get_value('UseCbreak');
  $self->{CPAN}->{CACHE} = $self->get_value('PacpanCache');
  $self->{Hackage}->{CACHE} = $self->get_value('HackageCache');
  $self->set_arch($self->get_value('Arch'));


  if ($self->get_pacvalue('-Q')
  and $self->get_pacvalue('-u'))
  {
    if ($self->get_value('AUR')
    or $self->get_value('CPAN')
    or $self->get_value('Hackage'))
    {
      my @ss = (sub{ return ['pacman', $self->query_upgrades] });

      if ($self->get_value('AUR'))
      {
        #$self->{AUR}->load_categories if $self->get_value('GetCategories');
        push(@ss, sub{ return ['aur', $self->query_aur_upgrades(@$clargs) ] });
      }

      if ($self->get_value('CPAN'))
      {
        push(@ss, sub{
          $self->{CPAN}->maybe_load_database();
          return ['cpan', $self->query_cpan_upgrades(@$clargs) ]
        });
      }

      if ($self->get_value('Hackage'))
      {
        push(@ss, sub{
          $self->{Hackage}->maybe_load_database();
          return ['hackage', $self->query_hackage_upgrades(@$clargs) ]
        });
      }

      my @results = &parallelize_subs(map {[$_]} @ss);
      my %results;
      $results{$_->[0]} = $_->[1] foreach @results;

      foreach ( qw/pacman cpan hackage aur/ )
      {
        print $results{$_} if defined($results{$_});
      }
    }
    else
    {
      print $self->query_upgrades;
      #exec $cmd;
    }
    exit;
  }
  elsif ($self->get_pacvalue('-S'))
  {
    # Check if the database needs to be refreshed before parsing the download.
    my $ys = $self->{PacmanConf}->get_cli_opt_count('-y');
    if ($ys > 0)
    {
      if ($self->get_value('Rebase'))
      {
        $self->maybe_lock_and_rebase(($ys>1));
      }
      else
      {
        my $cmd = $self->get_value('PacmanBin') . ' -S' . ('y' x $ys);
        $cmd = '/usr/bin/sudo '.$cmd if $ENV{'USER'} ne 'root' and -x '/usr/bin/sudo';
        system($cmd);
      }
      $self->{PacmanConf}->clear_cli_opt('-y');
      $self->reset_db;

      if ($self->get_value('CPAN'))
      {
        $self->drop_root_privileges_or_exit();
        $self->{CPAN}->maybe_update_database(($ys>1));
      }

      if ($self->get_value('Hackage'))
      {
        $self->drop_root_privileges_or_exit();
        $self->{Hackage}->maybe_update_database(($ys>1));
      }
    }

    if ($self->get_pacvalue('-s'))
    {
      #my $pac_args = (scalar(@$clargs)>1) ? '"('.join('|',@$clargs).')"' : join(' ',@$clargs);
      #my $cmd = $self->get_value('PacmanBin') . ' ' . join(' ',$self->{PacmanConf}->get_cli_opts). ' ' . $pac_args . ' 2>/dev/null';
      if ($self->get_value('AUR')
      or $self->get_value('CPAN')
      or $self->get_value('Hackage'))
      {
        #my @ss = (sub{ my $o = `$cmd`; return ['pacman',$o] });
        my @ss = (sub{ return ['pacman',$self->search_pacman($clargs)] });

        if ($self->get_value('AUR'))
        {
          push(@ss, sub{
            return ['aur', $self->search_aur($clargs) ]
          });
        }

        if ($self->get_value('CPAN'))
        {
          push(@ss, sub{
            $self->{CPAN}->maybe_load_database;
            return ['cpan', $self->search_cpan($clargs) ]
          });
        }

        if ($self->get_value('Hackage'))
        {
          push(@ss, sub{
            $self->{Hackage}->maybe_load_database;
            return ['hackage', $self->search_hackage($clargs) ]
          });
        }

        my @results = &parallelize_subs(map {[$_]} @ss);
        my %results;
        $results{$_->[0]} = $_->[1] foreach @results;

        my $found_something = 0;
        foreach ( qw/pacman cpan hackage aur/ )
        {
          if (exists($results{$_}) and $results{$_} ne '')
          {
            print $results{$_};
            print "\n";
          }
        }
      }
      else
      {
        #exec $cmd;
        print $self->search_pacman($clargs);
      }
      exit;
    }
    elsif ($self->get_pacvalue('-i'))
    {
      #my $cmd = $self->get_value('PacmanBin') . ' ' . join(' ',$self->{PacmanConf}->get_cli_input) . ' 2>/dev/null';
      if ($self->get_value('AUR')
      or $self->get_value('CPAN')
      or $self->get_value('Hackage'))
      {
        #my @ss = (sub{ my $o = `$cmd`; return ['pacman',$o] });
        my @ss = (sub{ return ['pacman',$self->search_pacman_info()] });

        if ($self->get_value('AUR'))
        {
          push(@ss, sub{
            $self->{AUR}->load_categories if $self->get_value('GetCategories');
            return ['aur', $self->search_aur_info($clargs) ]
          });
        }

        if ($self->get_value('CPAN'))
        {
          push(@ss, sub{
            $self->{CPAN}->maybe_load_database;
            return ['cpan', $self->search_cpan_info($clargs) ]
          });
        }

        if ($self->get_value('Hackage'))
        {
          push(@ss, sub{
            $self->{Hackage}->maybe_load_database;
            return ['hackage', $self->search_hackage_info($clargs) ]
          });
        }

        my @results = &parallelize_subs(map {[$_]} @ss);
        my %results;
        $results{$_->[0]} = $_->[1] foreach @results;

        my $found_something = 0;
        foreach ( qw/pacman cpan hackage aur/ )
        {
          if (exists($results{$_}) and $results{$_} ne '')
          {
            print $results{$_};
            $found_something = 1;
          }
        }
        exit(1) if not $found_something;
      }
      else
      {
        #exec $cmd;
        print $self->search_pacman_info();
      }
      exit;
    }
    elsif (not $self->operation_involves_possible_download)
    {
      $self->switch_to_pacman() if not $ys;
      exit;
    }
  }
  else
  {
    $self->switch_to_pacman();
    exit;
  }







  # Drop root privileges and only regain them when needed.
  $self->drop_root_privileges_or_exit();

  # Check for a writable cache and remember if it requires root.
  my $cachedir = $self->get_writable_cachedir;
  if ($cachedir)
  {
    $self->{CACHEDIR_NEEDS_ROOT} = 0;
  }
  else
  {
    &unsu();
    $cachedir = $self->get_writable_cachedir;
    $self->{CACHEDIR_NEEDS_ROOT} = 1 if $cachedir;
    $self->drop_root_privileges_or_exit();
  }
  if (not $cachedir)
  {
    my $dirlist = "    " . join("\n    ", $self->get_pacvalue('CacheDir'));
    print STDERR "error: no write permissions to any of the following cache directories:\n$dirlist\n";
    exit(1);
  }
  $self->{CACHEDIR} = $cachedir;



  if ($self->get_value('AUR') and $self->get_value('TrustedUsers'))
  {
    $self->pretty_message("Getting trusted user package lists...\n");
    $self->{'AUR'}->get_packages_by_maintainers($self->get_value('TrustedUsers'));
  }

  $self->{PowerpillConf}->set_value('BuildAll',1) if $self->get_value('--get-taurballs');
  my $sync_queue = $self->maybe_get_sync_queue_interactively();
  exit(0) if $self->get_value('--get-taurballs');


  my $unresolved = 0;
  foreach my $pkgname (keys(%$sync_queue))
  {
    if (defined($self->{DB}->{SKIP}->{$pkgname}))
    {
      $self->pretty_warning("skipping: $pkgname\n");
      #$unresolved ++;
      delete($sync_queue->{$pkgname});
      next;
    }
    if ($sync_queue->{$pkgname}->{unresolvable})
    {
      $self->pretty_warning('unresolved dependencies: '.$sync_queue->{$pkgname}->{unresolvable}."\n");
      $unresolved ++;
      delete($sync_queue->{$pkgname});
      next;
    }
  }
  if ($unresolved
  and not (
    $self->get_pacvalue('--noconfirm')
    or $self->pretty_confirm("Do you want to ignore the above packages for this operation?",'y')
  ))
  {
    exit();
  }


  my @sync_queue = sort {$sync_queue->{$b}->{order} <=> $sync_queue->{$a}->{order}}
                   keys(%$sync_queue);

  my @build_queue = ();
  my @download_queue = ();

#   if ($self->get_value('BuildAll'))
#   {
#     @build_queue = @sync_queue;
#   }
#   else
#   {
    foreach my $s (@sync_queue)
    {
      if ($sync_queue->{$s}->{location} ne 'repo'
      or $sync_queue->{$s}->{build})
      {
        push @build_queue, $s;
      }
      else
      {
        push @download_queue, $s;
      }
    }
#   }


  my @install_queue = @download_queue;

  if (@download_queue and not ($self->get_value('--get-metalink') or $self->get_value('--save-metalink')))
  {
    &unsu() if $self->{CACHEDIR_NEEDS_ROOT};
    @download_queue = $self->maybe_download_from_pkgds($cachedir,@download_queue);
    $self->drop_root_privileges_or_exit() if $self->{CACHEDIR_NEEDS_ROOT};
  }



  my $missing_build_sources = {};
  if (@build_queue)
  {
    $self->{SRCDEST} = $self->get_makepkg_config->get_value('SRCDEST');
    if (defined($self->{SRCDEST}))
    {
      my $dpath = $self->{SRCDEST};
      my $cpath = $self->get_value('MakepkgConfig');
      $self->pretty_message("detected SRCDEST in $cpath\n");
      &mkdir_p_or_exit($dpath);
    }
    $self->{PKGDEST} = $self->get_makepkg_config->get_value('PKGDEST');
    if (defined($self->{PKGDEST}))
    {
      my $dpath = $self->{PKGDEST};
      my $cpath = $self->get_value('MakepkgConfig');
      $self->pretty_message("detected PKGDEST in $cpath\n");
      &mkdir_p_or_exit($dpath);
    }
    $self->{PKGEXT} = $self->get_makepkg_config->get_value('PKGEXT');
    foreach my $pkgname (@build_queue)
    {
      my ($_pkgname) = &get_name_and_repo($pkgname);
      my $obj = $self->{PKGBUILD_OBJS}->{$_pkgname};
      next if not ref($obj->{source}) eq 'ARRAY';
      my $dir = defined($self->{SRCDEST}) ? $self->{SRCDEST} : $obj->{dpath};
      my @source = @{ $obj->{source} };
      SOURCE: for (my $i=0; $i<scalar(@source); $i++)
      {
        my $src = $source[$i];
        if ($src =~ m/^(?:ht|f)tp:\/\//)
        {
          my $fpath = &join_paths($dir,basename($src));
          foreach my $checksum (@PKGBUILD_CHECKSUMS)
          {
            if (defined($obj->{$checksum}->[$i]))
            {
              my $value = $obj->{$checksum}->[$i];
              # check if it has already been downloaded
              if (&checksum_checker($fpath,$checksum,$value))
              {
                delete($missing_build_sources->{$pkgname}->{$src}) if exists($missing_build_sources->{$pkgname}->{$src});
                next SOURCE;
              }
              $missing_build_sources->{$pkgname}->{$src}->{$checksum} = $value;
            }
          }
          $missing_build_sources->{$pkgname}->{$src}->{fpath} = $fpath;
        }
      }
      delete($missing_build_sources->{$pkgname}) if not keys(%{ $missing_build_sources->{$pkgname} });
    }
    $missing_build_sources = $self->load_build_source_sizes($missing_build_sources);
  }



  if (@download_queue)
  {
    # Check for writable cachedir and remove existing packages from queue.
    # The user can specify an empty cache to ignore existing files.
    @download_queue = $self->list_uncached_pkgs(@download_queue);
  }



  if (@download_queue or %$missing_build_sources)
  {

    $self->maybe_reflect_pkgnames(@download_queue);
    my $metalink = $self->get_metalink_from_pkgnames_and_build_sources(\@download_queue,$missing_build_sources);

    if ($self->get_value('--get-metalink')
    or $self->get_value('--save-metalink'))
    {
      my $cachedir = $self->{CACHEDIR};
      my $builddir = $self->get_value('BuildDir');
      if ($metalink =~ s/<file name="\Q$builddir\E/<file name="src/g)
      {
        $metalink =~ s/<file name="\Q$cachedir\E/<file name="pkg/g;
      }
      else
      {
        $metalink =~ s/<file name="\Q$cachedir\E\//<file name="/g;
      }
    }
    $metalink =~ s/<file name="\//<file name="/g;

    if ($self->get_value('--get-metalink'))
    {
      print $metalink;
      exit;
    }
    elsif ($self->get_value('--save-metalink'))
    {
      my $fpath = $self->get_value('--save-metalink');
      open(my $fh,'>',$fpath) or die "error: failed to open $fpath ($!)\n";
      print $fh $metalink;
      close($fh);
      exit;
    }

    exit unless $self->get_pacvalue('--noconfirm') or $self->approve_download(\@download_queue,$missing_build_sources);
    $self->maybe_reflect_pkgnames(@download_queue);

    my $i = 0;
    my $j = $self->get_value('RetryLimit');
    my $n = 1;
    while ($n)
    {
      # $> etc have no effect on external applications so don't even bother
      # changing them for the download operation
      &unsu();
      $self->download_metalink_with_aria2($metalink,'/');
      $self->drop_root_privileges_or_exit();

      # change permissions on source files if needed (uid != euid
      if ($< != $>)
      {
        my @fpaths;
        foreach my $pkg (values(%$missing_build_sources))
        {
          push (@fpaths,$_) foreach grep {-e $_} map {$_->{fpath}} values(%$pkg);
        }
        &unsu();
        $self->chown_to_builder(@fpaths);
        $self->drop_root_privileges_or_exit()
      }

      ($missing_build_sources, $n) = $self->check_missing_build_sources($missing_build_sources);
      @download_queue = $self->list_uncached_pkgs(@download_queue);
      $n += scalar(@download_queue);

      if ($self->get_pacvalue('--noconfirm'))
      {
        $i ++;
        $n = 0 if $i > $j;
      }
      elsif ($n)
      {
        if (not $self->pretty_confirm("$n unsuccessful download(s). Would you like to retry?", 'y'))
        {
          exit(0) if $self->pretty_confirm("Would you like to abort the operation?");
          last;
        }
      }
    }
    &unsu() if $self->{CACHEDIR_NEEDS_ROOT};
    $self->remove_aria2_control_files($cachedir);
    $self->drop_root_privileges_or_exit() if $self->{CACHEDIR_NEEDS_ROOT};
  }
  elsif ($self->get_value('--get-metalink')
  or $self->get_value('--save-metalink'))
  {
    $self->pretty_message("Nothing to download.\n");
    exit();
  }



  exit if $self->get_pacvalue('-w');



  if (@build_queue)
  {
    my @makedeps = ();
    my @installed = ();
    my @not_installed = ();
    my @explicit = ();
    my @asdeps = ();


    # This is for removing unneeded packages later.
    foreach my $pkgname (keys %$sync_queue)
    {
      push(@makedeps,$pkgname) if $sync_queue->{$pkgname}->{reason} eq 'mdep';
      $sync_queue->{$pkgname}->{installed} = 1 if $self->is_installed($pkgname);
    }

    # These do not need an explicit reason.
    foreach my $pkgname (@install_queue)
    {
      if ($self->is_installed($pkgname))
      {
        push(@installed,$pkgname);
      }
      else
      {
        push(@not_installed,$pkgname);
      }
    }

    foreach my $pkgname (@not_installed)
    {
      if ($sync_queue->{$pkgname}->{reason} eq 'exp')
      {
        push(@explicit,$pkgname);
      }
      else
      {
        push(@asdeps,$pkgname);
      }
    }


    push(@explicit,@installed);


    if (@explicit)
    {
      $self->pretty_message("Installing downloaded target packages...\n");
      # Installed packages and explicit packages do not require
      # additional flags. Also, install these first to let pacman pull
      # in new deps. It may avoid the second operation below.
      my $pacman_args = '-S';
      $pacman_args .= ' -f' if $self->get_pacvalue('-f');
      $self->run_pacman_as_root(\@explicit,$pacman_args) if (@explicit);
      $self->reset_db;
    }

    #Check if there are still missing deps (these will be makedeps
    # and deps for packages in the build queue):
    @asdeps = grep {not $self->is_installed($_)} @asdeps;

    if (@asdeps)
    {
      $self->pretty_message("Installing downloaded dependencies...\n");
      # Install these as deps.
      my $pacman_args = '-S --asdeps';
      $pacman_args .= ' -f' if $self->get_pacvalue('-f');
      $self->run_pacman_as_root(\@asdeps,$pacman_args);
    }

    my $makepkg_args  = '--config '.$self->get_value('MakepkgConfig');
    $makepkg_args .= ' '.$self->get_value('MakepkgArgs') if $self->get_value('MakepkgArgs');
    foreach my $pkgname (@build_queue)
    {
      my ($_pkgname) = &get_name_and_repo($pkgname);
      my $obj = $self->{PKGBUILD_OBJS}->{$_pkgname};
      my $build_dir = $obj->{dpath};
      my $dpath = defined($self->{PKGDEST}) ? $self->{PKGDEST} : $build_dir;
      my $filename = $obj->{pkgname}.'-'.$obj->{pkgver}.'-'.$obj->{pkgrel};
      my $arch = (grep {$_ eq 'any'} @{ $obj->{'arch'} }) ? 'any' : $self->{ARCH};
      $filename .= '-'.$arch.$self->{PKGEXT};
      my $fpath = &join_paths($dpath,$filename);
      if (not -e $fpath
          or ($makepkg_args
              and $makepkg_args =~ m/(?<!\S)(?:-f|--force)(?!\S)/))
      {
        my $cmd = "cd $build_dir && /usr/bin/makepkg $makepkg_args";

        $self->pretty_message("Building $pkgname...\n");
        my $e = $self->run_cmd_as_builder($cmd);

        if ($e)
        {
          #print STDERR "error: makepkg exited with an error ($e)\n";
          $self->pretty_error("makepkg exited with an error ($e)\n");
        }

        if (not -e $fpath)
        {
          # check for dev builds which change the pkgver
          $self->pkgbuild_to_obj(&join_paths($build_dir,'PKGBUILD'));
          $obj = $self->{PKGBUILD_OBJS}->{$pkgname};
          $filename = $obj->{pkgname}.'-'.$obj->{pkgver}.'-'.$obj->{pkgrel};
          $arch = (grep {$_ eq 'any'} @{ $obj->{'arch'} }) ? 'any' : $self->{ARCH};
          $filename .= '-'.$arch.$self->{PKGEXT};
          $fpath = &join_paths($dpath,$filename);

          if (not -e $fpath)
          {
            #print STDERR "error: could not find built package: $fpath\n";
            $self->pretty_error("no package found: $fpath\n");
            # Make it possible to keep going but do not do so by default with --noconfirm
            next if not ($self->get_pacvalue('--noconfirm') or $self->pretty_confirm('Abort?','y'));
            # Optionally remove makedeps and dependencies of the failed package.
            if (not $self->get_value('KeepMakedeps'))
            {
              my @orphans = grep {$sync_queue->{$_}->{reason} ne 'exp' and not $sync_queue->{$_}->{installed}} keys %$sync_queue;
              if (@orphans)
              {
                $self->reset_db;
                @orphans = grep {$self->is_installed($_)} @orphans;
                if (@orphans and $self->pretty_confirm('Remove leftover dependencies?','y'))
                {
                  $self->pretty_message("Removing leftover dependencies...\n");
                  $self->run_pacman_as_root(\@orphans,'-Ru');
                }
              }
            }
            exit(1);
          }
        }

        if ($self->get_value('CacheBuilt')
        and not $self->get_makepkg_config->get_value('PKGDEST'))
        {
          my $cache_fpath = &join_paths($self->{CACHEDIR},$filename);
          &unsu() if $self->{CACHEDIR_NEEDS_ROOT};
          if (move($fpath,$cache_fpath))
          {
            $self->pretty_message("moved $filename to the cache ($cache_fpath)\n");
            $fpath = $cache_fpath;
          }
          else
          {
            $self->pretty_warning("failed to move $filename to the cache ($fpath)\n");
          }
          $self->drop_root_privileges_or_exit() if $self->{CACHEDIR_NEEDS_ROOT};
        }
      }


      my $reason = $sync_queue->{$pkgname}->{reason};
      my $pacman_args = ($reason eq 'exp') ? '' : '--asdeps';
      $pacman_args .= ' -f' if $self->get_pacvalue('-f');
      $self->pretty_message("Installing $fpath...\n");
      $self->run_pacman_as_root([],"-U $pacman_args $fpath");
    }

    if (@makedeps and not $self->get_value('KeepMakedeps'))
    {
      $self->pretty_message("Removing makedepends...\n");
      $self->reset_db;
      @makedeps = grep {$self->is_installed($_)} @makedeps;
      $self->run_pacman_as_root(\@makedeps,'-Ru');
    }
    $self->pretty_message("Done\n");
    exit;
  }
  else
  {
    $self->switch_to_pacman;
    exit;
  }
}






sub switch_to_pacman
{
  my ($self) = @_;
  my $cmd = $self->get_value('PacmanBin');
  my @args = grep {not defined($self->{DB}->{SKIP}->{$_})} $self->{PacmanConf}->get_cli_input();
  $cmd .= ' '.join(' ',@args) if @args;
  &unsu();
  if ($self->operation_requires_root()
  and $ENV{'USER'} ne 'root'
  and -x '/usr/bin/sudo')
  {
    $cmd = '/usr/bin/sudo ' . $cmd;
  }
  $self->pretty_message("Switching to pacman: $cmd\n");
  exec $cmd;
  #print "cmd: $cmd\n";
  exit;
}



sub run_pacman_as_root
{
  my ($self,$pkgnames,$args) = @_;
  $args = '' if not $args;
  my $cmd = $self->get_value('PacmanBin');
  my $cachedirs = join(' ', map {"--cachedir \"$_\""} $self->get_pacvalue('CacheDir'));
  $cmd .= ' '.$cachedirs.' '.$args.' '.join(' ',@$pkgnames);
  &unsu();
  if ($ENV{'USER'} ne 'root'
  and -x '/usr/bin/sudo')
  {
    $cmd = '/usr/bin/sudo ' . $cmd;
  }
  $self->pretty_message("Running pacman: $cmd\n");
  my $e = system($cmd);
  if ($e)
  {
    print STDERR "error: pacman exited with an error ($e)\n";
    exit(1);
  }
  $self->drop_root_privileges_or_exit();
  #print "cmd: $cmd\n";
  #exit;
}



sub search_pacman
{
  my ($self,$args) = @_;
  my @args = @$args;

  my $pac_args = (scalar(@args)>1) ? '"('.join('|',@args).')"' : join(' ',@args);
  my $cmd = $self->get_value('PacmanBin') . ' ' . join(' ',$self->{PacmanConf}->get_cli_opts). ' ' . $pac_args . ' 2>/dev/null';

  my $w = &get_term_width;
  my $output = '';
  my $desc = '';
  if (open (my $fh, '-|', $cmd))
  {
    binmode($fh,":utf8");
    foreach my $line (<$fh>)
    {
      if (substr($line,0,1) ne ' ')
      {
        $output .= &wrap_line($w,'    ',$desc) if $desc;
        $desc = '';
        my $i = index($line,'/') + 1;
        my $j = index($line,' ');
        my $k = index($line,' ',$j+1);
        my $l = ($k < $j) ? length($line)-$j-2 : $k - $j;
        my $pkgname = substr($line,$i,$j-$i);
        my $pkgver = substr($line,$j+1,$l);
        my $status_str = $self->get_installation_status('search',$pkgname,$pkgver);
        if ($status_str)
        {
          chomp $line;
          $output .= $line . ' ' . $status_str . "\n";
        }
        else
        {
          $output .= $line;
        }
      }
      else
      {
        chomp $line;
        $line =~ s/^\s+//;
        $line =~ s/\s+$//;
        $desc = $desc ? $desc . ' ' . $line : $line;
      }
    }
    close($fh);
    $output .= &wrap_line($w,'    ',$desc) if $desc;
  }
  return $output;
}



sub search_pacman_info
{
  my ($self) = @_;

  my $cmd = $self->get_value('PacmanBin') . ' ' . join(' ',$self->{PacmanConf}->get_cli_input) . ' 2>/dev/null';

  my $w = &get_term_width;
  my $output = '';
  my $pkgname;
  my $pkgver;
  # use a counter to avoid problems with i18n
  my $i = 1;
  if (open (my $fh, '-|', $cmd))
  {
    binmode($fh,":utf8");
    foreach my $line (<$fh>)
    {
      if ($line eq "\n")
      {
        $pkgname = undef;
        $pkgver = undef;
        $i = 0;
      }
      $output .= $line;
      if ($i == 2)
      {
        $pkgname = substr($line,index($line,':')+2);
        chomp $pkgname;
      }
      elsif ($i == 3)
      {
        $pkgver = substr($line,index($line,':')+2);
        $output .= 'Installed      : ' . $self->get_installation_status('info',$pkgname,$pkgver) . "\n";
      }
      $i++
    }
    close($fh);
  }
  return $output;
}








sub approve_download
{
  my ($self,$download_queue,$missing_build_sources) = @_;
  return 1 if not(@$download_queue or %$missing_build_sources);
  my $w = &get_term_width();
  my $marker = $self->maybe_color('::','bold blue');
  my $marker2 = $self->maybe_color(':','bold blue');
  my $csize = 0;
  my $isize = 0;
  my $unknown = 0;
  my @pkgnames = @$download_queue;
  my $n = scalar(@pkgnames);
  my $s = 0;

  print "\n$marker The following sources will be downloaded:\n" if %$missing_build_sources;

  foreach my $pkgname (keys(%$missing_build_sources))
  {
    print "   $pkgname$marker2\n";
    foreach my $url (keys(%{ $missing_build_sources->{$pkgname} }))
    {
      print '     '.$url;
      my $size;
      if (defined($missing_build_sources->{$pkgname}->{$url}->{size}))
      {
        $size = $missing_build_sources->{$pkgname}->{$url}->{size};
        $csize += $size;
      }
      else
      {
        $unknown = 1;
      }
      if ($self->get_pacvalue('ShowSize'))
      {
        print $self->get_pkgsize_str($size);
      }
      print "\n";
      $s ++;
    }
  }

  print "\n$marker The following packages will be downloaded:\n" if $n;
  my $groups = $self->group_pkgnames_by_repo(@pkgnames);
  foreach my $reponame ($self->get_repo_list)
  {
    if (ref($groups->{$reponame}) eq 'ARRAY')
    {
      print '   '.$self->get_reponame_str($reponame)."\n";
      my @ps = ();
      foreach my $p (@{ $groups->{$reponame} })
      {
        my $size = $self->get_sync_db_pkg_csize($p);
        $csize += $size;
        $isize += $self->get_sync_db_pkg_isize($p);
        $p .= '-' . $self->get_sync_db_pkg_ver($p);
        if ($self->get_pacvalue('ShowSize'))
        {
          $p .= $self->get_pkgsize_str($size);
        }
        push(@ps,$p);
      }
      #print '     ',$_,"\n" foreach @ps;
      print &wrap_words($w,'     ',@ps);
    }
  }

  $csize = &format_bytes($csize,'MiB');
  $csize = ">=$csize" if $unknown;
  $isize = &format_bytes($isize,'MiB') if $n;

  print "\n";
  printf ("$marker Number of Sources:            %13d\n", $s) if $s;
  printf ("$marker Number of Packages:           %13d\n", $n) if $n;
  printf ("$marker Total Download Size:          %13s\n", $csize);
  printf ("$marker Binary Package Install Size:  %13s\n", $isize) if $n;
  print "\n";


  return $self->pretty_confirm('Proceed with download?','y');
}



sub select_abs_packages
{
  my ($self,$sync_queue) = @_;
  my @pkgnames;
  foreach my $pkgname (keys %$sync_queue)
  {
   if ($sync_queue->{$pkgname}->{location} eq 'repo'
   and not $sync_queue->{$pkgname}->{build})
   {
    push(@pkgnames,$pkgname);
   }
  }
  return $sync_queue if not @pkgnames;
  my $w = &get_term_width();
  my $marker = $self->maybe_color('::','bold blue');
  print "$marker ABS build queue:\n";
  my $groups = $self->group_pkgnames_by_repo(@pkgnames);
  foreach my $reponame ($self->get_repo_list)
  {
    if (ref($groups->{$reponame}) eq 'ARRAY')
    {
      print '   '.$self->get_reponame_str($reponame)."\n";
      print &wrap_words($w,'     ',@{$groups->{$reponame}});
    }
  }
  if ($self->get_pacvalue('--noconfirm') or $self->pretty_confirm('Build all?','y'))
  {
    foreach my $pkgname (sort @pkgnames)
    {
      $sync_queue->{$pkgname}->{build} = 1;
    }
  }
  else
  {
    foreach my $pkgname (@pkgnames)
    {
      $sync_queue->{$pkgname}->{build} = 1 if $self->pretty_confirm("Build $pkgname?",'y');
    }
  }
  return $sync_queue;
}



sub should_autobuild
{
  my ($self,$pkgname) = @_;
  return 1 if $self->get_value('BuildAll');
  if (not defined($self->{Autobuild}))
  {
    $self->{Autobuild} = {};
    $self->{Autobuild}->{$_} = 1 foreach $self->get_value('Autobuild');
  }
  return $self->{Autobuild}->{$pkgname};
}




sub check_missing_build_sources
{
  my ($self,$missing_build_sources) = @_;
  my $n = 0;
  foreach my $pkgname (keys(%$missing_build_sources))
  {
    SRC: foreach my $src (keys(%{$missing_build_sources->{$pkgname}}))
    {
      my $fpath = $missing_build_sources->{$pkgname}->{$src}->{fpath};
      next if not -e $fpath;
      foreach my $checksum (@PKGBUILD_CHECKSUMS)
      {
        if (defined($missing_build_sources->{$pkgname}->{$src}->{$checksum}))
        {
          my $value = $missing_build_sources->{$pkgname}->{$src}->{$checksum};
          if (&checksum_checker($fpath,$checksum,$value))
          {
            delete($missing_build_sources->{$pkgname}->{$src});
          }
          next SRC;
        }
      }
      # If there are no checksums and the path exists, consider the download
      # complete and let makepkg deal with eventual issues.
      delete($missing_build_sources->{$pkgname}->{$src});
    }
    my $i = scalar(keys(%{$missing_build_sources->{$pkgname}}));
    if ($i > 0)
    {
      $n += $i;
    }
    else
    {
      delete($missing_build_sources->{$pkgname});
    }
  }
  return ($missing_build_sources,$n);
}





########## MAKEPKG ##########

sub get_makepkg_config
{
  my ($self) = @_;
  my $oh = Xyne::Common::OptionHandler->new();
  $oh->declare_options(@MAKEPKG_OPTIONS);
  $oh->load_config($self->get_value('MakepkgConfig'),'');
  foreach my $o (@MAKEPKG_OPTIONS)
  {
    my $n = $o->{name};
    my $v = $oh->get_value($n);
    if ($v and $v =~ m/^(['"])(.*)\1$/)
    {
      $v = $2;
      my $q = $1;
      if ($q eq "'")
      {
        $v =~ s/'//g;
      }
      else
      {
        $v =~ s/\\([$`"\\\n])/$1/g;
      }
      $oh->set_value($n,$v);
    }
  }
  return $oh;
}









########## METALINK ##########

sub get_metalink_from_pkgnames_and_build_sources
{
  my ($self,$pkgnames,$missing_build_sources) = @_;
  my @pkgnames = @$pkgnames;
  my $metalink = Xyne::Common::Metalink -> new();
  my $split_size = $self->get_value('SplitSize');
  $split_size *= 1000000;
  my $dir = $self->{CACHEDIR};
  foreach my $pkgname (@pkgnames)
  {
    my $csize = $self->get_sync_db_pkg_csize($pkgname);
    my $filename = $self->get_sync_db_pkg_filename($pkgname);
    $filename = &join_paths($dir,$filename);

    my $metalink_args = {
      name         => $filename,
      identity     => $pkgname,
      version      => $self->get_sync_db_pkg_ver($pkgname),
      size         => $csize,
      verification => {
        md5 => $self->get_sync_db_pkg_md5sum($pkgname)
      }
    };

    $metalink_args->{maxconnections} = 1 if $split_size > $csize;

    my @urls = $self->get_reflected_sync_db_pkg_urls($pkgname);
    $metalink_args->{'urls'} = [@urls];
    $metalink->add_file($metalink_args);
  }


  foreach my $pkgname (keys(%$missing_build_sources))
  {
    my $obj = $self->{PKGBUILD_OBJS}->{$pkgname};
    #my $dir = defined($self->{SRCDEST}) ? $self->{SRCDEST} : $obj->{dpath};
    foreach my $url (keys(%{ $missing_build_sources->{$pkgname} }))
    {
      my $u = $missing_build_sources->{$pkgname}->{$url};
      my $size = $u->{size};
      #my $filename = &join_paths($dir,basename($url));
      my $filename = $missing_build_sources->{$pkgname}->{$url}->{fpath};

      my $metalink_args = {
        name     => $filename,
        identity => $pkgname,
        version  => $self->get_sync_db_pkg_ver($pkgname),
        urls     => [$url]
      };
      $metalink_args->{size} = $size if defined($size);
      foreach my $checksum (@PKGBUILD_CHECKSUMS)
      {
        next if not defined($u->{$checksum});
        my $abbrev = substr($checksum,0,index($checksum,'sum'));
        $metalink_args->{verification}->{$abbrev} = $u->{$checksum};
      }

      $metalink_args->{maxconnections} = 1 if $size and $split_size > $size;
      $metalink->add_file($metalink_args);
    }
  }

  return $metalink->get_metalink;
}






sub load_build_source_sizes
{
  my ($self,$missing_build_sources) = @_;
  $self->pretty_message("Determining sizes of source files...\n");

  my $s = sub {
    my ($p,$url) = @_;
    my $n = 0;
    while ($n < $HEAD_RETRIES)
    {
      $n++;
      my @head = head($url);
      return [$p,$url,$head[1]] if @head;
    }
    return [$p,$url,undef];
  };

  my @args = ();

  foreach my $p (keys(%$missing_build_sources))
  {
    foreach my $url (keys(%{ $missing_build_sources->{$p} }))
    {
      push(@args,[$p,$url]);
    }
  }

  foreach my $r (&parallelize($s,@args))
  {
    my ($p,$url,$size) = @$r;
    $missing_build_sources->{$p}->{$url}->{size} = $size;
  }

  return $missing_build_sources;
}






########## OPERATIONS ##########

sub resolve_sync_queue_deps
{
  my ($self,$sync_queue) = @_;
  $self->maybe_load_sync_queue_pkgbuild_objs($sync_queue);
  $self->build_deptree($_,$sync_queue) foreach (keys %$sync_queue);

  my $checked = {};
  foreach my $pkgname (keys(%$sync_queue))
  {
    $checked = $self->check_deptree($pkgname,$checked);
    if ($checked->{$pkgname} ne '')
    {
      $sync_queue->{$pkgname}->{unresolvable} = $checked->{$pkgname};
    }
  }
  $sync_queue = $self->insert_deps_from_deptree_into_sync_queue($sync_queue);
  return $sync_queue;
}

sub insert_deps_from_deptree_into_sync_queue
{
  my ($self,$sync_queue) = @_;
  my @queue = keys(%$sync_queue);
  # Prevent infinite recursion due to cyclical dependencies by checking that the
  # order never exceeds the number of keys in the hash.
  my $cyclical = 0;
  while (@queue)
  {
    my $pkgname = shift @queue;
    next if $sync_queue->{$pkgname}->{unresolvable};
    my $order = $sync_queue->{$pkgname}->{order};
    if ($order > scalar keys (%$sync_queue) + scalar @queue)
    {
      $cyclical = 1;
      next
    }
    $order += 1;
    my @deps = keys(%{$self->{DB}->{DEPTREE}->{$pkgname}});
    foreach my $dep (@deps)
    {
      $sync_queue->{$dep} = {} if not defined($sync_queue->{$dep});
      $sync_queue->{$dep}->{location} = $self->{DB}->{LOCATION}->{$dep}
        if not defined($sync_queue->{$dep}->{location});
      $sync_queue->{$dep}->{order} = $order if not defined($sync_queue->{$dep}->{order})
                                            or $order > $sync_queue->{$dep}->{order};
      $sync_queue->{$dep}->{reason} = $self->{sync_queue}->{reason}->{$dep};
      push(@queue,$dep) if not grep {$_ eq $dep} @queue;
    }
  }
  $self->pretty_warning("cyclical dependencies detected\n") if $cyclical;
  return $sync_queue;
}

sub check_deptree
{
  # checked is to prevent infinite recursion for cyclical dependencies
  my ($self,$pkgname,$checked) = @_;
  $checked = {} if not defined($checked);
  return $checked if defined($checked->{$pkgname});
  $checked->{$pkgname} = '';

  my $node = $self->{DB}->{DEPTREE}->{$pkgname};
  foreach my $dep (keys(%$node))
  {
    if (ref($node->{$dep}) eq 'HASH')
    {
      $checked = $self->check_deptree($dep,$checked);
      if ($checked->{$dep} ne '')
      {
        $checked->{$pkgname} = "$pkgname -> " .$checked->{$dep};
        return $checked;
      }
    }
    else
    {
      my $reason = $node->{$dep};
      $checked->{$pkgname} = "$pkgname -> $dep ($reason)";
      return $checked;
    }
  }
  return $checked;
}

sub build_deptree
{
  my ($self,$pkgname,$sync_queue,$reason) = @_;
  return 'skipped' if defined($self->{DB}->{SKIP}->{$pkgname});
  return $self->{DB}->{DEPTREE}->{$pkgname} if defined($self->{DB}->{DEPTREE}->{$pkgname});
  $self->{DB}->{DEPTREE}->{$pkgname} = {};
  my @deps = ();
  my @mdeps = ();
  my $location = $self->{DB}->{LOCATION}->{$pkgname};
  if ($location ne 'repo')
  {
    @deps = $self->get_pkgbuild_obj_depends($pkgname);
    @mdeps = $self->get_pkgbuild_obj_makedepends($pkgname);
  }
  else
  {
    # The stepwise check prevents Perl from creating keys.
    if
    (
      (
        exists($sync_queue->{$pkgname}) and
        defined($sync_queue->{$pkgname}->{'build'})
      )
      or $self->get_value('BuildAll')
    )
    {
      @deps = $self->get_pkgbuild_obj_depends($pkgname);
      @mdeps = $self->get_pkgbuild_obj_makedepends($pkgname);
    }
    else
    {
      @deps = $self->get_sync_db_pkg_deps($pkgname);
    }
  }
  my %reason;
  $reason{$_} = 'dep' foreach @deps;
  $reason{$_} = 'mdep' foreach @mdeps;

  # This is for parallelization in the next step.
  my @repo_deps = ();
  my @foreign_deps = ();
  foreach my $dep (@deps,@mdeps)
  {
    if (defined($self->resolve_dep($dep)))
    {
      next;
    }
    elsif ($self->is_effectively_in_sync_db($dep))
    {
      push(@repo_deps,$dep);
      $self->{DB}->{LOCATION}->{$dep} = 'repo';
    }
    else
    {
      push(@foreign_deps,$dep);
    }
  }

  my @aur_deps;
  my @cpan_deps;
  my @hackage_deps;
  my @remaining;
  @foreign_deps = map {(&get_name_and_version($_))[0]} @foreign_deps;

  if ($self->get_value('AUR'))
  {
    my ($aur_deps,$foreign_deps) = $self->{AUR}->get_aur_and_foreign_pkgnames(
      map {(&get_name_and_version($_))[0]} @foreign_deps
    );
    @aur_deps = @$aur_deps;
    @foreign_deps = @$foreign_deps;
  }

  foreach my $dep (@foreign_deps)
  {
    if ($self->get_value('Hackage')
    and defined($self->{Hackage}->get_pkg($dep)))
    {
      push @hackage_deps,$dep;
      $self->{DB}->{LOCATION}->{$_} = 'hackage'
    }
    elsif ($self->get_value('CPAN')
    and defined($self->{CPAN}->get_distribution($dep)))
    {
      push @cpan_deps,$dep;
      $self->{DB}->{LOCATION}->{$_} = 'cpan'
    }
    else
    {
      push @remaining,$dep;
    }
  }
  @foreign_deps = @remaining;

  if (@foreign_deps)
  {
    $self->{DB}->{DEPTREE}->{$pkgname}->{$_} = 'not found' foreach @foreign_deps;
    return $self->{DB}->{DEPTREE}->{$pkgname};
  }

  if (@aur_deps)
  {
    $self->{DB}->{LOCATION}->{$_} = 'aur' foreach @aur_deps;
  }

  foreach my $dep (@deps,@mdeps)
  {
    my $r = defined($reason) ? $reason : $reason{$dep};
    # make sure that all deps of makedeps are viewed as makedeps (unless
    # overridden by a direct dependency, see below)
    $reason = 'mdep' if $r eq 'mdep';
    my $d = $self->resolve_dep($dep);
    # avoid infinite recursions, just in case
    next if $d eq $pkgname;
    if (defined($d) and $d eq '')
    {
      next;
    }
    elsif ($d)
    {
      # Allow "mdep" to be changed to "dep" (squares and rectangles).
      if (not defined($self->{sync_queue}->{reason}->{$d})
      or $self->{sync_queue}->{reason}->{$d} eq 'mdep')
      {
        $self->{sync_queue}->{reason}->{$d} = $r;
      }
      if ($self->should_ignore($d))
      {
        $self->{DB}->{DEPTREE}->{$pkgname}->{$dep} = 'available but ignored';
        last;
      }
      else
      {
        if ($self->{DB}->{LOCATION}->{$d} eq 'aur')
        {
          $self->maybe_load_aur_pkgbuild_objs($d);
        }
        elsif ($self->{DB}->{LOCATION}->{$d} eq 'cpan')
        {
          $self->maybe_load_cpan_pkgbuild_objs($d);
        }
        elsif ($self->{DB}->{LOCATION}->{$d} eq 'hackage')
        {
          $self->maybe_load_hackage_pkgbuild_objs($d);
        }
        elsif ($self->get_value('BuildAll') or $self->should_autobuild($d))
        {
          $self->save_and_extract_abs_taurballs_for_pkgnames($d);
          $self->maybe_source_taurball_files('repo',$d);
        }
        $self->{DB}->{DEPTREE}->{$pkgname}->{$d} = $self->build_deptree($d,$sync_queue,$reason);
      }
    }
    else
    {
      $self->{DB}->{DEPTREE}->{$pkgname}->{$dep} = 'not found';
      last;
    }
  }
  return $self->{DB}->{DEPTREE}->{$pkgname};
}



sub resolve_dep
{
  my ($self,$dep) = @_;
  return $self->{DB}->{DEPS}->{$dep} if exists($self->{DB}->{DEPS}->{$dep});
  $self->pretty_message("resolving $dep... ");
  my ($n,$v,$i) = &get_name_and_ver_spec($dep);
  my $p = $self->is_effectively_installed($n,$v,$i);
  if (defined($p))
  {
    $self->display("installed");
    if ($p ne $n)
    {
      $self->display(" (provided by $p)");
    }
    $self->display("\n");
    $self->{DB}->{DEPS}->{$dep} = '';
    return '';
  }
  $p = $self->is_effectively_in_build_queue($n,$v,$i);
  if (defined($p))
  {
    $self->display("found in build queue");
    if ($p ne $n)
    {
      $self->display(" (provided by $p)");
    }
    $self->display("\n");
    $self->{DB}->{DEPS}->{$dep} = $p;
    return $p;
  }
  $p = $self->is_effectively_in_sync_db($n,$v,$i);
  if (defined($p)
  and (
    not $self->is_installed($p)
    or
      &compare_versions(
        $self->get_sync_db_pkg_ver($p),
        $self->get_local_db_pkg_ver($p)
      ) == 1
    )
  )
  {
    my $r = $self->get_sync_db_pkg_reponame($p);
    $self->display("found in [$r]");
    if ($p ne $n)
    {
      $self->display(" (provided by $p)");
    }
    $self->display("\n");
    $self->{DB}->{LOCATION}->{$p} = 'repo';
    $self->{DB}->{DEPS}->{$dep} = $p;
    return $p;
  }
  if ($self->get_value('AUR') and $self->{AUR}->is_effectively_in_aur($n,$v,$i))
  {
    $self->display("found in AUR\n");
    $self->{DB}->{LOCATION}->{$n} = 'aur';
    $self->{DB}->{DEPS}->{$dep} = $n;
    return $n;
  }
  if ($self->get_value('CPAN'))
  {
    my $d = $self->{CPAN}->get_distribution($n);
    if (defined($d))
    {
      $n = &Xyne::Arch::CPAN::module_to_pkgname($d);
      $self->display("found on CPAN ($d -> $n)\n");
      $self->{DB}->{LOCATION}->{$n} = 'cpan';
      $self->{DB}->{DEPS}->{$dep} = $n;
      return $n;
    }
  }
  if ($self->get_value('Hackage'))
  {
    my $pkg = $self->{Hackage}->get_pkg_name($n);
    if (defined($pkg))
    {
      $n = $self->{Hackage}->pkg_to_pkgname($pkg);
      $self->display("found on Hackage ($pkg -> $n)\n");
      $self->{DB}->{LOCATION}->{$n} = 'hackage';
      $self->{DB}->{DEPS}->{$dep} = $n;
      return $n;
    }
  }
  $self->display($self->maybe_color('unresolvable','bold red')."\n");
  $self->{DB}->{DEPS}->{$dep} = undef;
  return undef;
}



sub is_effectively_in_build_queue
{
  my ($self,$name,$ver,$ineq) = @_;
  foreach my $pkgname (keys(%{ $self->{PKGBUILD_OBJS} }))
  {
    my $obj = $self->{PKGBUILD_OBJS}->{$pkgname};
    my $_pkgname = $obj->{pkgname};
    my $pkgver = $obj->{pkgver};
    if
    (
      defined($_pkgname) and
      $_pkgname eq $name and
      (
        not $ver
        or
        &check_version_inequality($pkgver,$ver,$ineq)
      )
    )
    {
      return $pkgname;
    }
    if (defined($obj->{'provides'}))
    {
      foreach my $whatever (@{$obj->{'provides'}})
      {
        return $pkgname if &check_name_and_ver_spec($name,$ver,$ineq,$whatever);
      }
    }
  }
  return undef;
}





# $group_confirmation_sub must be a subroutine reference which accepts the
# download hash reference followed by the group name and then the list of
# pkgnames, and it must return the download hash reference. If not defined, all
# group members will be included in the download.
sub get_sync_queue
{
  my ($self,$group_confirmation_sub) = @_;
  return {} if not $self->get_pacvalue('-S');
  my @args = $self->{PacmanConf}->get_cli_args;
  if ($self->get_pacvalue('-u') and $self->get_value('--devel'))
  {
    push(@args,$self->get_devel_pkgnames());
  }
  my $sync_queue = {};
  my %foreign = ();
  foreach my $arg (@args)
  {
    my $pkgname = $self->is_effectively_in_sync_db($arg);
    if ($pkgname)
    {
      $sync_queue->{$pkgname} = {
        target => 1,
        order => 0,
        location => 'repo',
        reason => $self->get_sync_queue_install_reason($pkgname)
      };
    }
    else
    {
      my @pkgnames = $self->get_sync_db_group($arg) if $arg !~ m/^[^\/]+\//;
      if (@pkgnames)
      {
        if (ref($group_confirmation_sub) eq 'CODE')
        {
          $self->cbreak(0);
          @pkgnames = $group_confirmation_sub->($arg,@pkgnames);
        }
        $sync_queue->{$_} = {
          target => 1,
          order => 0,
          location => 'repo',
          reason => $self->get_sync_queue_install_reason($_)
        } foreach @pkgnames;
      }
      else
      {
        if ($self->get_value('AUR')
        or $self->get_value('CPAN')
        or $self->get_value('Hackage'))
        {
          $self->pretty_warning("$arg does not match any package or group in sync db\n");
          $foreign{$arg} = 1;
        }
        else
        {
          print STDERR "error: $arg does not match any package or group\n";
          exit(1);
        }
      }
    }
  }
  my @foreign = keys(%foreign);
  my @hackage = ();
  my @cpan = ();
  my @aur = ();
  if (@foreign)
  {
    if ($self->get_value('AUR'))
    {
      $self->pretty_message("Searching for AUR packages...\n");
      @foreign = map {$_ =~ s/^AUR\///i; $_} @foreign;
      my ($aur,$foreign) = $self->{AUR}->get_aur_and_foreign_pkgnames(@foreign);
      @aur = @$aur;
      @foreign = @$foreign;
    }

    if ($self->get_value('Hackage'))
    {
      $| = 1;
      $self->pretty_message("Loading Hackage database...\n");
      $self->{Hackage}->maybe_load_database;
      $self->pretty_message("Searching for Hackage packages...\n");
      my @remaining;
      foreach my $f (@foreign)
      {
        my $pkg = $self->{Hackage}->get_pkg_name($f);
        if (defined($pkg))
        {
          push(@hackage,$self->{Hackage}->pkg_to_pkgname($pkg));
        }
        else
        {
          push @remaining,$f
        }
      }
      @foreign = @remaining;
    }

    if ($self->get_value('CPAN'))
    {
      $| = 1;
      $self->pretty_message("Loading CPAN database...\n");
      $self->{CPAN}->maybe_load_database;
      $self->pretty_message("Searching for CPAN packages...\n");
      my @remaining;
      foreach my $f (@foreign)
      {
        my $d = $self->{CPAN}->get_distribution($f);
        if (defined($d))
        {
          push @cpan,&Xyne::Arch::CPAN::module_to_pkgname($d);
        }
        else
        {
          push @remaining,$f
        }
      }
      @foreign = @remaining;
    }

    if (@foreign)
    {
      $self->pretty_warning('ignoring the following unrecognized packages: '.join('" "',@foreign)."\"\n");
    }
  }

  if ($self->get_pacvalue('--needed'))
  {
    foreach my $pkgname (grep {$sync_queue->{$_}->{location} eq 'repo'} keys(%{$sync_queue}))
    {
      delete($sync_queue->{$pkgname}) if not $self->is_upgradable($pkgname);
    }
    @aur = $self->{AUR}->get_upgradable(1,@aur) if @aur;
    @hackage = $self->{Hackage}->get_upgradable(1,@hackage) if @hackage;
    @cpan = $self->{CPAN}->get_upgradable(1,@cpan) if @cpan;
  }

  $sync_queue->{$_} = {
    target => 1,
    order => 0,
    location => 'aur',
    reason => $self->get_sync_queue_install_reason($_)
  } foreach @aur;

  $sync_queue->{$_} = {
    target => 1,
    order => 0,
    location => 'hackage',
    reason => $self->get_sync_queue_install_reason($_)
  } foreach @hackage;

  $sync_queue->{$_} = {
    target => 1,
    order => 0,
    location => 'cpan',
    reason => $self->get_sync_queue_install_reason($_)
  } foreach @cpan;

  if ($self->get_pacvalue('-u'))
  {
    $self->pretty_message("Searching for upgradable repo packages...\n");
    foreach my $pkgname ($self->get_upgradable_pkgnames(1))
    {
      $sync_queue->{$pkgname} = {
        target => 1,
        order => 0,
        location => 'repo',
        reason => $self->get_sync_queue_install_reason($pkgname)
      } if not defined($sync_queue->{$pkgname});
    }

    if ($self->get_value('AUR'))
    {
      $self->pretty_message("Searching for upgradable AUR packages...\n");
      foreach my $pkgname ($self->{AUR}->get_upgradable(1))
      {
        next if defined($sync_queue->{$pkgname});
        $sync_queue->{$pkgname} = {
          target => 1,
          order => 0,
          location => 'aur',
          reason => $self->get_sync_queue_install_reason($pkgname)
        };
      }
    }

    if ($self->get_value('Hackage'))
    {
      $self->pretty_message("Searching for upgradable Hackage packages...\n");
      foreach my $pkgname ($self->{Hackage}->get_upgradable(1))
      {
        next if defined($sync_queue->{$pkgname});
        $sync_queue->{$pkgname} = {
          target => 1,
          order => 0,
          location => 'hackage',
          reason => $self->get_sync_queue_install_reason($pkgname)
        };
      }
    }

    if ($self->get_value('CPAN'))
    {
      $self->pretty_message("Searching for upgradable CPAN packages...\n");
      foreach my $pkgname ($self->{CPAN}->get_upgradable(1))
      {
        next if defined($sync_queue->{$pkgname});
        $sync_queue->{$pkgname} = {
          target => 1,
          order => 0,
          location => 'cpan',
          reason => $self->get_sync_queue_install_reason($pkgname)
        };
      }
    }
  }
  foreach my $pkgname (keys %$sync_queue)
  {
    $sync_queue->{$pkgname}->{build} = 1 if $self->should_autobuild($pkgname);
  }
  $sync_queue = $self->select_abs_packages($sync_queue) if $self->get_value('ABS') and not $self->get_value('BuildAll');

  if (not $self->get_pacvalue('-d'))
  {
    $sync_queue = $self->resolve_sync_queue_deps($sync_queue);
  }
  else
  {
    $self->maybe_load_sync_queue_pkgbuild_objs($sync_queue);
  }

  foreach my $pkgname (keys %$sync_queue)
  {
    $sync_queue->{$pkgname}->{build} = 1 if $self->should_autobuild($pkgname);
  }
  return $sync_queue;
}



sub get_sync_queue_install_reason
{
  my ($self,$pkgname) = @_;
  if (not $self->is_installed($pkgname))
  {
    return ($self->get_pacvalue('--asdeps') ? 'dep' : 'exp')
  }
  return ($self->is_explicitly_installed($pkgname) ? 'exp' : 'dep')
}



sub maybe_load_sync_queue_pkgbuild_objs
{
  my ($self,$sync_queue) = @_;
  my @abs_pkgnames = ();
  my @aur_pkgnames = ();
  my @hackage_pkgnames = ();
  my @cpan_pkgnames = ();
  delete($self->{sync_queue});
  foreach my $pkgname (keys(%$sync_queue))
  {
    if ($sync_queue->{$pkgname}->{location} eq 'aur')
    {
      push(@aur_pkgnames,$pkgname);
    }
    elsif ($sync_queue->{$pkgname}->{location} eq 'hackage')
    {
      push(@hackage_pkgnames,$pkgname);
    }
    elsif ($sync_queue->{$pkgname}->{location} eq 'cpan')
    {
      push(@cpan_pkgnames,$pkgname);
    }
    elsif ($sync_queue->{$pkgname}->{build})
    {
      push(@abs_pkgnames,$pkgname);
    }
    $self->{DB}->{LOCATION}->{$pkgname} = $sync_queue->{$pkgname}->{location};
  }

  if (@aur_pkgnames)
  {
    $self->pretty_message("Loading AUR taurball data...\n");
    $self->maybe_load_aur_pkgbuild_objs(@aur_pkgnames);
  }

  if (@hackage_pkgnames)
  {
    $self->pretty_message("Loading Hackage taurball data...\n");
    $self->maybe_load_hackage_pkgbuild_objs(@hackage_pkgnames);
  }

  if (@cpan_pkgnames)
  {
    $self->pretty_message("Loading CPAN taurball data...\n");
    $self->maybe_load_cpan_pkgbuild_objs(@cpan_pkgnames);
  }

  if (@abs_pkgnames)
  {
    $self->pretty_message("Loading ABS taurball data...\n");
    $self->save_and_extract_abs_taurballs_for_pkgnames(@abs_pkgnames);
    $self->maybe_source_taurball_files('repo',@abs_pkgnames);
  }
}










########## AUR OPERATIONS ##########

sub search_aur_info
{
  my ($self,$args,$type) = @_;
  $type = 'info' if not $type;
  my @pkgs = $self->{AUR}->quietly( sub{$self->{AUR}->get_cached_json(@_)}, $type, @$args);
  my $output = '';
  my @ks;
  if (@pkgs)
  {
    my %pkg = %{$pkgs[0]};
    # Make sure that the predefined keys exist.
    # This is to avoid errors if the JSON interface changes slightly.
    @ks = grep {exists($pkg{$_})} @INFO_KEYS;
    # Append the rest in order.
    my $info_keys_str = " @INFO_KEYS ";
    push @ks, sort grep {index($info_keys_str, " $_ ") == -1} keys(%pkg);
  }
  else
  {
    return $output;
  }

  my $l = 14;
  # find the longest key to format the output;
  foreach (map {length($_)} @ks) {$l = $_ if $_ > $l;}
  @pkgs = sort {$a->{'Name'} cmp $b->{'Name'}} @pkgs;

  my $w = &get_term_width();
  my $spacer = ' ' x ($l+3);
  foreach my $pkg (@pkgs)
  {
    my $pkgname;
    foreach my $k (@ks)
    {
      my $line = sprintf("%-${l}s : %s", $k, $pkg->{$k});
      if (length($line) > $w)
      {
        $output .= &rstrip(substr($line,0,$w)) . "\n";
        $output .= &wrap_line($w, $spacer, &lstrip(substr($line,$w)));
      }
      else
      {
        $output .= $line . "\n";
      }
      $pkgname = $pkg->{$k} if $k eq 'Name';
      if ($k eq 'Version')
      {
        $output .= sprintf("%-${l}s : %s\n", 'Installed', $self->get_installation_status('info',$pkgname,$pkg->{$k}));
      }
    }
    $output .= "\n";
  }
  return $output;
}



sub query_aur_upgrades
{
  my ($self,@args) = @_;
  my $output = "";
  @args = $self->{AUR}->get_upgradable(1,@args);
  return $output if not @args;

  my @pkgs = $self->{AUR}->quietly( sub{$self->{AUR}->get_cached_json(@_)}, 'info', @args);
  @pkgs = sort {$a->{'Name'} cmp $b->{'Name'}} @pkgs;

  foreach my $pkg (@pkgs)
  {
    my $pkgname = $pkg->{'Name'};
    my $aur_ver = $pkg->{'Version'};
    my $local_ver = $self->get_local_db_pkg_ver($pkgname);
    $output .= 'AUR/'.$pkgname;
    if (not $self->get_pacvalue('-q'))
    {
      #$output .= ' ' . $self->maybe_color("$local_ver -> $aur_ver", 'bold green');
      $output .= " $local_ver -> $aur_ver";
    }
    $output .= "\n";
  }
  $output .= "\n" if $output;
  return $output;
}



sub search_aur
{
  my ($self,$args,$type) = @_;
  $type = 'search' if not $type;
  my @pkgs = ();
  my @results = $self->{AUR}->quietly( sub{$self->{AUR}->get_cached_json(@_)}, $type, @$args);
  foreach my $result (@results)
  {
    # results are returned in a JSON array
    if (ref($result) eq 'ARRAY')
    {
      push(@pkgs, @{$result});
    }
    else
    {
      push(@pkgs, $result);
    }
  }
  my %unique = ();
  $unique{$_->{'Name'}} = $_ foreach @pkgs;

  my $output = "";
  my $w = &get_term_width;
  foreach my $pkg (map {$unique{$_}} sort keys(%unique))
  {
    my $pkgname = $pkg->{'Name'};
    my $pkgver = $pkg->{'Version'};

    my $installed = $self->get_installation_status('search',$pkgname,$pkgver);

    if ($self->get_pacvalue('-q'))
    {
      $output .= "$pkgname\n";
    }
    else
    {
#       $output .= $self->maybe_color('AUR/','bold magenta') .
#                  $pkg->{'Name'} . ' ' .
#                  $self->maybe_color($pkg->{'Version'}, 'bold green') .
#                  "\n".&wrap_line($w,'    ',$pkg->{'Description'});
      $output .= 'AUR/' .
                 $pkgname . ' ' .
                 $pkgver .
                 "$installed\n" .
                 &wrap_line($w,'    ',$pkg->{'Description'});
    }
  }
  return $output;
}









sub pkgbuild_objs_to_info
{
  my ($self,$objs,$repo) = @_;
  my @objs = sort {$a->{pkgname} cmp $b->{pkgname}} @$objs;

  my $output = '';
  my $l = 14;
  my $w = &get_term_width();
  my $spacer = ' ' x ($l+3);
  foreach my $obj (@objs)
  {
    my @fields = (
      ['Repository',$repo],
      ['Name',$obj->{pkgname}],
      ['Version',$obj->{pkgver}],
      ['Installed',$self->get_installation_status('info',$obj->{pkgname},$obj->{pkgver})],
      ['Description',$obj->{pkgdesc}],
      ['URL',$obj->{url}],
      ['Licenses',join(' ',@{$obj->{license}})],
      ['Architecture',join(' ',@{$obj->{arch}})],
#      ['Packager','Xyne::Arch::CPAN'],
      ['Groups',defined($obj->{groups}) ? join(' ',@{$obj->{groups}}) : 'None'],
      ['Provides',defined($obj->{provides}) ? join(' ',@{$obj->{provides}}) : 'None'],
      ['Depends On',defined($obj->{depends}) ? join(' ',@{$obj->{depends}}) : 'None'],
      ['Build Deps',defined($obj->{makedepends}) ? join(' ',@{$obj->{makedepends}}) : 'None'],
      ['Optional Deps',defined($obj->{optdepends}) ? join(' ',@{$obj->{optdepends}}) : 'None'],
      ['Conflicts With',defined($obj->{conflicts}) ? join(' ',@{$obj->{conflicts}}) : 'None'],
      ['Replaces',defined($obj->{replaces}) ? join(' ',@{$obj->{replaces}}) : 'None']
    );
    foreach my $f (@fields)
    {
      my $line = sprintf("%-${l}s : %s", $f->[0], $f->[1]);
      if (length($line) > $w)
      {
        my $n = rindex($line,' ',$w);
        $n = $w if $n < 18;
        $output .= &rstrip(substr($line,0,$n)) . "\n";
        $output .= &wrap_line($w, $spacer, &lstrip(substr($line,$n)));
      }
      else
      {
        $output .= $line . "\n";
      }
    }
    $output .= "\n";
  }
  return $output;
}






########## CPAN OPERATIONS ##########

sub search_cpan_info
{
  my ($self,$args) = @_;
  my $cpan = $self->{CPAN};
  my $output = '';
  return $output if not $args or not @$args;
  my @objs = $cpan->get_pkgbuild_objs(@$args);
  return '' if not @objs;
  return $self->pkgbuild_objs_to_info(\@objs,'CPAN');
}



sub query_cpan_upgrades
{
  my ($self,@pkgnames) = @_;
  my $cpan = $self->{CPAN};
  if (not @pkgnames)
  {
    foreach my $d (keys %{ $cpan->{Distributions} })
    {
      my $pkgname = &Xyne::Arch::CPAN::module_to_pkgname($d);
      push (@pkgnames,$pkgname) if $self->is_in_local_db($pkgname);
    }
  }
  @pkgnames = grep {not $self->should_ignore($_)} @pkgnames;
  my $output = '';
  foreach my $pkgname (@pkgnames)
  {
    my $d = $cpan->get_distribution($pkgname);
    my $cpan_ver = $cpan->{Distributions}->{$d}->{version};
    next if $cpan_ver eq 'undef';
    my $local_ver = $self->get_local_db_pkg_ver($pkgname);
    my $cmpver = &compare_versions($cpan_ver,$local_ver);
    if ($cmpver == 1)
    {
      $output .= "CPAN/$pkgname" ;
      if (not $self->get_pacvalue('-q'))
      {
        $output .= " $local_ver -> $cpan_ver";
      }
      $output .= "\n";
    }
  }
  $output .= "\n" if $output;
  return $output;
}



sub search_cpan
{
  my ($self,$args) = @_;
  my @args = @$args;
  return '' if not @args;

  my $output = '';
  my $w = &get_term_width;
  @args = map {$_=~s/::/-/g; $_} @args;
  foreach my $d (sort keys %{$self->{CPAN}->{Distributions}})
  {
    my $pkgname = &Xyne::Arch::CPAN::module_to_pkgname($d);
    my $pkgver = $self->{CPAN}->{Distributions}->{$d}->{version};
    my $matched = 0;
    foreach my $a (@args)
    {
      if ($pkgname =~ m/$a/i)
      {
        $matched = 1;
        last
      }
    }
    next if not $matched;

    my $installed = $self->get_installation_status('search',$pkgname,$pkgver);

    if ($self->get_pacvalue('-q'))
    {
      $output .= "$pkgname\n";
    }
    else
    {
      $output .= 'CPAN/' .
                 $pkgname . ' ' .
                 $pkgver .
                 "$installed\n    " .
                 $self->{CPAN}->get_distribution_url($d) . "\n";
    }
  }
  return $output;
}





########## HACKAGE OPERATIONS ##########

sub search_hackage_info
{
  my ($self,$args) = @_;
  return "install cabal2arch to enable full Hackage support" if not $CABAL2ARCH;
  return '' if not $args or not @$args;
  #my @objs = $self->{Hackage}->get_pkgbuild_objs(@$args);

  my $hackage = $self->{Hackage};
  my $cabal_dpath = &join_paths($hackage->{CACHE},'cabal_files');
  &mkdir_p_or_exit($cabal_dpath);
  my @cabal_urls = map {$hackage->get_cabal_url($_)} @$args;
  &parallel_mirror_urls($cabal_dpath,@cabal_urls);
  my $build_dir = $self->get_hackage_build_dir();
  &mkdir_p_or_exit($build_dir);
  my @objs;
  foreach my $pkgname (@$args)
  {
    my $pkg_name = $hackage->get_pkg_name($pkgname);
    my $cabal_fpath = &join_paths($cabal_dpath,$pkg_name.'.cabal');
    $self->run_cmd_as_builder("cd \"$build_dir\" && $CABAL2ARCH \"$cabal_fpath\" >/dev/null 2>&1");
    $self->pkgbuild_to_obj(&join_paths($build_dir,$pkgname,'PKGBUILD'));
    push(@objs,$self->{PKGBUILD_OBJS}->{$pkgname});
  }
  return '' if not @objs;
  return $self->pkgbuild_objs_to_info(\@objs,'Hackage');
}



sub query_hackage_upgrades
{
  my ($self,@pkgnames) = @_;
  my $hackage = $self->{Hackage};
  @pkgnames = grep {not $self->should_ignore($_)} $hackage->get_upgradable(@pkgnames);
  my $output = '';
  foreach my $pkgname (@pkgnames)
  {
    my $pkg = $hackage->get_pkg($pkgname);
    my $hackage_ver = $pkg->{ver};
    my $local_ver = $self->get_local_db_pkg_ver($pkgname);
    my $cmpver = &compare_versions($hackage_ver,$local_ver);
    if ($cmpver == 1)
    {
      $output .= "Hackage/$pkgname" ;
      if (not $self->get_pacvalue('-q'))
      {
        $output .= " $local_ver -> $hackage_ver";
      }
      $output .= "\n";
    }
  }
  $output .= "\n" if $output;
  return $output;
}



sub search_hackage
{
  my ($self,$args) = @_;
  my @args = @$args;
  return '' if not @args;

  my $output = '';
  my $w = &get_term_width;
  @args = map {$_=~s/::/-/g; $_} @args;
  my $hackage = $self->{Hackage};
  foreach my $pkg (sort keys %{$hackage->{DB}})
  {
    my $pkgname = $hackage->pkg_to_pkgname($pkg);
    my $pkgver = $hackage->{DB}->{$pkg}->{ver};
    my $desc = $hackage->{DB}->{$pkg}->{desc};
    my $matched = 0;
    foreach my $a (@args)
    {
      if ($pkgname =~ m/$a/i)
      {
        $matched = 1;
        last
      }
      elsif ($desc =~ m/$a/i)
      {
        $matched = 1;
        last
      }
    }
    next if not $matched;

    my $installed = $self->get_installation_status('search',$pkgname,$pkgver);

    if ($self->get_pacvalue('-q'))
    {
      $output .= "$pkgname\n";
    }
    else
    {
      $output .= 'Hackage/' .
                 $pkgname . ' ' .
                 $pkgver .
                 "$installed\n    " .
                 $hackage->{DB}->{$pkg}->{desc} . "\n";
    }
  }
  return $output;
}





########## TAURBALLS ##########

# bsdtar is a dependency of makepkg so we can count on it to be installed
sub run_tar_as_builder
{
  my ($self,$args) = @_;
  my $builder = $self->get_unprivileged_user_if_root();
  my $cmd = "bsdtar $args 2>/dev/null";
  if (defined($builder))
  {
    $cmd = "su $builder -c '$cmd'";
  }
  my $output = `$cmd`;
  if ($?)
  {
    print STDERR "error: tar exited with $? ($cmd)\n";
    exit(1);
  }
  return $output;
}

sub save_aur_taurballs
{
  my ($self,@pkgnames) = @_;
  my $dpath = $self->get_aur_taurball_dir();
  &mkdir_p_or_exit($dpath);
  $self->{AUR}->save_taurballs($dpath,@pkgnames);
}

sub extract_aur_taurballs
{
  my ($self,@pkgnames) = @_;
  foreach my $pkgname (@pkgnames)
  {
    my ($filename) = $self->{AUR}->get_filenames($pkgname);
    my $fpath = $self->get_aur_taurball_dir($filename);
    my $dpath = $self->get_aur_build_dir();
    $self->mkdir_p_or_exit($dpath);
    $self->run_tar_as_builder("-zf $fpath -C $dpath -x --no-same-owner --no-same-permissions");

  }
}

sub maybe_source_taurball_files
{
  my ($self,$location,@pkgnames) = @_;
  my $trust = 0;
  my $autopatch = 0;
  my $tag;
  if ($location eq 'repo')
  {
    $trust = 1 if $self->get_value('TrustABS')
               or ($self->get_value('--blindly-trust-everything-when-building-packages-despite-the-inherent-danger'));
                  #and $self->get_pacvalue('--noconfirm'));
    $tag = "trusting ABS";
  }
  foreach my $pkgname (@pkgnames)
  {
    next if exists($self->{PKGBUILD_OBJS_LIST}->{$pkgname});
    $self->{PKGBUILD_OBJS_LIST}->{$pkgname} = 1;
    my ($_pkgname,$reponame) = &get_name_and_repo($pkgname);
    my $dpath;
    if ($location eq 'repo')
    {
      $reponame = $self->get_sync_db_pkg_reponame($pkgname) if not $reponame;
      my $pkgbase = $self->{ABS}->{pkglist}->{&join_paths($reponame,$_pkgname)};
      $dpath = &join_paths($self->get_value('BuildDir'),$pkgbase);
    }
    else
    {
      if ($self->get_value('--blindly-trust-everything-when-building-packages-despite-the-inherent-danger'))
      {
        $trust = 1;
        $tag = $self->maybe_color('blindly trusting all packages','bold red');
      }
      elsif ($location eq 'hackage')
      {
        $trust = 1 if $self->get_value('TrustHackage');
        $tag = "trusting Hackage";
      }
      elsif ($location eq 'cpan')
      {
        $trust = 1 if $self->get_value('TrustCPAN');
        $tag = "trusting CPAN";
      }
      else
      {
        # reset it after each loop
        $trust = 0;
        my @TUs = $self->get_value('TrustedUsers');
        my @ms = $self->{AUR}->get_maintainers($pkgname);
        if (@TUs and @ms)
        {
          my $maintainer = $ms[0];
          if (grep {$_ eq $maintainer} @TUs)
          {
            $trust = 1;
            $tag = "trusting user: $maintainer";
          }
        }
      }
      $dpath = &join_paths($self->get_value('BuildDir'),$location,$_pkgname);
    }

    my $fpath = &join_paths($dpath,'PKGBUILD');
    my $mtime =  (stat($fpath))[9];
    next if defined($self->{PKGBUILD_OBJS}->{$pkgname}->{mtime})
    and $mtime == $self->{PKGBUILD_OBJS}->{$pkgname}->{mtime};

    if ($trust)
    {
      foreach my $p ($self->get_value('Autopatch'))
      {
        if ($p eq $pkgname)
        {
          $autopatch = 1;
          $self->maybe_patch($fpath);
          last;
        }
      }
      $| = 1;
      # Don't bother sourcing it if it's a CPAN pkgbuild because it was generated
      # from a PKGBUILD obj. "next" also skips the redundant install file check
      # below.
      next if $location eq 'cpan';
      $self->pretty_message("sourcing $fpath ($tag)\n");
    }
    else
    {
      $self->pretty_warning("about to source $fpath\n");
      if (not $self->maybe_edit_and_confirm($fpath))
      {
        $self->{DB}->{SKIP}->{$pkgname} = 1;
        next;
      }
    }
    if ($location eq 'cpan')
    {
      # skip sourcing it if it wasn't changed
      next if (stat($fpath))[9] == $mtime;
    }
    my @ns = $self->pkgbuild_to_obj($fpath);

    foreach my $n (@ns)
    {
      my $install_file = $self->{PKGBUILD_OBJS}->{$n}->{'install'};
      if (defined($install_file))
      {
        if (ref($install_file) eq 'ARRAY')
        {
          my $pkgname = $self->{PKGBUILD_OBJS}->{$n}->{'pkgname'};
          my $n = scalar(@$install_file);
          $self->pretty_error("The PKGBUILD for $pkgname is incorrect. Please inform the packager that the \"install\" variable should be a string, not an array\n");
          if ($n == 1)
          {
            $self->pretty_message("fixing the packager's mistake...\n");
            $install_file = $install_file->[0];
          }
          else
          {
            print STDERR "error: unable to safely determine the correct install file\n";
            exit(1);
          }
        }
        $fpath = &join_paths($dpath,$install_file);
        $self->pretty_message("an associated install file was found for $n: $fpath\n");
        if ($trust)
        {
          $self->maybe_patch($fpath) if $autopatch;
        }
        else
        {
          next if not $self->maybe_edit_and_confirm($fpath);
        }
      }
    }
  }
}

sub maybe_patch
{
  my ($self,$fpath) = @_;
  return if not -w $fpath;
  my $patchdir = $self->get_value('PatchDir');

  if (defined($patchdir) and -x '/usr/bin/patch')
  {
    my $l = length($self->get_value('BuildDir'));
    my $patchfile = $patchdir . substr($fpath,$l).".patch";
    if (-f $patchfile)
    {
      my $e = system("patch $fpath < $patchfile");
      if ($e)
      {
        $self->pretty_warn("error: failed to patch $fpath ($e)\n");
      }
      else
      {
        $self->pretty_message("patched $fpath\n");
      }
    }
  }
}

sub maybe_edit_and_confirm
{
  my ($self,$fpath) = @_;
  if (not -e $fpath)
  {
    print STDERR "error: $fpath does not exist\n";
    exit(1);
  }
  elsif (not -r $fpath)
  {
    print STDERR "error: unable to read $fpath\n";
    exit(1);
  }
  elsif (not -w $fpath)
  {
    print STDERR "error: unable to edit $fpath (permission denied)\n";
    exit(1);
  }
  $self->cbreak(1);
  PROMPT: while (1)
  {
    local $| = 1;
    my $w = &get_term_width();
    my $patchdir = $self->get_value('PatchDir');
    my $patchfile;
    my @words = ("view [v] ","backup [b] ", "edit [e] ", "continue [ok] ","skip [s] ", "abort [x] ");
    if (-f $fpath.'.bak')
    {
      push(@words,"restore [r] ");
      if (defined($patchdir) and -x '/usr/bin/diff')
      {
        push(@words,"diff [d] ");
      }
    }
    if (defined($patchdir) and -x '/usr/bin/patch')
    {
      my $l = length($self->get_value('BuildDir'));
      $patchfile = $patchdir . substr($fpath,$l).".patch";
      push(@words,"patch [p] ") if -e $patchfile;
    }

    my $prompt =&wrap_words($w,'', @words);
    chomp $prompt;

    my @valid_answers = map { m/\[(\w+)\] $/; $1 } @words;

    print $prompt . ' ';
    my $ans = '';
    if ($self->get_value('UseCbreak'))
    {
      while (1)
      {
        my $c = getc(STDIN);
        my $n = ord($c);

        if ($n == 10)
        {
          next PROMPT;
        }
        if ($n == 127)
        {
          print "\b \b\b \b";
          if ($ans)
          {
            chop $ans;
            print "\b \b";
          }
        }
        else
        {
          $ans .= $c;
          #print $c;
        }
        if (grep {$ans eq $_} @valid_answers)
        {
          print "\n";
          last;
        }
      }
    }
    else
    {
      $ans = <STDIN>;
      chomp $ans;
    }

    #next if not grep {$ans eq $_} @valid_answers;

    if ($ans eq 'v')
    {
      if (open(my $fh,'<',$fpath))
      {
        my $viewer = $self->get_value('Viewer');
        if ($viewer)
        {
          if (open(my $vh,'|-',$viewer))
          {
            print $vh $_ while <$fh>;
            close($vh);
          }
          else
          {
            print STDERR "error: failed to open pipe to $viewer ($!)\n";
            close($fh);
            exit(1);
          }
        }
        else
        {
          print "\n";
          print $_ while <$fh>;
          close $fh;
          print "\n";
        }
        close $fh;
      }
      else
      {
        print STDERR "error: failed to open $fpath ($!)\n";
        exit(1);
      }
    }
    elsif ($ans eq 'b')
    {
      if (not copy($fpath, $fpath.'.bak'))
      {
        print STDERR "error: failed to copy $fpath: $!\n";
        exit(1)
      }
      next;
    }
    elsif ($ans eq 'r')
    {
      if (not copy($fpath.'.bak', $fpath))
      {
        print STDERR "error: failed to restore $fpath: $!\n";
        exit(1)
      }
      next;
    }
    elsif ($ans eq 'd')
    {
      if (defined($patchfile))
      {
        &mkdir_p_or_exit(dirname($patchfile));
        my $e = system("diff -u $fpath.bak $fpath > $patchfile");
        #print STDERR "error: failed to create patch ($e)\n" if $e;
      }
      next;
    }
    elsif ($ans eq 'p')
    {
      if (defined($patchfile))
      {
        my $e = system("patch $fpath < $patchfile");
        print STDERR "error: failed to apply patch ($e)\n" if $e;
      }
      next;
    }
    elsif ($ans eq 'e')
    {
      my $editor = $self->get_value('Editor');
      $editor = $ENV{'EDITOR'} if not defined($editor);
      $editor = &which($editor);
      while (not $editor)
      {
        $self->pretty_error("no valid editor specified: enter the name or path of an editor or hit enter to exit\n");
        $editor = <STDIN>;
        chomp $editor;
        exit if not $editor;
        $editor = &which($editor);
        $self->{PowerpillConf}->set_value('Editor',$editor);
        #print STDERR "error: editor path is not executable ($editor)\n";
        #exit(1);
      }
      system("$editor \"$fpath\"");
      next;
    }
    elsif ($ans eq 'x')
    {
      $self->abort();
    }
    elsif ($ans eq 'ok')
    {
      return 1;
    }
    elsif ($ans eq 's')
    {
      return 0;
    }
  }
}

sub pkgbuild_to_obj
{
  my ($self,$fpath) = @_;
  $self->drop_root_privileges_or_exit();
  my @pkgnames = ();
  my $mtime =  (stat($fpath))[9];
  foreach my $obj (&parse_pkgbuild($fpath))
  {
    # remove depends from the makedepends array (packager error)
    if (defined($obj->{depends}) and defined($obj->{makedepends}))
    {
      my @deps = @{ $obj->{depends} };
      my @mdeps = @{ $obj->{makedepends} };
      my %mdeps;
      $mdeps{$_} = 2 foreach @deps;
      $mdeps{$_} += 1 foreach @mdeps;
      $obj->{makedepends} = [grep {$mdeps{$_} == 1} keys(%mdeps)]
    }
    my $pkgname = defined($obj->{'pkgname'}) ? $obj->{'pkgname'} : next;
    $obj->{'mtime'} = $mtime;
    $obj->{'dpath'} = dirname($fpath);
    $self->{PKGBUILD_OBJS}->{$pkgname} = $obj;
    push (@pkgnames,$pkgname);
  }
  return @pkgnames;
}

sub get_pkgbuild_obj_key
{
  my ($self,$pkgname,$k) = @_;
  return $self->{PKGBUILD_OBJS}->{$pkgname}->{$k} if defined($self->{PKGBUILD_OBJS}->{$pkgname}->{$k});
  return undef;
}

sub get_pkgbuild_obj_depends
{
  my ($self,$pkgname) = @_;
  my $deps = $self->get_pkgbuild_obj_key($pkgname,'depends');
  return (defined($deps) ? @$deps : ());
}

sub get_pkgbuild_obj_makedepends
{
  my ($self,$pkgname) = @_;
  my $deps = $self->get_pkgbuild_obj_key($pkgname,'makedepends');
  return (defined($deps) ? @$deps : ());
}

sub get_pkgbuild_obj_source
{
  my ($self,$pkgname) = @_;
  my $deps = $self->get_pkgbuild_obj_key($pkgname,'source');
  return (defined($deps) ? @$deps : ());
}

sub get_pkgbuild_obj_md5sums
{
  my ($self,$pkgname) = @_;
  my $deps = $self->get_pkgbuild_obj_key($pkgname,'md5sums');
  return (defined($deps) ? @$deps : ());
}

sub check_pkgbuild_obj_architecture
{
  my ($self,$pkgname) = @_;
  my $arch = $self->get_pkgbuild_obj_key($pkgname,'arch');
  return 0 if not defined($arch);
  return scalar (grep {$_ eq 'any' or $_ eq $self->{ARCH}} @$arch);
}

sub maybe_load_aur_pkgbuild_objs
{
  my ($self,@pkgnames) = @_;
  return if not @pkgnames;
  $self->save_aur_taurballs(@pkgnames);
  $self->extract_aur_taurballs(@pkgnames);
  $self->maybe_source_taurball_files('aur',@pkgnames);
}

sub get_aur_pkgbuild_objs
{
  my ($self,@pkgnames) = @_;
  $self->maybe_load_aur_pkgbuild_objs(@pkgnames);
  return map {$self->{PKGBUILD_OBJS}->{$_}} @pkgnames;
}

sub get_aur_build_dir
{
  my $self = shift;
  return &join_paths($self->get_value('BuildDir'),'aur',@_);
}

sub get_aur_taurball_dir
{
  my $self = shift;
  return &join_paths($self->get_value('TaurballDir'),'aur',@_);
}






########## CPAN ##########

sub maybe_load_cpan_pkgbuild_objs
{
  my ($self,@pkgnames) = @_;
  return if not @pkgnames;
  my @objs = $self->{CPAN}->get_pkgbuild_objs(@pkgnames);
  foreach my $obj (@objs)
  {
    my $pkgname = $obj->{pkgname};
    $self->{PKGBUILD_OBJS}->{$pkgname} = $obj;

    my $dpath = $self->get_cpan_build_dir($pkgname);
    $obj->{dpath} = $dpath;
    &mkdir_p_or_exit($dpath);
    my $fpath = &join_paths($dpath,'PKGBUILD');
    if (open(my $fh,'>',$fpath))
    {
      print $fh $self->{CPAN}->pkgbuild_obj_to_pkgbuild($obj);
      close($fh);
    }
    else
    {
      print STDERR "error: failed to save PKGBUILD $fpath ($!)\n";
      exit(1);
    }
  }
  $self->maybe_source_taurball_files('cpan',@pkgnames);
}

sub get_cpan_build_dir
{
  my $self = shift;
  return &join_paths($self->get_value('BuildDir'),'cpan',@_);
}






########## Hackage ##########

# This can be improved once cabal2arch matures to more options.
sub maybe_load_hackage_pkgbuild_objs
{
  my ($self,@pkgnames) = @_;
  return if not @pkgnames;
  if (not $CABAL2ARCH)
  {
    print STDERR "error: this operation requires cabal2arch\n";
    exit(1);
  }
  my $hackage = $self->{Hackage};
  my $cabal_dpath = &join_paths($hackage->{CACHE},'cabal_files');
  &mkdir_p_or_exit($cabal_dpath);
  my @cabal_urls = map {$hackage->get_cabal_url($_)} @pkgnames;
  &parallel_mirror_urls($cabal_dpath,@cabal_urls);
  my $build_dir = $self->get_hackage_build_dir();
  &mkdir_p_or_exit($build_dir);
  foreach my $pkgname (@pkgnames)
  {
    my $pkg_name = $hackage->get_pkg_name($pkgname);
    my $cabal_fpath = &join_paths($cabal_dpath,$pkg_name.'.cabal');
    $self->run_cmd_as_builder("cd \"$build_dir\" && $CABAL2ARCH \"$cabal_fpath\" >/dev/null 2>&1");
  }
  $self->maybe_source_taurball_files('hackage',@pkgnames);
}

sub get_hackage_build_dir
{
  my $self = shift;
  return &join_paths($self->get_value('BuildDir'),'hackage',@_);
}







########## ABS ##########

sub save_abs_taurballs
{
  my ($self,$reponames,$arch) = @_;
  my @reponames = @$reponames;
  return () if not @reponames;

  my $dpath = $self->get_value('TaurballDir');
  &mkdir_p_or_exit($dpath);
  my @args = ();
  my @checklist = ();
  foreach my $reponame (@reponames)
  {
    my $a = $self->get_repo_archive($reponame,'abs',$arch);
    my $fpath = &join_paths($dpath,$a->{'filename'});
    if (-e $fpath and defined($self->{ABS}->{archives}->{$reponame}->{'last_check'})
    and (time() - $self->{ABS}->{archives}->{$reponame}->{'last_check'}) < $REFRESH_INTERVAL)
    {
      next;
    }
    else
    {
      push(@args,[$reponame,$a,$fpath]);
      push(@checklist,$reponame);
    }
  }
  if (@checklist)
  {
    my $list = join(' ',map {$self->get_reponame_str($_)} @checklist);
    $self->pretty_message("Checking for abs archive updates for the following repos: $list\n");
  }

  my $s = sub {
    my ($reponame,$a,$fpath) = @_;
    my $e = $a->mirror_content($fpath);
    return [$e,$reponame]
  };

  my @output = &parallelize($s,@args);

  my $t = time();
  $self->{ABS}->{archives}->{$_}->{'last_check'} = $t foreach @reponames;

  foreach my $reponame (@reponames)
  {
    my $fpath = &join_paths($self->get_value('TaurballDir'),$reponame.'.abs'.$Xyne::Arch::Pacman::ARCHIVE_EXT);
    my $mtime =  (stat($fpath))[9];
    if (not defined($self->{ABS}->{archives}->{$reponame}->{mtime}) or $mtime > $self->{ABS}->{archives}->{$reponame}->{mtime})
    {
      $self->{ABS}->{archives}->{$reponame} = {
      } if not defined($self->{ABS}->{archives}->{$reponame});
      $self->{ABS}->{archives}->{$reponame}->{mtime} = $mtime;

      my $filelist = $self->run_tar_as_builder("-ztf $fpath");

      $self->{ABS}->{archives}->{$reponame}->{filelist} = $filelist;
      # Each key contains the repo name in its path thus allowing all repos to
      # share the same namespace.
      $self->{ABS}->{pkglist} = {} if not defined($self->{ABS}->{pkglist});
      foreach my $line (split(/\s*\n\s*/,$filelist))
      {
        if ($line =~ m/^(.+)\/PKGBUILD$/)
        {
          # This may seem redundant here, but this will also be used to match
          # split packages to PKGBUILDs, thus providing a unified search.
          $self->{ABS}->{pkglist}->{$1} = $1;
        }
      }
    }
  }
  return @output;
}

sub save_and_extract_abs_taurballs_for_pkgnames
{
  my ($self,@repo_pkgnames) = @_;
  return if not @repo_pkgnames;
  my $groups = $self->group_pkgnames_by_repo(@repo_pkgnames);
  my @reponames = keys(%{ $groups });
  $self->save_abs_taurballs(\@reponames);
  $self->extract_abs_taurballs_for_repo_groups($groups);
}

sub extract_abs_taurballs_for_pkgnames
{
  my ($self,@pkgnames) = @_;
  return if not @pkgnames;
  my $groups = $self->group_pkgnames_by_repo(@pkgnames);
  $self->extract_abs_taurballs_for_repo_groups($groups);
}

sub extract_abs_taurballs_for_repo_groups
{
  my ($self,$groups) = @_;
  my $dpath = $self->get_value('BuildDir');

  while (my ($reponame,$pkgnames) = each(%{ $groups }))
  {
    my $fpath = &join_paths($self->get_value('TaurballDir'),$reponame.'.abs'.$Xyne::Arch::Pacman::ARCHIVE_EXT);
    my @pkgnames = @$pkgnames;
    my $args = '';
    # This is to match split packages to their PKGBUILDs.
    PKGNAME: foreach my $pkgname (@pkgnames)
    {
      my ($_pkgname) = &get_name_and_repo($pkgname);
      my $p = &join_paths($reponame,$_pkgname);
      if (defined($self->{ABS}->{pkglist}->{$p}))
      {
        $args .= $self->{ABS}->{pkglist}->{$p} . ' ';
      }
      else
      {
        my $pkgbase = $self->get_sync_db_pkg_pkgbase($pkgname);
        if (defined($pkgbase))
        {
          my $pb = &join_paths($reponame,$pkgbase);
          if (defined($self->{ABS}->{pkglist}->{$pb}))
          {
            $args .= $self->{ABS}->{pkglist}->{$pb} . ' ';
            $self->{ABS}->{pkglist}->{$p} = $self->{ABS}->{pkglist}->{$pb};
            next PKGNAME;
          }
        }
        print STDERR "error: could not find PKGBUILD for $_pkgname in $fpath\n";
        exit 1;
      }
    }
    $self->mkdir_p_or_exit($dpath);
    $self->run_tar_as_builder("-zxf $fpath -C $dpath --no-same-owner --no-same-permissions $args");
  }
}

sub get_abs_build_dir
{
  my ($self,$pkgname) = @_;
  my $reponame = $self->get_sync_db_pkg_reponame($pkgname);
  return &join_paths($self->get_value('BuildDir'),$reponame,$pkgname);
}













########## MISC ##########

sub set_arch
{
  my ($self,$arch) = @_;
  $self->{ARCH} = $arch if defined($arch);
}

sub mkdir_p_or_exit
{
  my ($dpath,@subdirs) = @_;
  $dpath = &join_paths($dpath,@subdirs) if @subdirs;
  if (not &mkdir_p($dpath))
  {
    print STDERR "error: failed to create $dpath\n";
    exit(1);
  }
}



sub get_devel_pkgnames
{
  my ($self) = @_;
  return grep {$_ =~ m/-(?:svn|cvs|hg|git|bzr|darcs)$/} $self->get_local_db_pkgname_list();
}



sub get_installation_status
{
  my ($self,$type,$pkgname,$pkgver) = @_;
  if ($type eq 'info')
  {
    if ($self->is_installed($pkgname))
    {
      if (defined($pkgver))
      {
        my $local_ver = $self->get_local_db_pkg_ver($pkgname);
        my $cmp = &compare_versions($local_ver,$pkgver);
        if ($cmp == 0)
        {
          return 'yes';
        }
        else
        {
          return $local_ver;
        }
      }
      else
      {
        return 'yes';
      }
    }
    else
    {
      return 'no';
    }
  }
  else
  {
    if ($self->is_installed($pkgname))
    {
      if (defined($pkgver))
      {
        my $local_ver = $self->get_local_db_pkg_ver($pkgname);
        my $cmp = &compare_versions($local_ver,$pkgver);
        if ($cmp == 0)
        {
          return ' (installed)';
        }
        else
        {
          return " (installed: $local_ver)";
        }
      }
      else
      {
        return ' (installed)';
      }
    }
    else
    {
      return '';
    }
  }
}










########## PERMISSIONS ##########

sub get_unprivileged_user
{
  my ($self) = @_;
  my ($ruid,$rgid) = (getpwnam('root'))[2,3];
  my $uid = $ruid;
  my $gid = $rgid;
  my $user = $self->get_value('BuildAs');
  if (not defined($user) and defined($ENV{SUDO_USER}))
  {
    $user = $ENV{SUDO_USER};
  }
  my $was_unset = 0;
  if (defined($user))
  {
    ($uid,$gid) = (getpwnam($user))[2,3];
  }
  else
  {
    print "No unprivileged user specified with the \"BuildAs\" option\n";
  }
  while (not $uid or not $gid or $uid == $ruid or $gid == $rgid)
  {
    print "Please specify an unprivileged user for this operation: ";
    $user = <STDIN>;
    next if not defined($user);
    chomp $user;
    $self->{PowerpillConf}->set_value('BuildAs',$user);
    ($uid,$gid) = (getpwnam($user))[2,3];
  }
  return $user;
}

sub get_unprivileged_user_if_root
{
  my ($self) = @_;
  my ($ruid,$rgid) = (getpwnam('root'))[2,3];
  if ($< == $ruid or $( == $rgid)
  {
    return $self->get_unprivileged_user;
  }
  else
  {
    return undef;
  }
}

sub drop_root_privileges_or_exit
{
  my ($self) = @_;
  my $user = $self->get_unprivileged_user_if_root;
  if (defined($user))
  {
    my ($uid,$gid) = (getpwnam($user))[2,3];
    $) = $gid;
    $> = $uid;
  }
}

sub chown_to_builder
{
  my ($self,@paths) = @_;
  return if not @paths;
  my $user = $self->get_unprivileged_user_if_root;
  if (defined($user))
  {
    my ($uid,$gid) = (getpwnam($user))[2,3];
    chown $uid,$gid,@paths;
  }
}

sub grant_access_to_builder
{
  my ($self,$path,$umask) = @_;
  return if not defined($path);
  my $user = $self->get_unprivileged_user_if_root;
  if (defined($user))
  {
    my ($uid,$gid) = (getpwnam($user))[2,3];
    &unsu();
    find(
      sub{
        &chmod_for_builder($_,$umask);
        chown $uid,$gid,$_;
      },
      $path
    );
    $self->drop_root_privileges_or_exit();
  }
}

sub chmod_for_builder
{
  my ($path,$umask) = @_;
  return if not $path;
  my $mode = (stat($path))[2];
  return if not defined($mode);
  $umask = umask if not defined($umask);
  $umask = 022 if not defined($umask);
  # if a directory
  if ($mode & 040000)
  {
    $mode = ($mode | 0777) ^ $umask;;
  }
  else
  {
    $mode = ($mode | 0666) ^ $umask;;
  }
  chmod $mode, $path;
}

sub find_chmod_for_builder
{
  find(sub{&chmod_for_builder($_);},@_);
}


sub su
{
  my ($username) = @_;
  my ($uid,$gid) = (getpwnam($username))[2,3];
  $) = $gid;
  $> = $uid;
  $) = $gid;
}

sub unsu
{
  $> = $<;
  $) = $(;
}

sub maybe_log
{
  my ($self,@msgs) = @_;
  my $lpath = $self->get_value('Log');
  if (defined($lpath))
  {
    # preserve permissions from entry on exit, but elevate if possible for logging
    my $drop;
    if ($< != $>)
    {
      &unsu();
      $drop = 1;
    }
    $self->{LOG} = Xyne::Common::Log->new($lpath) if not $self->{LOG};
    $self->{LOG}->log(@msgs);
    $self->drop_root_privileges_or_exit() if $drop;
  }
}

sub run_cmd_as_builder
{
  my ($self,$cmd) = @_;
  my $builder = $self->get_unprivileged_user_if_root();
  if (defined($builder))
  {
    $cmd = "su $builder -c '$cmd'";
  }
  return system($cmd);
}

1;

__END__

=head1 NAME

Xyne::Arch::Bauerbill - A pacman wrapper and download-accelerator class with ABS
and AUR support.

=head1 SYNOPSIS

  use Xyne::Arch::Bauerbill;

=head1 DESCRIPTION

Bauerbill extends the functionality of Powerpill by adding support for building
packages from the AUR and ABS.

=head2 EXPORT

None by default.

=head1 AUTHOR

Xyne <lt>xyne at archlinux dot ca<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009,2010  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.