# Copyright (C) 2010  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


package Xyne::Arch::CPAN;
use strict;
use warnings;

require Exporter;
our @ISA = qw/Exporter/;
our @EXPORT_OK = qw/module_to_pkgname path_to_distribution_and_version format_perl_ver format_cpan_ver/;

use Xyne::Arch::Pacman qw/compare_versions get_db_pkg_filename_by_section parse_db_file format_db_file save_db_file/;
use Xyne::Common::Functions qw/confirm join_paths refpush str find_files mkdir_p/;
use Xyne::Common::Threads qw/parallel_mirror_urls/;

use File::Basename qw/basename dirname/;
use IO::Uncompress::Gunzip qw/gunzip/;
use LWP::Simple;
use Parse::CPAN::Meta;

my $DISTRIBUTION_LIST_URL = 'http://www.cpan.org/modules/01modules.index.html';
my $MODULE_LIST_URL = 'http://www.cpan.org/modules/02packages.details.txt.gz';
my $CPAN_DATABASE_FILE = 'CPAN_database';
my $META_YAML_BASE_URL = 'http://search.cpan.org/src/';
my $SRC_BASE_URL = 'http://search.cpan.org/CPAN/authors/id/';
my $DIST_BASE_URL = 'http://search.cpan.org/dist/';

my @ARCHITECTURES = qw/i686 x86_64/;
my @ARCHIVE_EXTS = qw/.bz2 .gz .tar.gz .tgz .zip/;
@ARCHIVE_EXTS = sort {length($b) <=> length($a)} @ARCHIVE_EXTS;
my @S_INC = sort {length($b) <=> length($a)} @INC;

# for parsing the distribution list
my @MONTHS = qw/Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec/;
my %MONTHS_TO_INT;
@MONTHS_TO_INT{@MONTHS} = map {sprintf "%02d", $_} (1..12);

my %CUSTOM_PKGNAMES = (
  'perl' => 'perl',
  'libwww-perl' => 'perl-libwww',
  'perl-ldap' => 'perl-ldap'
);

our $VERSION = '0.07';






sub new {
  my ($proto,$args) = @_;
  my $class = ref($proto) || $proto;
  my $self  = {};
  bless ($self, $class);

  if (defined($args->{Pacman}))
  {
    $self->{Pacman} = $args->{Pacman};
  }
  else
  {
    $self->{Pacman} = Xyne::Arch::Pacman->new({
      autoload => $args->{'pacman_autoload'},
      config_path => $args->{'pacman_config_path'}
    });
  }


  if (defined($args->{Cache}))
  {
    $self->{CACHE} = $args->{Cache};
  }
  else
  {
    $self->{CACHE} = '/tmp/bauerbill/pacpan';
  }

  return $self;
}








########## OPERATIONS ##########

sub get_upgradable
{
  my ($self,$ignore,@pkgnames) = @_;
  my $pacman = $self->{Pacman};
  if (not @pkgnames)
  {
    foreach my $d (keys %{ $self->{Distributions} })
    {
      my $pkgname = &module_to_pkgname($d);
      push (@pkgnames,$pkgname) if $pacman->is_in_local_db($pkgname);
    }
  }
  @pkgnames = grep {not $pacman->should_ignore($_)} @pkgnames if $ignore;
  my @upgradable = ();
  foreach my $pkgname (@pkgnames)
  {
    my $d = $self->get_distribution($pkgname);
    my $cpan_ver = $self->{Distributions}->{$d}->{version};
    next if $cpan_ver eq 'undef';
    my $local_ver = $pacman->get_local_db_pkg_ver($pkgname);
    my $cmpver = &compare_versions($cpan_ver,$local_ver);
    push(@upgradable,$pkgname) if $cmpver == 1;
  }
  return @upgradable;
}









########## Module Methods ##########

sub maybe_update_database
{
  my ($self,$force) = @_;
  my @urls = ($DISTRIBUTION_LIST_URL,$MODULE_LIST_URL);

  if ($force)
  {
    foreach my $url (@urls)
    {
      my $fpath = &join_paths($self->{CACHE},basename($url));
      unlink $fpath if -e $fpath;
    }
  }
  my @responses = &parallel_mirror_urls($self->{CACHE},@urls);

  my $is_new = 0;
  foreach my $response (@responses)
  {
    if ($response->[0] == 200)
    {
      $is_new = 1;
    }
    elsif ($response->[0] != 304)
    {
      print STDERR "error: failed to mirror $response->[1] (response code: $response->[0])\n";
      exit(1);
    }
  }
  my $fpath = &join_paths($self->{CACHE},$CPAN_DATABASE_FILE);
  $is_new = 1 if not -f $fpath;
  if ($is_new)
  {
    $self->format_and_save_database_list();
  }
  return $is_new;
}



sub maybe_update_and_load_database
{
  my ($self,$force) = @_;
  my $is_new = $self->maybe_update_database($force);
  $self->load_database if not $is_new;
  return $is_new;
}



sub load_distribution_list
{
  my ($self) = @_;
  my $fpath = &join_paths($self->{CACHE}, basename($DISTRIBUTION_LIST_URL));
  my $distributions;
  if (open (my $fh, '<', $fpath))
  {
    while (defined(my $line = <$fh>))
    {
      if ($line =~ m/\.\.\/authors\/id\/([A-Z]\/[A-Z]{2}\/[A-Z]+\/[^"]+)".*(\d\d) (\w{3}) (\d{4})/)
      {
        my $p = $1;
        my ($n,$v) = &path_to_distribution_and_version($p);
        my $d = $4.$MONTHS_TO_INT{$3}.$2;
        if (exists $distributions->{$n} and $distributions->{$n}->{date} >= $d)
        {
          next;
        }
        $distributions->{$n} = {
          date => $d,
          path => $p
        };
      }
    }
    close($fh);
    return $distributions;
  }
  else
  {
    print STDERR "error: failed to load $fpath ($!)\n";
    exit(1);
  }
}



sub format_and_save_database_list
{
  my ($self) = @_;

  my $distributions = $self->load_distribution_list();
  my $fpath = &join_paths($self->{CACHE}, basename($MODULE_LIST_URL));
  my $list;
  &gunzip($fpath,\$list);

  my $modules = {};
  my $sources = {};

  foreach my $line (split("\n",$list))
  {
    my @columns = split(/\s+/,$line);
    if (scalar(@columns) == 3)
    {
      my ($module,$version,$path) = @columns;
      $modules->{$module} = &format_cpan_ver($version);
      &refpush($sources, $path, $module);
      #$self->{PkgnameToModule}->{&module_to_pkgname($module)} = $module;
    }
  }

  foreach my $path (keys(%$sources))
  {
    my ($name,$version) = &path_to_distribution_and_version($path);

    my $bname = basename($path);
    my @modules = @{$sources->{$path}};
    my @candidate = ();

    my $is_distro = 0;

    # If this is a distribution (including "perl"), use the distributions name
    # for the package and flag it to take precedence over non-distributions.
    if
    (
      (
        exists($distributions->{$name})
        and
        $distributions->{$name}->{path} eq $path
      )
      or
      substr($path,0,15) eq 'D/DA/DAPM/perl-'
    )
    {
      $is_distro = 1;
    }
    else
    {
      # This relies on the list being in alphabetical order, which means that
      # the first module to match the path name will also be the highest up
      # in the hierarchy.
      foreach my $module (@modules)
      {
        my $fname = &module_to_pathname($module);
        # If we find a module name that matches the source code file, then we
        # consider it the distribution name.
        if (index($bname,$fname) eq 0)
        {
          $name = &module_to_pathname($module);
          last;
        }
        # Otherwise we need to find the most appropriate name of among the modules
        # which are part of a common distribution.
        my @components = split('::',$module);
        if (@candidate)
        {
          my $a = index($bname,$components[0]) == 0;
          my $b = index($bname, $candidate[0]) == 0;
          my $l = scalar(@components);
          if (
            # If the first component matches the path name and the candidate doesn't
            ($a and not $b) or
            # or both match and the candidate is longer
            ($a and $b and $l < scalar(@candidate)) or
            lc($components[0]) eq 'bundle'
          )
          {
            # then replace the candidate
            @candidate = @components;
          }
          elsif ($candidate[0] eq $components[0])
          {
            my @new_candidate = ();
            my $n = 0;
            while (defined($candidate[$n])
            and defined($components[$n])
            and $candidate[$n] eq $components[$n])
            {
              push(@new_candidate,$candidate[$n]);
              $n++;
            }
            @candidate = @new_candidate;
          }
        }
        else
        {
          @candidate = @components;
        }
      }
      $name = join('-', @candidate) if not $name;
      if ($version eq 'undef' and exists($modules->{$name}))
      {
        $version = $modules->{$name};
      }
    }
    if (not exists($self->{Distributions}->{$name})
    or $is_distro)
    {
      # Make sure that pkgnames map to the most recent distribution.
      # This is necessary because of name collisions due to the pkgname being
      # all lowercase.
      my $pkgname = &module_to_pkgname($name);
      if (not defined($self->{PkgnameToDistribution}->{$pkgname}))
      {
        $self->{PkgnameToDistribution}->{$pkgname} = $name;
      }
      else
      {
        my $other_name = $self->{PkgnameToDistribution}->{$pkgname};
        if ($name ne $other_name)
        {
          my $a = $distributions->{$name}->{date} if exists($distributions->{$name});
          $a = 0 if not defined($a);
          my $b = $distributions->{$other_name}->{date} if exists($distributions->{$other_name});
          $b = 0 if not defined($b);

          if ($a > $b)
          {
            $self->{PkgnameToDistribution}->{$pkgname} = $name;
          }
          else
          {
            next;
          }
        }
      }
      $self->{Distributions}->{$name} = {
        path => $path,
        version => $version,
        modules => {}
      };
      foreach my $m (@modules)
      {
        $self->{Distributions}->{$name}->{modules}->{$m} = $modules->{$m};
        $self->{ModuleToDistribution}->{$m} = $name;
        my $pkgname = &module_to_pkgname($m);
        $self->{PkgnameToDistribution}->{$pkgname} = $name if not defined($self->{PkgnameToDistribution}->{$pkgname});
      }
    }
  }
  if (open(my $fh, '>', &join_paths($self->{CACHE},$CPAN_DATABASE_FILE)))
  {
    foreach my $d (sort keys %{$self->{Distributions}})
    {
      my $version = $self->{Distributions}->{$d}->{version};
      next if $version eq 'undef';
      print $fh $d;
      print $fh " ";
      print $fh $version;
      print $fh " ";
      print $fh $self->{Distributions}->{$d}->{path};
      foreach my $m (sort keys %{$self->{Distributions}->{$d}->{modules}})
      {
        next if $m eq $d;
        my $v = $self->{Distributions}->{$d}->{modules}->{$m};
        print $fh " $m $v";
      }
      print $fh "\n"
    }
    close($fh);
  }
}



sub maybe_load_database
{
  my ($self) = @_;
  $self->load_database if not defined $self->{Distributions};
}



sub load_database
{
  my ($self) = @_;
  my $fpath = &join_paths($self->{CACHE},$CPAN_DATABASE_FILE);
  if (not -f $fpath)
  {
    $self->maybe_update_and_load_database;
  }
  else
  {
    if (open(my $fh, '<', $fpath))
    {
      while(defined(my $line = <$fh>))
      {
        chomp $line;
        my ($distribution,$version,$path,@modules) = split(' ', $line);
        my $href = {
          version => $version,
          path => $path,
          modules => {}
        };
        $self->{PkgnameToDistribution}->{&module_to_pkgname($distribution)} = $distribution;
        #$self->{PkgnameToModule}->{&module_to_pkgname($distribution)} = $distribution;
        my $modules = {};
        while(@modules)
        {
          my $module = shift(@modules);
          my $version = shift(@modules);
          $href->{modules}->{$module} = $version;
          $self->{ModuleToDistribution}->{$module} = $distribution;
          #$self->{PkgnameToModule}->{&module_to_pkgname($module)} = $module;
          my $pkgname = &module_to_pkgname($module);
          $self->{PkgnameToDistribution}->{$pkgname} = $distribution if not defined($self->{PkgnameToDistribution}->{$pkgname});
        }
        $self->{Distributions}->{$distribution} = $href;
      }
      close($fh);
    }
  }
}



sub mirror_and_load_module_list
{
  my ($self) = @_;
  my $e = $self->mirror_module_list;
  if ($e == 304)
  {
    $self->load_module_list;
  }
}



########## Metadata ##########


sub get_distribution
{
  my ($self,$whatever) = @_;
  if (substr($whatever,0,5) eq 'perl-' and exists($self->{PkgnameToDistribution}->{$whatever}))
  {
    $whatever = $self->{PkgnameToDistribution}->{$whatever};
  }
  return $whatever if exists($self->{Distributions}->{$whatever});
  if (exists($self->{ModuleToDistribution}->{$whatever}))
  {
    $whatever = $self->{ModuleToDistribution}->{$whatever};
    return $whatever if exists($self->{Distributions}->{$whatever});
  }
  return undef;
}



sub get_distribution_url
{
  my ($self,$d) = @_;
  return &join_paths($DIST_BASE_URL,&module_to_pathname($d))
}



sub get_meta_yaml_url
{
  my ($self,$distribution) = @_;
  my $obj = $self->{Distributions}->{$distribution};

  my $path = $obj->{path};
  my ($author,$filename) = ($path =~ m/^\w\/\w{2}\/(\w+)\/.*?([^\/]+)$/);
  my ($dir) = ($filename =~ m/^([^\.]+(?:\.\d+)*)/);
  return &join_paths($META_YAML_BASE_URL,$author,$dir,'META.yml');
}



sub get_meta_yaml_fpath
{
  my ($self,$distribution) = @_;
  my $version = $self->{Distributions}->{$distribution}->{version};
  return &join_paths($self->{CACHE},'META.yml',&module_to_pathname($distribution),$version);
}



sub maybe_mirror_meta_yaml
{
  my ($self,$distribution) = @_;
  my $fpath = $self->get_meta_yaml_fpath($distribution);
  return (1,$distribution) if -f $fpath;
  my ($e) = $self->mirror_meta_yaml($distribution);
  return (($e == 200),$distribution);# or $e == 304); # should never be 304
}



sub mirror_meta_yaml
{
  my ($self,$distribution) = @_;
  my $meta_url = $self->get_meta_yaml_url($distribution);
  my $fpath = $self->get_meta_yaml_fpath($distribution);
  my $dpath = dirname($fpath);
  if (-d $dpath)
  {
    if (opendir(my $dh,$dpath))
    {
      my $filename = basename($fpath);
      foreach my $i (readdir($dh))
      {
        next if $i eq '.' or $i eq '..';
        unlink &join_paths($dpath,$i) if $i ne $filename;
      }
      closedir($dh);
    }
  }
  else
  {
    if (not &mkdir_p($dpath))
    {
      print STDERR "error: failed to create $dpath\n";
      exit(1);
    }
  }
  my $e = &mirror($meta_url,$fpath);
  # If the file has changed, the corresponding checksums may have also changed
  # so we flag the checksums file to make sure that it's mirrored upon next access.
  if ($e != 304)
  {
    my ($a,$f) = $self->get_author_path_and_filename($distribution);
    my $fp = $self->get_checksums_fpath($a).'.check';
    if (not -f $fp)
    {
      my $fh;
      open($fh,'>',$fp) and close($fh);
    }
  }
  return ($e,$distribution);
}



sub get_meta_yaml_obj
{
  my ($self,$distribution) = @_;
  my $fpath = $self->get_meta_yaml_fpath($distribution);
  # The eval block prevents the process from dying horribly, which is problematic
  # with incorrect META.yml files and thread.
  no warnings;
  my @yaml;
  eval("\@yaml = \&Parse::CPAN::Meta::LoadFile(\$fpath)");
  use warnings;
  if (not defined($yaml[0]))
  {
    #print STDERR "error: failed to parse $fpath\n";
    #print STDERR "       Inform the CPAN author that there is an error in\n       $distribution\'s META.yml file\n\n";
    #exit(1);
    $self->maybe_display_error("failed to parse $fpath\n", "inform the CPAN author that there is an error in ${distribution}'s META.yml file\n");
  }
  return $yaml[0];
}



sub get_author_path_and_filename
{
  my ($self,$distribution) = @_;
  my $path = $self->{Distributions}->{$distribution}->{path};
  my ($a,$f) = ($path =~ m/^(\w\/\w{2}\/\w+)\/(.+)$/);
  return [$a,$f];
}



sub get_checksums_url
{
  my ($self,$apath) = @_;
  return &join_paths($SRC_BASE_URL,$apath,'CHECKSUMS');
}



sub get_checksums_fpath
{
  my ($self,$apath) = @_;
  return &join_paths($self->{CACHE},'CHECKSUMS',$apath,'CHECKSUMS');
}



sub maybe_mirror_checksums
{
  my ($self,$apath) = @_;
  my $fpath = $self->get_checksums_fpath($apath);
  my $flagged = -f $fpath.'.check';
  return (1,$apath) if -f $fpath and not $flagged;
  my ($e) = $self->mirror_checksums($apath);
  return (($e == 200 or $e == 304),$apath);
}



sub mirror_checksums
{
  my ($self,$apath) = @_;
  my $url = $self->get_checksums_url($apath);
  my $fpath = $self->get_checksums_fpath($apath);
  my $flag = $fpath.'.check';
  my $flagged ;
  my $dpath = dirname($fpath);
  if (-d $dpath)
  {
    $flagged = -f $flag;
  }
  else
  {
    $flagged = 0;
    if (not &mkdir_p($dpath))
    {
      print STDERR "error: failed to create $dpath\n";
      exit(1);
    }
  }
  my $e = &mirror($url,$fpath);
  unlink $flag if $flagged and ($e == 200 or $e == 304);
  return ($e,$apath);
}



sub get_checksums_obj
{
  my ($self,$apath) = @_;
  $self->maybe_mirror_checksums($apath);
  my $fpath = $self->get_checksums_fpath($apath);
  my $obj = {};
  # Read the file instead of sourcing it to avoid a security hole.
  if (open(my $fh, '<', $fpath))
  {
    my $section = 0;
    while(defined(my $line = <$fh>))
    {
      if (not $section)
      {
        if ($line =~ m/^  '(.+)' => \{$/)
        {
          $section = $1;
        }
      }
      else
      {
        if ($line =~ m/^    '(.+?)' => '(.+?)',?$/)
        {
          $obj->{$section}->{$1} = $2;
        }
        elsif ($line =~ m/^  \},?$/)
        {
          undef $section;
        }
      }
    }
    close($fh);
  }
  return $obj;
}


# The overhead of thread creation each time this is run is not worth the benefits
# that threading provides.
sub get_pkgbuild_objs
{
  my ($self,@whatever) = @_;
  my @objs = ();
  my %distributions;
  my %apaths;
  my %failed;
  $distributions{$_} = 1 foreach grep {$_} map {$self->get_distribution($_)} @whatever;
  $apaths{$_} = 1 foreach map {$self->get_author_path_and_filename($_)->[0]} keys %distributions;

#  my @ss;
  foreach my $k (keys %distributions)
  {
    my ($e) = $self->maybe_mirror_meta_yaml($k);
    next if $e == 1;
    #delete $distributions{$k};
    $failed{$self->get_meta_yaml_url($k)} = $e;
  }
  foreach my $k (keys %apaths)
  {
    my ($e) = $self->maybe_mirror_checksums($k);
    next if $e == 1;
    #delete $apaths{$k};
    $failed{$self->get_checksums_url($k)} = $e;
  }

  if (%failed)
  {
    my $error_msg = "failed to download the following:\n";
    $error_msg .= '  '.$_.' (response: '.$failed{$_}.")\n" foreach keys %failed;
    $self->maybe_display_error($error_msg);
    #exit(1);
  }

  my $checksums;
  foreach my $a (keys %apaths)
  {
    $checksums->{$a} = $self->get_checksums_obj($a);
  }
  foreach my $d (keys %distributions)
  {
    my $y = $self->get_meta_yaml_obj($d);
    my ($a,$f) = @{ $self->get_author_path_and_filename($d) };
    my $c;
    if (exists($checksums->{$a}->{$f}))
    {
      $c = $checksums->{$a}->{$f};
    }
    else
    {
      $c = {};
    }
    push @objs, $self->meta_yaml_obj_to_pkgbuild_obj($d,$y,$c);
  }
  return @objs;
}



sub meta_yaml_obj_to_pkgbuild_obj
{
  my ($self,$distribution,$yaml,$checksums) = @_;
  my $modules = $self->{Distributions}->{$distribution}->{modules} if exists($self->{Distributions}->{$distribution}->{modules});
  my $pkgname = &module_to_pkgname($distribution);
  my $obj = {};
  $obj->{'META.yml error'} = 1 if not defined($yaml);
  $obj->{pkgname} = $pkgname;
  $obj->{url} = &join_paths($DIST_BASE_URL,&module_to_pathname($distribution)).'/';
  my $ver_a = &format_cpan_ver($yaml->{version});
  my $ver_b = $self->{Distributions}->{$distribution}->{version};
  my $pkgver = (defined($ver_a) and &Xyne::Arch::Pacman::compare_versions($a,$b) == 1) ? $ver_a : $ver_b;
  $obj->{pkgver} = $pkgver;
  $obj->{pkgrel} = 1;
  $obj->{pkgdesc} = defined($yaml->{abstract}) ? $yaml->{abstract} : "CPAN package (no description found)";
  $obj->{arch} = [@ARCHITECTURES];
  $obj->{license} = defined($yaml->{license}) ? [$yaml->{license}] : ["unknown"];
  $obj->{source} = [&join_paths($SRC_BASE_URL,$self->{Distributions}->{$distribution}->{path})];
  $obj->{md5sums} = [$checksums->{md5}] if defined($checksums->{md5});
  $obj->{sha256sums} = [$checksums->{sha256}] if defined($checksums->{sha256});


  $obj->{depends} = [$self->href_to_pkgname_array($yaml->{requires},'>=')] if exists($yaml->{requires});
  my %makedepends = ();
  %makedepends = %{ $yaml->{build_requires} } if defined($yaml->{build_requires});
  %makedepends = (%{ $yaml->{configure_requires} }, %makedepends) if defined($yaml->{configure_requires});
  $obj->{makedepends} = [$self->href_to_pkgname_array(\%makedepends,'>=')] if %makedepends;
  $obj->{optdepends} = [$self->href_to_pkgname_array($yaml->{recommends})] if defined($yaml->{recommends});
  if (defined($modules))
  {
    my $provides = {};
    my $not_empty = 0;
    foreach my $m (keys %$modules)
    {
      next if &module_to_pkgname($m) eq $pkgname;
      $provides->{$m} = $modules->{$m};
      $not_empty = 1;
    }
    $obj->{provides} = [$self->href_to_pkgname_array($provides,'=')] if $not_empty;
  }
  return $obj;
}



sub href_to_pkgname_array
{
  my ($self,$modules,$ineq) = @_;
  my %pkgnames;
  while (my ($m,$v) = each(%$modules))
  {
    # Check that whatever has been specified maps to a valid distribution. If not,
    # ignore it. This deals with META.yml file which specify "Configure" in their
    # "requires" section, among other things.
    my $d = $self->get_distribution($m);
    if (not defined($d))
    {
      $self->maybe_display_warning("ignoring $m which does not map to a CPAN distribution\n");
      next
    }
    $m = &module_to_pkgname($m);
    $v = undef if not $v or $v eq 'undef';
    if ($v)
    {
      # Rarely needed but a showstopper if it's missed.
      $v =~ s/^(?:>=|>|=|<=|<\s+)//;
      $v =~ s/\s+/_/g;
      $v = '0'.$v if substr($v,0,1) eq '.';
    }
    if ($v)
    {
      $v = ($m eq 'perl') ? &format_perl_ver($v) : &format_cpan_ver($v);
    }
    if (exists($pkgnames{$m}))
    {
      next if not $v or &compare_versions($pkgnames{$m},$v) == 1;
    }
    $pkgnames{$m} = $v;
  }

  my @pkgnames = ();
  while (my ($p,$v) = each(%pkgnames))
  {
    $p .= $ineq.$v if $ineq and $v;
    push(@pkgnames,$p);
  }
  return sort @pkgnames;
}



sub get_pkgbuild
{
  my ($self,$whatever) = @_;
  my ($obj) = $self->get_pkgbuild_objs($whatever);
  return undef if not $obj;
  return $self->pkgbuild_obj_to_pkgbuild($obj);
}



sub pkgbuild_obj_to_pkgbuild
{
  my ($self,$obj) = @_;
  my $pkgbuild = "# Generated by Xyne::Arch::CPAN $VERSION\n\n";
  if (defined($obj->{'META.yml error'}))
  {
    my $d = $self->get_distribution($obj->{pkgname});
    my $url = $self->get_meta_yaml_url($d);
    $pkgbuild .= <<WARNING;
################################################################################
# WARNING: An error occured when either retrieving or parsing the META.yml file
# for this package. The information below is probably incomplete. Please contact
# the author on CPAN and kindly inform him/her that this file is missing,
# incorrect or malformatted:
#   CPAN Distribution: $d
#   Distribution url: $obj->{url}
#   META.yml url:  $url
################################################################################

WARNING
  }
  $pkgbuild .= "pkgname=$obj->{pkgname}\n";
  $pkgbuild .= "pkgver=$obj->{pkgver}\n";
  $pkgbuild .= "pkgrel=$obj->{pkgrel}\n";

  my $pkgdesc = $obj->{pkgdesc};
  $pkgdesc =~ s/"/\\"/g;
  $pkgbuild .= "pkgdesc=\"$pkgdesc\"\n";

  $pkgbuild .= "arch=('".join(' ',@{$obj->{arch}})."')\n";

  $pkgbuild .= "url=\"$obj->{url}\"\n";

  my $license = join(' ',map{"'$_'"} @{$obj->{license}});
  $pkgbuild .= "license=($license)\n";

  if (defined($obj->{source}))
  {
    my $source = join(' ',map{"'$_'"} @{$obj->{source}});
    $pkgbuild .= "source=($source)\n";
  }

  if (defined($obj->{md5sums}))
  {
    my $md5sums = join(' ',map{"'$_'"} @{$obj->{md5sums}});
    $pkgbuild .= "md5sums=($md5sums)\n";
  }

  if (defined($obj->{sha256sums}))
  {
    my $sha256sums = join(' ',map{"'$_'"} @{$obj->{sha256sums}});
    $pkgbuild .= "sha256sums=($sha256sums)\n";
  }

  if (defined($obj->{depends}))
  {
    my $depends = join(' ',map{"'$_'"} @{$obj->{depends}});
    $pkgbuild .= "depends=($depends)\n";
  }

  if (defined($obj->{makedepends}))
  {
    my $makedepends = join(' ',map{"'$_'"} @{$obj->{makedepends}});
    $pkgbuild .= "makedepends=($makedepends)\n";
  }

  if (defined($obj->{optdepends}))
  {
    my $optdepends = join(' ',map{"'$_'"} @{$obj->{optdepends}});
    $pkgbuild .= "optdepends=($optdepends)\n";
  }

  if (defined($obj->{provides}))
  {
    my $provides = join(' ',map{"'$_'"} @{$obj->{provides}});
    $pkgbuild .= "provides=($provides)\n";
  }

  $pkgbuild .= "options=(!emptydirs)\n";

  $pkgbuild .= "\n";
  $pkgbuild .= <<'BUILD';
build() {
  _dir=$(find $srcdir -maxdepth 2 -type f -name 'Makefile.PL')
  if [ ! -z "$_dir" ]; then
    cd $(dirname "$_dir")
    PERL_MM_USE_DEFAULT=1 perl Makefile.PL INSTALLDIRS=vendor || return 1
    make  || return 1
    make install DESTDIR="${pkgdir}" || return 1

  else
  _dir=$(find $srcdir -maxdepth 2 -type f -name 'Build.PL')
  if [ ! -z "$_dir" ]; then
    cd $(dirname "$_dir")
    PERL_MM_USE_DEFAULT=1 perl Build.PL INSTALLDIRS=vendor || return 1
    ./Build  || return 1
    ./Build install destdir=${pkgdir} || return 1

  else
    echo "error: failed to detect build method for $pkgname"
    echo "you may be able to fix this by editing the PKGBUILD"
    return 1
  fi fi

  # remove perllocal.pod and .packlist
  find ${pkgdir} -name perllocal.pod -delete
  find ${pkgdir} -name .packlist -delete
}
BUILD
  $pkgbuild .= "\n";


  $pkgbuild .= "# vim:set ts=2 sw=2 et:\n";
  return $pkgbuild;
}


########## Package Methods ##########

sub local_module_to_pkg
{
  my ($self,$module) = @_;
  my $owner;
  my $mfile = $module;
  $mfile =~ s/::/\//g;
  $mfile .= '.pm';
  foreach my $dpath (@INC)
  {
    my $mpath = &join_paths($dpath,$mfile);
    if (-e $mpath)
    {
      $owner = $self->{Pacman}->get_local_db_owner(substr($mpath,1));
      last if defined($owner);
    }
  }
  $self->{ModuleToPkgname}->{$module} = $owner;
  return $owner;
}



# This is arranged so that sync db packages always take priority.
sub module_to_pkg
{
  my ($self,$module) = @_;
  return undef if not defined($module);
  if (exists($self->{ModuleToPkgname}->{$module}))
  {
    return $self->{ModuleToPkgname}->{$module};
  }

  my $p = $self->{Pacman};
  my $distribution = $self->get_distribution($module);
  my $dcandidate;
  my $candidate;
  my $pkgname;
  if ($distribution)
  {
    $dcandidate = &module_to_pkgname($distribution);
    $pkgname = $p->is_effectively_in_sync_db($dcandidate);
    # If the distribution was found, map all its modules to that package;
    if ($pkgname)
    {
      $self->{ModuleToPkgname}->{$distribution} = $pkgname;
      $self->{ModuleToPkgname}->{$_} = $pkgname foreach keys(%{ $self->{Distributions}->{$distribution}->{'modules'} });
      return $pkgname;
    }
  }

  $candidate = &module_to_pkgname($module);
  $pkgname = $p->is_effectively_in_sync_db($candidate);
  if ($pkgname)
  {
    $self->{ModuleToPkgname}->{$module} = $pkgname;
    return $pkgname;
  }

  if ($dcandidate)
  {
    $pkgname = $p->is_effectively_in_local_db($dcandidate);
    # If the distribution was found, map all its modules to that package;
    if ($pkgname)
    {
      $self->{ModuleToPkgname}->{$distribution} = $pkgname;
      $self->{ModuleToPkgname}->{$_} = $pkgname foreach keys(%{ $self->{Distributions}->{$distribution}->{'modules'} });
      return $pkgname;
    }
  }

  $pkgname = $p->is_effectively_in_local_db($candidate);
  $pkgname = $self->local_module_to_pkg($module) if not $pkgname;
  $pkgname = $dcandidate if not $pkgname;
  $self->{ModuleToPkgname}->{$module} = $pkgname;
  return $pkgname;
}



#sub pkgname_to_module
#{
#  my ($self,$pkgname) = @_;
#  return exists($self->{PkgnameToModule}->{$pkgname}) ? $self->{PkgnameToModule}->{$pkgname} : undef;
#}



sub detect_local_packages
{
  my ($self) = @_;
  my $pkgnames;
  my %matched;
  foreach my $dpath (@INC)
  {
    next if $dpath eq '.';
    foreach my $file (&find_files($dpath))
    {
      next if defined($matched{$file});
      next if substr($file,-3) ne '.pm';
      my $owner = $self->{Pacman}->get_local_db_owner(substr($file,1));
      if (defined($owner))
      {
        $pkgnames->{$owner} = 1;
        $matched{'/'.$_}=1 foreach @{ $self->{Pacman}->get_local_db_pkg($owner)->{'get'}->('FILES') };
      }
    }
  }
  return keys %$pkgnames;
}



sub check_local_pkgs
{
  my ($self,$fix,$noconfirm) = @_;
  foreach my $pkgname (sort $self->detect_local_packages)
  {
    $| = 1;
    my $distribution = $self->get_distribution($pkgname);
    my $section = $self->{Pacman}->get_local_db_pkg($pkgname)->{'get'}->('PROVIDES');
    my @provides = ref($section) eq 'ARRAY' ? sort @{ $section } : ();

    my %provides;
    foreach my $p (@provides)
    {
      my ($n,$v) = ($p =~ m/^([^>=<]+)(.*)$/);
      next if $n eq $pkgname;
      $v = '' if not $v;
      $provides{$n} = $v;
    }
    my @cpan_provides = sort $self->get_installed_pkg_provides_array($pkgname);
    my %missing;
    my $found_missing;
    my %correct_names;
    foreach my $cp (@cpan_provides)
    {
      my ($n,$v) = ($cp =~ m/^([^>=<]+)(.*)$/);
      my $d = $self->get_distribution($n);
      my $rn = &module_to_pkgname($d);
      $correct_names{$rn}=1;
      next if $n eq $pkgname;
      $v = '' if not $v;
      if (exists($provides{$n}))
      {
        if ($v ne $provides{$n})
        {
          $found_missing = 1;
          $provides{$n} = $v;
        }
      }
      else
      {
        $found_missing = 1;
        $missing{$cp} = 1;
      }
    }
    my @new_provides;
    while (my ($n,$v) = each(%provides))
    {
      push(@new_provides,$n.$v);
    }
    push(@new_provides,keys(%missing));

    @new_provides = sort(@new_provides);

    print "$pkgname\n";
    if (not $distribution)
    {
      if (@cpan_provides)
      {
        print "  CPAN package: yes\n";
        print "  valid package: no\n";
        if (%correct_names)
        {
          my @names = sort keys %correct_names;
          print "  recommended replacement(s): @names\n";
        }
      }
      else
      {
        print "  CPAN package: no\n\n";
        next;
      }
    }
    else
    {
      print "  CPAN package: yes\n";
      my $correct_pkgname = &module_to_pkgname($distribution);
      if ($pkgname ne $correct_pkgname)
      {
        $correct_names{$pkgname} = 1;
        my @names = sort keys %correct_names;
        print "  valid package: no\n";
        print "  recommended replacement(s): @names\n";
      }
      else
      {
        print "  valid package: yes\n";
      }
    }
    if ($found_missing)
    {
      print "  current provides array (incomplete):\n";
      print "    $_\n" foreach @provides;
      print "  provides array recommendation:\n";
      print "    $_\n" foreach @new_provides;
      if ($fix)
      {
        my $dpath = $self->{Pacman}->get_local_db_pkg($pkgname)->{'dpath'};
        my $fname = &get_db_pkg_filename_by_section('PROVIDES');
        my $fpath = &join_paths($dpath,$fname);
        my $contents = &parse_db_file($fpath);
        my $before = &format_db_file($contents);
        $contents->{PROVIDES} = \@new_provides;
        my $after = &format_db_file($contents);
        print "\n$fpath:\n\n";
        print "########## BEFORE ##########\n";
        print $before;
        print "############################\n\n";
        print "########## AFTER ##########\n";
        print $after;
        print "###########################\n\n";
        if ($noconfirm or &confirm('Apply changes?','y'))
        {
          &save_db_file($fpath,$contents);
        }
      }
    }
    else
    {
      print "  current provides array (complete):\n";
      print "    $_\n" foreach @provides;
    }
    print "\n";
  }
}


sub check_pkg_validity
{
  my ($self,@pkgnames) = @_;
  my %unrecognized;
  my %mergible;
  foreach my $pkgname (@pkgnames)
  {
    my $distribution = $self->get_distribution(@pkgnames);
    if (not $distribution)
    {
      $unrecognized{$pkgname}++;
    }
    elsif ($pkgname ne &module_to_pkgname($distribution))
    {
      $mergible{$pkgname} = $distribution;
    }
  }

  print "unrecognized:\n";
  print "  $_\n" foreach sort keys %unrecognized;
  print "\nmergible:\n";
  print "  $_ -> ".&module_to_pkgname($mergible{$_})."\n" foreach sort keys %mergible;
}





########## Convenience Methods ##########

sub get_installed_pkg_provides_array
{
  my ($self,$pkgname) = @_;
  my $p = $self->{Pacman};
  my @fs = @{$p->get_local_db_pkg($pkgname)->{'get'}->('FILES')};
  my %provides;
  # Check all files which belong to the package:
  foreach my $f (@fs)
  {
    my $m;
    my $n;
    my $inc;
    # Skip anything which isn't a module.
    next if substr($f,-3) ne '.pm';
    $f = '/'.$f;
    # Remove the inclusion path to get the relative path,
    # which matches the module name.
    foreach my $i (@S_INC)
    {
      if ($f =~ m/^\Q$i\E\/(.+)\.pm$/)
      {
        $inc = $i;
        $m = $1;
        $n = $1.'.pm';
        $m =~ s/\//::/g;
        last;
      }
    }
    next if not $m;
    # If the module corresponds to a CPAN module, include it after checking
    # the local version.
    if (exists($self->{ModuleToDistribution}->{$m}))
    {
      my $v;
      my $lf;

      # If the module has already been loaded
      if (exists($INC{$n}))
      {
        # but the path is undefined in %INC, ignore the version
        if (not defined($INC{$n}))
        {
          $provides{$m} = undef;
          next;
        }
        # if the path matches, get the version
        elsif ($INC{$n} eq $f)
        {
          no warnings;
          $v = eval($m.'->VERSION');
          use warnings;
          $provides{$m} = $v;
          next;
        }
        else
        {
          # otherwise it's the wrong module (rare, but possible)
          $lf = $INC{$n};
          delete $INC{$n};
        }
      }
      # Set @INC to load the right module
      my @inc = @INC;
      @INC = ($inc);
      my $warnings_sub = $SIG{__WARN__};
      $SIG{__WARN__} = sub{};
      $v = eval("require $m;".$m.'->VERSION');
      @INC = @inc;
      $provides{$m} = $v;

      # reload the original module if it was loaded
      if ($lf)
      {
        delete $INC{$n};
        eval("require $m");
      }
      $SIG{__WARN__} = $warnings_sub;
    }
  }

  # If every module in a distribution is installed, consider the distribution
  # installed. This applies mostly to bundles, e.g. Bundle::LWP
  my %tracker;
  MODULE: foreach my $m (keys(%provides))
  {
    next if exists($tracker{$m});
    my $d = $self->{ModuleToDistribution}->{$m};
    next if exists($tracker{$d});
    foreach my $dm (keys(%{$self->{Distributions}->{$d}->{modules}}))
    {
      next MODULE if not exists($provides{$dm});
      $tracker{$dm} = 1;
    }
    $provides{$d} = undef;
    $tracker{$d} = 1;
  }
  delete($provides{$pkgname}) if exists($provides{$pkgname});
  my @provides = $self->href_to_pkgname_array(\%provides,'=');
  return @provides;
}







########## Formatting ##########

sub module_to_pkgname
{
  my ($name) = @_;
  return $CUSTOM_PKGNAMES{$name} if defined($CUSTOM_PKGNAMES{$name});
  $name =~ s/::/-/g;
  return 'perl-'.lc($name);
}

sub module_to_pathname
{
  my ($name) = @_;
  $name =~ s/::/-/g;
  return $name;
}

sub format_perl_ver
{
  my $ver = (@_) ? shift : $];
  return $ver if ($ver =~ m/\d+\.\d+\.\d+/);
  my ($maj,$min) = ($ver =~ m/(\d+)\.(\d+)/);
  my $a = int($min/1000);
  my $b = $min - $a * 1000;
  return "$maj.$a.$b";
}

sub format_cpan_ver
{
  my $ver = shift;
  return $ver if not defined($ver);
  if ($ver =~ m/^[vV]?(\d+\.)(\d+)$/)
  {
    my ($a,$b) = ($1,$2);
    my $l = length($b);
    $b .= ('0' x (6-$l)) if $l < 6;
    $ver = $a.int(substr($b,0,3)).'.'.int(substr($b,3));
  }
  elsif ($ver =~ m/^[vV]?(\d+)((?:\.\d+)*)(_\d+)$/)
  {
    my ($a,$b,$c) = ($1,$2,$3);
    my @bs = map {int($_)} grep {$_} split(/\./,$b);
    $c = 'a.' . int(substr($c,1));
    $ver = $a;
    $ver .= '.'.$_ foreach @bs;
    $ver .= $c;
  }
  return $ver
}

sub path_to_distribution_and_version
{
  my ($path) = @_;
  my $bname = basename($path);
  my $blength = length($bname);
  foreach my $ext (@ARCHIVE_EXTS)
  {
    my $l = length($ext);
    if ($blength >= $l and substr($bname,-$l) eq $ext)
    {
      $bname = substr($bname,0,-$l);
      if ($bname =~ m/^(.+?)[\.\-_v](\d.*)$/ or $bname =~ m/^(.+?)-V(\d.*)$/)
      {
        my ($name,$ver) = ($1,$2);
        $ver = ($name eq 'perl') ? &format_perl_ver($ver) : &format_cpan_ver($ver);
        return ($name,$ver);
      }
      last;
    }
  }
  return ($bname,'undef');
}



########## MESSAGES ##########

sub maybe_display_warning
{
  my $self = shift;
  return if $self->{CALLBACK}->{QUIET} or not defined($self->{CALLBACK}->{WARNING});
  foreach my $arg (@_)
  {
    $self->{CALLBACK}->{WARNING}->($arg);
  }
}

sub maybe_display_error
{
  my $self = shift;
  return if not defined($self->{CALLBACK}->{ERROR});
  foreach my $arg (@_)
  {
    $self->{CALLBACK}->{ERROR}->($arg);
  }
}

1;

__END__

=head1 NAME

Xyne::Arch::CPAN - Subroutines for interfacing with CPAN.

=head1 SYNOPSIS

  use Xyne::Arch::CPAN;

=head1 DESCRIPTION

This module provides several subroutines for interfacing with the CPAN.

=head2 EXPORT

None by default.

=head1 AUTHOR

Xyne <lt>xyne at archlinux dot ca<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2010  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
