# Copyright (C) 2010  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


package Xyne::Arch::Hackage;
use strict;
use warnings;

require Exporter;
our @ISA = qw/Exporter/;
our @EXPORT_OK = qw/module_to_pkgname path_to_distribution_and_version/;

use Xyne::Arch::Pacman qw/compare_versions/;
use Xyne::Common::Functions qw/join_paths/;
use Xyne::Common::Threads qw/parallel_mirror_urls/;

use File::Basename qw/basename/;

our $BASE_URL = 'http://hackage.haskell.org/packages/';
our $PKG_LIST_URL = $BASE_URL . 'archive/pkg-list.html';
our $LOG_URL = $BASE_URL . 'archive/log';
our $HACKAGE_DB = 'hackage_db';




our $VERSION = '0.01';






sub new {
  my ($proto,$args) = @_;
  my $class = ref($proto) || $proto;
  my $self  = {};
  bless ($self, $class);

  if (defined($args->{Pacman}))
  {
    $self->{Pacman} = $args->{Pacman};
  }
  else
  {
    $self->{Pacman} = Xyne::Arch::Pacman->new({
      autoload => $args->{'pacman_autoload'},
      config_path => $args->{'pacman_config_path'}
    });
  }


  if (defined($args->{Cache}))
  {
    $self->{CACHE} = $args->{Cache};
  }
  else
  {
    $self->{CACHE} = '/tmp/bauerbill/hackage';
  }

  return $self;
}








########## OPERATIONS ##########

sub get_upgradable
{
  my ($self,$ignore,@pkgnames) = @_;
  my $pacman = $self->{Pacman};
  if (not @pkgnames)
  {
    foreach my $pkgname (keys %{ $self->{PKGNAMES} })
    {
      push (@pkgnames,$pkgname) if $pacman->is_in_local_db($pkgname);
    }
  }
  @pkgnames = grep {not $pacman->should_ignore($_)} @pkgnames if $ignore;
  my @upgradable = ();
  foreach my $pkgname (@pkgnames)
  {
    my $pkg = $self->{PKGNAMES}->{$pkgname};
    next if not defined($pkg);
    my $hackage_ver = $self->{DB}->{$pkg}->{ver};
    my $local_ver = $pacman->get_local_db_pkg_ver($pkgname);
    my $cmpver = &compare_versions($hackage_ver,$local_ver);
    push(@upgradable,$pkgname) if $cmpver == 1;
  }
  return @upgradable;
}









########## Database Functions and Methods ##########

sub get_versions_from_log
{
  my ($fpath) = @_;
  if(open(my $fh,'<',$fpath))
  {
    my $versions = {};
    while(defined(my $line = <$fh>))
    {
      $versions->{$1} = $2 if $line =~ m/(\S+) (\S+)$/;
    }
    close($fh);
    return $versions;
  }
  else
  {
    print STDERR "error: failed to open $fpath ($!)";
    exit(1);
  }
}



sub get_pkg_list
{
  my ($fpath,$versions) = @_;
  if(open(my $fh,'<',$fpath))
  {
    my $pkgs = {};
    my $cat;
    my $pkg;
    my $type;
    my $desc;
    while(defined(my $line = <$fh>))
    {
      if ($line =~ m/name="cat:([^"]+)"/)
      {
        $cat = $1;
      }
      elsif ($line =~ m/href="\/package\/([^"]+)"/)
      {
        $pkg = $1;
      }
      elsif ($line =~ m/> (.+?): (.+)<\/li/)
      {
        ($type,$desc) = ($1,$2);
        $pkgs->{$pkg} = {
          cat => $cat,
          type => $type,
          desc => $desc,
          ver => $versions->{$pkg}
        };
      }
    }
    close($fh);
    return $pkgs;
  }
  else
  {
    print STDERR "error: failed to open $fpath ($!)";
    exit(1);
  }
}



sub save_db
{
  my ($fpath,$pkgs) = @_;
  if (open(my $fh,'>',$fpath))
  {
    foreach my $pkg (sort keys(%$pkgs))
    {
      print $fh join(' ', $pkg, $pkgs->{$pkg}->{ver}, $pkgs->{$pkg}->{cat}, $pkgs->{$pkg}->{type}.':', $pkgs->{$pkg}->{desc})."\n";
    }
    close($fh);
  }
  else
  {
    print STDERR "error: failed to open $fpath ($!)";
    exit(1);
  }
}



sub maybe_update_database
{
  my ($self,$force) = @_;
  my @urls = ($PKG_LIST_URL,$LOG_URL);

  if ($force)
  {
    foreach my $url (@urls)
    {
      my $fpath = &join_paths($self->{CACHE},basename($url));
      unlink $fpath if -e $fpath;
    }
  }
  my @responses = &parallel_mirror_urls($self->{CACHE},@urls);

  my $is_new = 0;
  foreach my $response (@responses)
  {
    if ($response->[0] == 200)
    {
      $is_new = 1;
    }
    elsif ($response->[0] != 304)
    {
      print STDERR "error: failed to mirror $response->[1] (response code: $response->[0])\n";
      exit(1);
    }
  }
  my $fpath = &join_paths($self->{CACHE},$HACKAGE_DB);
  $is_new = 1 if not -f $fpath;
  if ($is_new)
  {
    $self->format_and_save_database();
  }
  return $is_new;
}



sub maybe_update_and_load_database
{
  my ($self,$force) = @_;
  my $is_new = $self->maybe_update_database($force);
  $self->load_database if not $is_new;
  return $is_new;
}



sub format_and_save_database
{
  my ($self) = @_;
  my $log_fpath = &join_paths($self->{CACHE}, basename($LOG_URL));
  my $pkg_list_fpath = &join_paths($self->{CACHE}, basename($PKG_LIST_URL));
  my $db_fpath = &join_paths($self->{CACHE}, $HACKAGE_DB);

  my $versions = &get_versions_from_log($log_fpath);
  my $pkgs = &get_pkg_list($pkg_list_fpath,$versions);
  &save_db($db_fpath,$pkgs);
  $self->{DB} = $pkgs;
  $self->load_pkgname_list();
}



sub load_database
{
  my ($self) = @_;
  my $fpath = &join_paths($self->{CACHE}, $HACKAGE_DB);
  if (not -f $fpath)
  {
    $self->maybe_update_and_load_database;
  }
  elsif(open(my $fh,'<',$fpath))
  {
    my $pkgs = {};
    while(defined(my $line = <$fh>))
    {
      my ($pkg,$ver,$cat,$type,$desc) = ($line =~ m/^(\S+) (\S+) (\S+) (.+?): (.+)$/);
      $pkgs->{$pkg} = {
        cat => $cat,
        type => $type,
        desc => $desc,
        ver => $ver
      };
    }
    close($fh);
    $self->{DB} = $pkgs;
    $self->load_pkgname_list();
  }
  else
  {
    print STDERR "error: failed to open $fpath ($!)";
    exit(1);
  }

}


sub maybe_load_database
{
  my ($self) = @_;
  $self->load_database if not defined $self->{DB};
}










########## PKGNAMES ##########

sub load_pkgname_list
{
  my ($self) = @_;
  foreach my $pkg (keys(%{$self->{DB}}))
  {
    $self->{PKGNAMES}->{$self->pkg_to_pkgname($pkg)} = $pkg;
  }
}

sub pkg_to_pkgname
{
  my ($self,$pkg) = @_;
  my $obj = $self->{DB}->{$pkg};
  return undef if not $obj;
  my $type = $obj->{type};
  my $prefix = (lc(substr($type,0,7)) eq 'library') ? 'haskell-' : '';
  return $prefix.lc($pkg);
}

sub pkgname_to_pkg
{
  my ($self,$pkgname) = @_;
  return $self->{PKGNAMES}->{$pkgname};
}

sub get_pkg
{
  my ($self,$pkg) = @_;
  return $self->{DB}->{$pkg} if defined($self->{DB}->{$pkg});
  $pkg = $self->pkgname_to_pkg($pkg);
  return $self->{DB}->{$pkg} if $pkg and defined($self->{DB}->{$pkg});
}

sub get_pkg_name
{
  my ($self,$pkg) = @_;
  return $pkg if defined($self->{DB}->{$pkg});
  return $self->pkgname_to_pkg($pkg);
}

sub get_pkg_url
{
  my ($self,$pkg) = @_;
  $pkg = $self->pkgname_to_pkg($pkg) if not defined($self->{DB}->{$pkg});
  return '' if not defined($pkg);
  return $BASE_URL.$pkg;
}

sub get_cabal_url
{
  my ($self,$pkg) = @_;
  $pkg = $self->pkgname_to_pkg($pkg) if not defined($self->{DB}->{$pkg});
  return '' if not defined($self->{DB}->{$pkg});
  my $ver = $self->{DB}->{$pkg}->{ver};
  return $BASE_URL.'archive/'.$pkg.'/'.$ver.'/'.$pkg.'.cabal';
}




















########## MESSAGES ##########

sub maybe_display_warning
{
  my $self = shift;
  return if $self->{CALLBACK}->{QUIET} or not defined($self->{CALLBACK}->{WARNING});
  foreach my $arg (@_)
  {
    $self->{CALLBACK}->{WARNING}->($arg);
  }
}

sub maybe_display_error
{
  my $self = shift;
  return if not defined($self->{CALLBACK}->{ERROR});
  foreach my $arg (@_)
  {
    $self->{CALLBACK}->{ERROR}->($arg);
  }
}

1;

__END__

=head1 NAME

Xyne::Arch::Hackage - Subroutines for interfacing with Hackage.

=head1 SYNOPSIS

  use Xyne::Arch::Hackage;

=head1 DESCRIPTION

This module provides several subroutines for retrieving data from Hackage.

=head2 EXPORT

None by default.

=head1 AUTHOR

Xyne <lt>xyne at archlinux dot ca<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2010  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
