# Copyright (C) 2010  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


package Xyne::Arch::PKGBUILD;
use strict;
use warnings;

use File::Basename qw/dirname basename/;
use IPC::Open2;
use POSIX qw/uname/;

use Xyne::Common::Functions qw/str strip parse_list_of_strings/;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = do {
  no strict 'refs';
  grep exists &$_ && /[a-z]/, keys %{ __PACKAGE__ . '::'};
};

our $VERSION = '0.03';

our $VAR_REGEX = '([A-Za-z0-9_]+)';
our $FUNC_REGEX = '(?:function )?([A-Za-z0-9_-]+) \\(\\)';








sub parse_pkgbuild
{
  my ($fpath) = @_;
  
  # may as well be thorough here
  if (not -e $fpath)
  {
    print STDERR "error: $fpath does not exist\n";
    exit(1);
  }
  elsif (not -f $fpath)
  {
    print STDERR "error: $fpath is not a file\n";
    exit(1);
  }
  elsif (not -r $fpath)
  {
    print STDERR "error: $fpath is not readable\n";
    exit(1);
  }
  
  # Check the syntax to make sure the PKGBUILD is correctly formatted. This
  # should prevent PKGBUILDs from escaping the wrapping function below.
  
  my $e = system("/bin/bash -n $fpath 2>/dev/null\n");
  {
    if ($e)
    {
      print STDERR "error: malformed Bash syntax ($fpath)\n";
      exit(1);
    }
  }
  
  # A hash ref to store all variables.
  my $vars = {
    'CARCH' => [uname()]->[4]
  };
  # A hash ref that points to hash refs nested in $vars, for scoping.
  my $_vars = $vars;
  # A flat hash ref which tracks all variables within the local scope.
  my $_local_vars = {};
  
  # Open pipes to bash... I don't know if "--restricted" does anything useful
  # here, but I doubt it hurts.
  my ($pin,$pout,$pid);
  if ($pid=open2($pout,$pin,'/bin/bash --restricted'))
  {
    local $| = 1;
    
    open (my $fh,'<',$fpath) or die "$!\n";
    
    # Wrap the entire PKGBUILD in a function to prevent actual code execution.
    print $pin "pkgbuild() {\n";
    while (defined(my $line = <$fh>))
    {
      print $pin $line;
    }
    close $fh;
    print $pin "\n}\nset\n";
    close $pin;
    
    my @functions = ();
    my $maybe_function = 0;
    while(defined(my $line = <$pout>))
    {
      chomp $line;
      if ($line =~ m/^$VAR_REGEX=(.*[^;])?$/)
      {
        my $var = $1;
        my $val = $2;
        next if not $val;
        my $is_array = 0;
        if (substr($val,0,1) eq '(')
        {
          $val =~ s/\[\d+\]=//g;
          $is_array = 1;
        }
        my @vals = &parse_list_of_strings($val);
        $vars->{$var} = ($is_array) ? [@vals] : $vals[0];
      }
      elsif ($line =~ m/(?:^|.*\s)$VAR_REGEX=(.*?);?$/)
      {
        my $var = $1;
        my $val = $2;
        
        next if not defined($val);
        $val = &safe_backticks($val);
        
        my $is_array = 0;
        if (substr($val,0,1) eq '(')
        {
          $is_array = 1;
          $val = substr($val,1,rindex($val,')')-1);
          $val = &parse($val,$_local_vars);
        }
        # This also handles escapes which is why it is used for non-arrays too.
        my @vals = &parse_list_of_strings($val);
        my @parsed_vals = ();
  
        foreach my $v (@vals)
        {
          push (@parsed_vals,&parse($v,$_local_vars));
        }
        
        if ($line =~ m/(?:^|.*\s)eval\s+\Q$var\E=(.*?);?$/)
        {
          @parsed_vals = safe_eval(@parsed_vals);
        }
        
        $_vars->{$var} = ($is_array) ? [@parsed_vals] : $parsed_vals[0];
        $_local_vars->{$var} = $_vars->{$var};
      }
      elsif ($line =~ m/^\s*$FUNC_REGEX/)
      {
        push(@functions,$1);
        $maybe_function = 1;
        next;
      }
      elsif ($line =~ m/^\s*\{\s*$/)
      {
        my $f = $functions[-1];
        $_vars->{$f} = {} if not defined($_vars->{$f});
        $_vars = $_vars->{$f};
        foreach (keys(%{$vars}))
        {
          $_local_vars->{$_} = $vars->{$_} if not defined($_local_vars->{$_});
        }
        $_local_vars->{$_} = $_vars->{$_} foreach keys(%{$_vars});
      }
      elsif ($line =~ m/^\s*\};?\s*$/)
      {
        pop(@functions);
        $_vars = $vars;
        $_local_vars = $vars;
        $_local_vars->{$_} = $_vars->{$_} foreach keys(%{$_vars});
        foreach my $f (@functions)
        {
          $_vars->{$f} = {} if not defined($_vars->{$f});
          $_vars = $_vars->{$f};
          $_local_vars->{$_} = $_vars->{$_} foreach keys(%{$_vars});
        }
        
      }
      elsif ($maybe_function)
      {
        pop(@functions);
      }
      $maybe_function = 0;
    }
    close $pout;
    waitpid $pid,0;
  }
  else
  {
    print STDERR "error: failed to open pipes to /bin/bash ($!)\n";
    exit(1);
  }
    
  my $pkgbuild = $vars->{pkgbuild};
  my @pkgnames;
  if (ref($pkgbuild->{pkgname}) eq 'ARRAY')
  {
    @pkgnames = @{ $pkgbuild->{pkgname} };
  }
  else
  {
    @pkgnames = ($pkgbuild->{pkgname});
  }
  
  my $common = {};
  while(my($k,$v) = each(%$pkgbuild))
  {
    # save scalars and arrays
    $common->{$k} = $v if ref($v) ne 'HASH';
  }
  
  my @pkgbuild_objs = ();
  foreach my $pkgname (@pkgnames)
  {
    my $obj = {
      %$common,
      'pkgname' => $pkgname,
    };
    if (ref($pkgbuild->{'package_'.$pkgname}) eq 'HASH')
    {
      while(my($k,$v) = each(%{$pkgbuild->{'package_'.$pkgname}}))
      {
        $obj->{$k} = $v if ref($v) ne 'HASH';
      }
    }
    push (@pkgbuild_objs,$obj);
  }
  return @pkgbuild_objs;
}

sub parse
{
  my ($v,$_local_vars) = @_;
  # variable interpolation
  $v =~ s/(\$\{([^}]+)\})/&interpolate($_local_vars,$1,$2)/eg;
  $v =~ s/(\$$VAR_REGEX)/&interpolate($_local_vars,$1,$2)/eg;
  
  # substrings
  $v =~ s/(\$\{$VAR_REGEX:(\d+)(?::(\d+))?\})/&get_substring($_local_vars,$1,$2,$3)/eg;
  
  # replacements
  $v =~ s/(\$\{$VAR_REGEX (\/[\/#%]?) ((?:\\.|[^\\\/])+)  (?:\/ ((?:\\.|[^\\\/])*))?  \}  )/&replace($_local_vars,$1,$2,$3,$4,$5)/xeg;
  return $v;
}

sub interpolate
{
  my ($local_vars,$str,$var) = @_;
  my $index;
  if ($var =~ m/^(.+?)\[([^\]]+)\]$/)
  {
    $var = $1;
    $index = $2;
  }
  return $str if not defined($local_vars->{$var});
  my $ivar = $local_vars->{$var};
  if (ref($ivar) eq 'ARRAY')
  {
    if (@$ivar)
    {
      return $ivar->[0] if not defined($index);
      return join(' ',@$ivar) if $index eq '@';
      return @$ivar[$index];
    }
    return '';
  }
  return $ivar;
}

sub get_substring
{
  my ($local_vars,$str,$var,$start,$length) = @_;
  return $str if not defined($local_vars->{$var}) or not defined($start);
  if (ref($local_vars->{$var}) eq 'ARRAY')
  {
    if (@{$local_vars->{$var}})
    {
      $str = $local_vars->{$var}->[0];
    }
    else
    {
      return $str;
    }
  }
  else
  {
    $str = ($local_vars->{$var});
  }
  if (defined($length))
  {
    return substr($str,$start,$length);
  }
  else
  {
    return substr($str,$start);
  }
}

sub replace
{
  my ($local_vars,$str,$var,$opener,$a,$b) = @_;
  return $str if not defined($local_vars->{$var});
  if (ref($local_vars->{$var}) eq 'ARRAY')
  {
    if (@{$local_vars->{$var}})
    {
      $str = $local_vars->{$var}->[0];
    }
    else
    {
      return $str;
    }
  }
  else
  {
    $str = ($local_vars->{$var});
  }
  $b = '' if not defined($b);
  $a =~ s/\\(.)/$1/g;
  $b =~ s/\\(.)/$1/g;
  if ($opener eq '/')
  {
    $str =~ s/\Q$a\E/$b/;
  }
  elsif ($opener eq '//')
  {
    $str =~ s/\Q$a\E/$b/g;
  }
  elsif ($opener eq '/#')
  {
    $str =~ s/^\Q$a\E/$b/g;
  }
  elsif ($opener eq '/%')
  {
    $str =~ s/\Q$a\E$/$b/g;
  }
  return $str;
}

sub safe_eval
{
  my (@values) = @_;
  my @parsed_values = ();
  foreach my $v (@values)
  {
    # handles expansion of foo{bar,baz}whatever
    if ($v =~ m/^(.*[^\$])?\{([^\{\}]+)\}(.*)$/)
    {
      my $pre = $1;
      my $post = $3;
      my @args = map {$pre.$_.$post} split(',',$2);
      push(@parsed_values,&safe_eval(@args));
    }
    else
    {
      push(@parsed_values,$v);
    }
  }
  return @parsed_values;
}

sub safe_backticks
{
  my ($val) = @_;
  while ($val =~ m/`([^`]*)`/g)
  {
    my $cmd = $1;
    my $output = '';
    if ($cmd =~ m/^\s*uname(?:\s+[^&|]+)?$/)
    {
      $output = `$cmd`;
    }
    $val =~ s/`\Q$cmd\E`/$output/g;
  }
  return $val;
}


1;

__END__

=head1 NAME

Xyne::Arch::PKGBUILD; - Arch Linux PKGBUILD Parser

=head1 SYNOPSIS

  use Xyne::Arch::PKGBUILD;

=head1 DESCRIPTION

This module contains functions to parse PKGBUILDs and extract useful information.

=head2 EXPORT

None by default.

=head1 AUTHOR

Xyne <lt>xyne at archlinux dot ca<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009,2010  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.