# Copyright (C) 2009,2010  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


package Xyne::Arch::Pacman;
use strict;
use warnings;

use Xyne::Arch::PacmanConf;
use Xyne::Common::Functions qw/confirm get_term_width join_paths slurp_file refpush str wrap_words/;
use Xyne::Common::RemoteFile;

require Exporter;
our @ISA = qw/Exporter/;
our @EXPORT_OK = qw/check_version_inequality check_cmpver_inequality compare_versions get_name_and_version get_name_and_ver_spec parse_db_file get_name_and_repo save_db_file format_db_file get_db_pkg_filename_by_section check_name_and_ver_spec/;

our $VERSION = '0.47';

our @ARCHITECTURES = qw/i686 x86_64/;
our $ARCHIVE_EXT = '.tar.gz';

our $LOCK_FILE = 'db.lck';


########## CONSTRUCTOR ##########

sub new
{
  my ($proto,$args) = @_;
  my $class = ref($proto) || $proto;
  my $self  = {};
  bless ($self, $class);

  $self->reset_db;


  if (defined($args->{PacmanConf}))
  {
    $self->{PacmanConf} = $args->{PacmanConf};
  }
  else
  {
    $self->{PacmanConf} = Xyne::Arch::PacmanConf->new();
  }
  $self->load_config($args->{'config_path'}) if defined($args->{'autoload'});

  return $self;
}


sub reset_db
{
  my ($self) = @_;
  $self->{DB}={};
}



########## DEBUGGING ##########
sub get_report
{
  my ($self,$report) = @_;
  $report = &Xyne::Common::Functions::get_report($report, 'Pacman', ['Database', $self->{DB}]);
  $report = $self->{PacmanConf}->get_report($report);
  return $report;
}

sub abort
{
  my ($self,$sig) = @_;
  print "\naborting...\n";
  if ($self->{DB_LOCKED})
  {
    print "removing database lock\n";
    $self->unlock_db;
  }
  exit(1);
}










########## CONVENIENCE METHODS ##########

sub get_repo_list
{
  my ($self) = @_;
  return $self->{PacmanConf}->get_repo_list()
}

sub load_config
{
  my ($self) = @_;
  return $self->{PacmanConf}->load_config();
}







########## URLS ##########

sub get_repo_server_urls
{
  my ($self,$reponame,$arch) = @_;
  return () if not defined($reponame);
  my @urls = $self->{PacmanConf}->get_repo_servers($reponame);
  @urls = &replace_url_architecture($arch,@urls) if defined($arch);
  return &replace_url_repo($reponame,@urls);
}

sub get_repo_file_urls
{
  my ($self,$reponame,$filename,$arch) = @_;
  return () if not defined($filename);
  return map {$_.'/'.$filename} $self->get_repo_server_urls($reponame,$arch);
}













########## ARCHIVES ##########

sub get_repo_archive
{
  my ($self,$reponame,$archive,$arch) = @_;
  return $self->{ARCHIVES}->{$reponame}->{$archive} if defined($self->{ARCHIVES}->{$reponame}->{$archive});
  return undef if not defined($reponame) or not defined($archive);
  my $filename = $reponame.'.'.$archive.$ARCHIVE_EXT;
  my $a = Xyne::Common::RemoteFile->new({
    'filename' => $filename,
    'urls' => [$self->get_repo_file_urls($reponame,$filename,$arch)]
  });
  $self->{ARCHIVES} = {} if not defined($self->{ARCHIVES});
  $self->{ARCHIVES}->{$reponame} = {} if not defined($self->{ARCHIVES}->{$reponame});
  $self->{ARCHIVES}->{$reponame}->{$archive} = $a;
  return $a;
}

sub get_repo_db_remote_mtime
{
  my ($self,$reponame,$arch) = @_;
  my $a = $self->get_repo_archive($reponame,'db',$arch);
  return $a->get_mtime;
}






########## OPERATIONS ##########

# returns true if the package is not locally installed as well
sub is_upgradable
{
  my ($self,$pkgname,$ignore) = @_;
  return 1 if not $self->is_installed($pkgname);
  return 0 if $ignore and $self->should_ignore($pkgname);
  my $local_ver = $self->get_local_db_pkg_ver($pkgname);
  my $sync_ver = $self->get_sync_db_pkg_ver($pkgname);
  my $cmpver = &compare_versions($sync_ver,$local_ver);
  if ($cmpver > 0 or ($cmpver != 0 and $self->is_forced($pkgname)))
  {
    return 1;
  }
  else
  {
    return 0;
  }
}

sub get_upgradable_pkgnames
{
  my ($self,$ignore) = @_;
  my @upgradable = ();
  foreach my $pkgname ($self->get_local_db_pkgname_list)
  {
    if ($self->is_upgradable($pkgname,$ignore))
    {
      push(@upgradable,$pkgname);
    }
#     else
#     {
#       my $replacement = $self->get_sync_db_replacement($pkgname);
#       push(@upgradable,$replacement) if defined($replacement);
#     }
  }
  return @upgradable;
}

sub is_ignored
{
  my ($self,$pkgname) = @_;
  return 0 if not defined($pkgname);
  my @ignored = $self->{PacmanConf}->get_value('IgnorePkg');
  return 0 if not @ignored;
  return 1 if grep {$pkgname eq $_} @ignored;
  return 0;
}

sub is_ignored_group
{
  my ($self,$groupname) = @_;
  return 0 if not defined($groupname);
  my @groups = $self->{PacmanConf}->get_value('IgnoreGroup');
  foreach (@groups)
  {
    return 1 if $_ eq $groupname;
  }
  return 0;
}

sub should_ignore
{
  my ($self,$pkgname) = @_;
  return 1 if $self->is_ignored($pkgname);
  my $pkg = $self->get_local_db_pkg($pkgname);
  return 0 if not defined($pkg);

  my @groups = $self->{PacmanConf}->get_value('IgnoreGroup');

  if (@groups)
  {
    foreach my $group (&get_pkg_groups($pkg))
    {
      return 1 if grep {$group eq $_} @groups;
    }
  }
  return 0;
}






sub resolve_sync_queue_deps
{
  my ($self,$sync_queue) = @_;
  my @pkgnames = keys(%$sync_queue);
  $self->build_deptree($_) foreach @pkgnames;
  my $checked = {};
  foreach my $pkgname (@pkgnames)
  {
    $checked = $self->check_deptree($pkgname,$checked);
    if ($checked->{$pkgname} ne '')
    {
      $sync_queue->{$pkgname}->{unresolvable} = $checked->{$pkgname};
    }
    else
    {
      $sync_queue->{$_}->{dep} = 1 foreach $self->get_deps_from_deptree($pkgname);
    }
  }
  return $sync_queue;
}

sub get_deps_from_deptree
{
  my ($self,$pkgname) = @_;
  my %deps = ();
  my @queue = ($pkgname);
  while (@queue)
  {
    my $n = shift @queue;
    next if defined($deps{$n});
    push (@queue,keys(%{$self->{DB}->{DEPTREE}->{$n}}));
    $deps{$n} = 1;
  }
  delete($deps{$pkgname});
  return keys(%deps);
}

sub check_deptree
{
  # checked is to prevent infinite recursion for cyclical dependencies
  my ($self,$pkgname,$checked) = @_;
  $checked = {} if not defined($checked);
  return $checked if defined($checked->{$pkgname});
  $checked->{$pkgname} = '';

  my $node = $self->{DB}->{DEPTREE}->{$pkgname};
  foreach my $dep (keys(%$node))
  {
    if (ref($node->{$dep}) eq 'HASH')
    {
      $checked = $self->check_deptree($dep,$checked);
      if ($checked->{$dep} ne '')
      {
        $checked->{$pkgname} = "$pkgname -> " .$checked->{$dep};
        return $checked;
      }
    }
    else
    {
      my $reason = $node->{$dep};
      $checked->{$pkgname} = "$pkgname -> $dep ($reason)";
      return $checked;
    }
  }
  return $checked;
}

sub build_deptree
{
  my ($self,$pkgname) = @_;
  return $self->{DB}->{DEPTREE}->{$pkgname} if defined($self->{DB}->{DEPTREE}->{$pkgname});
  $self->{DB}->{DEPTREE}->{$pkgname} = {};
  foreach my $dep ($self->get_sync_db_pkg_deps($pkgname))
  {
    my $d = $self->resolve_dep($dep);
    if ($d eq '')
    {
      next
    }
    elsif ($d)
    {
      if ($self->should_ignore($d))
      {
        $self->{DB}->{DEPTREE}->{$pkgname}->{$dep} = 'available but ignored';
        last;
      }
      else
      {
        $self->{DB}->{DEPTREE}->{$pkgname}->{$d} = $self->build_deptree($d);
      }
    }
    else
    {
      $self->{DB}->{DEPTREE}->{$pkgname}->{$dep} = 'not found';
      last;
    }
  }
  return $self->{DB}->{DEPTREE}->{$pkgname};
}

# Returns an empty string if satisfied by an installed package,
# else the name of a sufficient sync db package, otherwise undef.
sub resolve_dep
{
  my ($self,$dep) = @_;
  return $self->{DB}->{DEPS}->{$dep} if exists($self->{DB}->{DEPS}->{$dep});
  my ($n,$v,$i) = &get_name_and_ver_spec($dep);
  if ($self->is_effectively_installed($n,$v,$i))
  {
    $self->{DB}->{DEPS}->{$dep} = '';
    return '';
  }
  my $p = $self->is_effectively_in_sync_db($n,$v,$i);
  if (defined($p))
  {
    $self->{DB}->{DEPS}->{$dep} = $p;
    return $p;
  }
  $self->{DB}->{DEPS}->{$dep} = undef;
  return undef;
}







sub get_sync_queue
{
  my ($self,$group_confirmation_sub) = @_;
  return {} if not $self->{PacmanConf}->get_value('-S');
  my @args = $self->{PacmanConf}->get_cli_args;
  my $sync_queue = {};
  foreach my $arg (@args)
  {
    my $pkgname = $self->is_effectively_in_sync_db($arg);
    if ($pkgname)
    {
      $sync_queue->{$pkgname}->{target} = 1;
    }
    else
    {
      my @pkgnames = $self->get_sync_db_group($arg);
      if (@pkgnames)
      {
        if (ref($group_confirmation_sub) eq 'CODE')
        {
          @pkgnames = $group_confirmation_sub->($arg,@pkgnames);
        }
        $sync_queue->{$_}->{target} = 1 foreach @pkgnames;
      }
      else
      {
        print STDERR "error: $arg does not match any package or group\n";
        exit(1);
      }
    }
  }
  if ($self->{PacmanConf}->get_value('--needed'))
  {
    foreach my $pkgname (keys(%{$sync_queue}))
    {
      delete($sync_queue->{$pkgname}) if not $self->is_upgradable($pkgname);
    }
  }
  if ($self->{PacmanConf}->get_value('-u'))
  {
    $sync_queue->{$_}->{target} = 1 foreach $self->get_upgradable_pkgnames(1);
  }
  if (not $self->{PacmanConf}->get_value('-d'))
  {
    $sync_queue = $self->resolve_sync_queue_deps($sync_queue);
  }
  return $sync_queue;
}



sub maybe_get_sync_queue_interactively
{
  my ($self) = @_;
  if ($self->{PacmanConf}->get_value('--noconfirm'))
  {
    return $self->get_sync_queue();
  }
  else
  {
    return $self->get_sync_queue_interactively();
  }
}



sub get_sync_queue_interactively
{
  my ($self) = @_;
  return $self->get_sync_queue(sub{
    $self->select_group_packages(@_)
  });
}



sub select_group_packages
{
  my ($self,$group,@pkgnames) = @_;
  my @selected = ();
  my $w = &get_term_width();
  print ":: group $group (including ignored packages):\n";
  print &wrap_words($w,'    ',@pkgnames);
  if (&confirm(':: Install whole content?','y'))
  {
    return @pkgnames;
  }
  else
  {
    foreach my $pkgname (@pkgnames)
    {
      push(@selected,$pkgname) if &confirm(":: Install $pkgname from group $group?",'y');
    }
  }
  return @selected;
}








sub operation_requires_root
{
  my ($self) = @_;
  if ($self->{PacmanConf}->get_value('-S'))
  {
    return not (
      $self->{PacmanConf}->get_value('-p') or
      $self->{PacmanConf}->get_value('-i') or
      $self->{PacmanConf}->get_value('-s') or
      $self->{PacmanConf}->get_value('-l')
    );
  }
  elsif ($self->{PacmanConf}->get_value('-R')
  or     $self->{PacmanConf}->get_value('-U'))
  {
    return 1;
  }
  else
  {
    return 0;
  }
}


sub operation_involves_possible_download
{
  my ($self) = @_;
  if ($self->{PacmanConf}->get_value('-S'))
  {
    return 1 if $self->{PacmanConf}->get_value('-u');
    return 0 if $self->{PacmanConf}->get_value('-i');
    return 0 if $self->{PacmanConf}->get_value('-p');
    return 0 if $self->{PacmanConf}->get_value('-s');
    return 0 if $self->{PacmanConf}->get_value('-l');
    return 0 if not $self->{PacmanConf}->get_cli_args();
    return 1;
  }
  return 0;
}








########## DATABASE ##########

sub get_db_path
{
  my ($self,@path) = @_;
  return join_paths($self->{PacmanConf}->get_dbpath(),@path);
}

sub get_db_lock_path
{
  my ($self) = @_;
  return $self->get_db_path($LOCK_FILE);
}

sub lock_db
{
  my ($self) = @_;
  my $fpath = $self->get_db_lock_path;
  if (-e $fpath)
  {
    print STDERR "error: unable to lock database... an existing lock file was found ($fpath)\n";
    exit(1);
  }
  my $fh;
  if (open($fh,'>',$fpath))
  {
    close($fh);
  }
  else
  {
    print STDERR "error: failed to lock database ($!)\n";
    exit(1);
  }
  $self->{DB_LOCKED} = 1;
}

sub unlock_db
{
  my ($self) = @_;
  my $fpath = $self->get_db_lock_path;
  if (not -e $fpath)
  {
    print STDERR "error: unable to unlock database... the lock file could not be found)\n";
    exit(1);
  }
  unlink($fpath);
  if (-e $fpath)
  {
    print STDERR "error: failed to unlock database ($!)\n";
    exit(1);
  }
  delete $self->{DB_LOCKED};
}

sub read_db_dir
{
  my ($self,@path) = @_;
  my $dbpath = $self->get_db_path(@path);
  return () if not -r $dbpath;
  if (opendir(my $dh,$dbpath))
  {
    my @items = grep {substr($_,0,1) ne '.'} readdir($dh);
    closedir($dh);
    return @items;
  }
  else
  {
    return ();
  }
}

sub pkg_info_search
{
  my ($self,$pkg,$section,$name,$ver,$ineq) = @_;
  if (defined($ver))
  {
    my $section_data = $pkg->{'get'}->($section);
    return 0 if not defined($section_data);
    foreach my $whatever (@$section_data)
    {
      return 1 if $name eq $whatever;
      return 1 if &check_name_and_ver_spec($name,$ver,$ineq,$whatever);
    }
  }
  else
  {
    return $pkg->{'check'}->($section,$name);
  }
  return 0;
}

sub pkgs_info_search
{
  my ($self,$pkgs,$single,@args) = @_;
  my @pkgs = @$pkgs;
  if ($single)
  {
    foreach my $pkg (@pkgs)
    {
      return $pkg if $self->pkg_info_search($pkg,@args);
    }
  }
  else
  {
    return grep {$self->pkg_info_search($_,@args)} @pkgs;
  }
  return undef;
}

sub is_in_db
{
  my ($self,$db,$name,$ver,$ineq) = @_;
  return 0 if not defined($db->{$name});
  return 1 if not defined($ver);
  return &check_version_inequality($ver,$db->{$name}->{'version'},$ineq);
}










########## LOCAL DATABASE ##########

sub get_local_db
{
  my ($self) = @_;
  if (not defined($self->{DB}->{LOCAL}))
  {
    $self->{DB}->{LOCAL} = {};
    foreach my $i ($self->read_db_dir('local'))
    {
      my $subpath = 'local/'.$i;
      my $dpath = $self->get_db_path($subpath);
      my ($pkgname,$pkgver) = &get_name_and_version($i);
      $self->{DB}->{LOCAL}->{$pkgname} = {
        'name' => $pkgname,
        'version' => $pkgver,
        'dpath' => $dpath
      };
      $self->{DB}->{LOCAL}->{$pkgname}->{'get'} = sub {
        return $self->db_file_get($subpath,@_);
      };
      $self->{DB}->{LOCAL}->{$pkgname}->{'check'} = sub {
        return $self->db_file_check($subpath,@_);
      };
    }
  }
  return $self->{DB}->{LOCAL};
}

sub get_local_db_pkg_ver
{
  my ($self,$pkgname) = @_;
  my $local_db = $self->get_local_db();
  return $local_db->{$pkgname}->{'version'} if exists($local_db->{$pkgname});
}

sub get_local_db_pkg
{
  my ($self,$pkgname,$noload) = @_;
  my $local_db = $self->get_local_db();
  return undef if not defined($local_db->{$pkgname});
  return $local_db->{$pkgname};
}

sub get_local_db_pkgname_list
{
  my ($self) = @_;
  my $local_db = $self->get_local_db();
  return keys(%$local_db) if defined($local_db);
  return ();
}

sub get_local_db_pkg_list
{
  my ($self) = @_;
  my $local_db = $self->get_local_db();
  return values(%$local_db) if defined($local_db);
  return ();
}













sub local_db_pkg_info_search
{
  my $self = shift;
  return $self->pkgs_info_search([$self->get_local_db_pkg_list()],@_);
}

sub get_local_db_provider
{
  my ($self,$pkgname,@args) = @_;
  my $pkg = $self->local_db_pkg_info_search(1,'PROVIDES',$pkgname,@args);
  return $pkg->{'name'} if defined($pkg);
  return undef;
}

sub get_local_db_owner
{
  my ($self,$fpath) = @_;
  my $pkg = $self->local_db_pkg_info_search(1,'FILES',$fpath);
  return $pkg->{'name'} if defined($pkg);
  return undef;
}


sub get_local_db_group
{
  my $self = shift;
  return map {$_->{'name'}} $self->local_db_pkg_info_search(0,'GROUPS',@_);
}

# accepts pkgname and optionally pkgver and inequality
sub is_in_local_db
{
  my ($self,$pkgname,@args) = @_;
  my $db = $self->get_local_db();
  return $self->is_in_db($db,$pkgname,@args);
}

sub is_installed
{
  return &is_in_local_db(@_);
}

# accepts pkgname and optionally pkgver and inequality
sub is_effectively_in_local_db
{
  my ($self,$pkgname,@args) = @_;
  return $pkgname if $self->is_in_local_db($pkgname,@args);
  return $self->get_local_db_provider($pkgname,@args);
}

sub is_effectively_installed
{
  return &is_effectively_in_local_db(@_);
}

sub get_local_pkg_file_bool
{
  my ($self,$pkgname,$what) = @_;
  my $whatever = $self->get_local_db_pkg($pkgname)->{'get'}->($what);
  return defined($whatever);
}

sub is_explicitly_installed
{
  my ($self,$pkgname) = @_;
  return not $self->get_local_pkg_file_bool($pkgname, 'REASON');
}










########## SYNC DATABASE ##########

sub get_sync_db
{
  my ($self) = @_;
  if (not defined($self->{DB}->{SYNC}))
  {
    foreach my $i ($self->read_db_dir('sync'))
    {
      $self->{DB}->{SYNC}->{$i} = undef;
    }
  }
  return $self->{DB}->{SYNC};
}

sub get_sync_db_repo
{
  my ($self,$repo) = @_;
  return undef if not defined($repo);
  my $sync_db = $self->get_sync_db();
  if (not defined($sync_db->{$repo}))
  {
    foreach my $i ($self->read_db_dir('sync',$repo))
    {
      my $subpath = 'sync/'.$repo.'/'.$i;
      my $dpath = $self->get_db_path($subpath);
      my ($pkgname,$pkgver) = &get_name_and_version($i);
      $sync_db->{$repo}->{$pkgname} = {
        'name' => $pkgname,
        'version' => $pkgver,
        'dpath' => $dpath
      };
      $sync_db->{$repo}->{$pkgname}->{'get'} = sub {
        return $self->db_file_get($subpath,@_);
      };
      $sync_db->{$repo}->{$pkgname}->{'check'} = sub {
        return $self->db_file_check($subpath,@_);
      };
      # simple way to enable package specification by "$repo/$pkgname"
      $sync_db->{$repo}->{$repo.'/'.$pkgname} = $sync_db->{$repo}->{$pkgname};
    }
  }
  return $sync_db->{$repo};
}

sub get_sync_db_repos
{
  my ($self) = @_;
  return map {$self->get_sync_db_repo($_)} $self->get_repo_list();
}

sub get_sync_db_repo_lastupdate
{
  my ($self,$repo) = @_;
  return 0 if not defined($repo);
  my $fpath = $self->get_db_path('sync',$repo,'.lastupdate');
  return &get_lastupdate($fpath);
}

sub set_sync_db_repo_lastupdate
{
  my ($self,$repo,$t) = @_;
  return 0 if not defined($repo);
  my $fpath = $self->get_db_path('sync',$repo,'.lastupdate');
  return &set_lastupdate($fpath,$t);
}

sub get_sync_db_pkg
{
  my ($self,$pkgname) = @_;
  my $reponame;
  ($pkgname,$reponame) = &get_name_and_repo($pkgname);
  my @reponames = defined($reponame) ? ($reponame) : $self->get_repo_list();

  foreach $reponame (@reponames)
  {
    my $repo = $self->get_sync_db_repo($reponame);
    if (defined($repo->{$pkgname}))
    {
      return $repo->{$pkgname};
    }
  }
  return undef;
}

sub get_sync_db_pkg_reponame
{
  my ($self,$pkgname) = @_;
  return undef if not defined($pkgname);
  my $reponame;
  ($pkgname,$reponame) = &get_name_and_repo($pkgname);
  return $reponame if defined($reponame);
  foreach $reponame ($self->get_repo_list())
  {
    my $repo = $self->get_sync_db_repo($reponame);
    return $reponame if defined($repo->{$pkgname});
  }
  return undef;
}

sub get_sync_db_pkg_reponames
{
  my $self = shift @_;
  return () if not @_;
  my @pkgnames = ();
  my %reponames = ();
  my $reponame;
  foreach my $pkgname (@_)
  {
    ($pkgname,$reponame) = &get_name_and_repo($pkgname);
    if (defined($reponame))
    {
      $reponames{$reponame} = 1;
    }
    else
    {
      push(@pkgnames,$pkgname);
    }
  }
  foreach $reponame ($self->get_repo_list())
  {
    my $repo = $self->get_sync_db_repo($reponame);
    if (not exists($reponames{$reponame}))
    {
      my @queue = ();
      while (@pkgnames)
      {
        my $pkgname = shift(@pkgnames);
        if (defined($repo->{$pkgname}))
        {
          $reponames{$reponame} = 1;
        }
        else
        {
          push(@queue,$pkgname)
        }
      }
      push(@pkgnames,@queue);
    }
    @pkgnames = grep {not exists($repo->{$_})} @pkgnames;
  }
  return keys %reponames;
}

sub get_sync_db_repo_pkgname_list
{
  my ($self,$repo) = @_;
  $repo = $self->get_sync_db_repo($repo);
  return grep {index($_,'/') == -1} keys(%$repo) if defined($repo);
  return ();
}

sub get_sync_db_pkg_list
{
  my ($self) = @_;
  my @pkgs = ();
  foreach my $repo ($self->get_repo_list())
  {
    my $repo = $self->get_sync_db_repo($repo);
    push @pkgs, map {$repo->{$_}} grep {index($_,'/') == -1} keys(%$repo);
    #push (@pkgs, values(%$repo)) if defined($repo);
  }
  return @pkgs;
}

sub group_pkgnames_by_repo
{
  my ($self,@pkgnames) = @_;
  return {} if not @pkgnames;
  my $href = {};
  foreach my $pkgname (sort @pkgnames)
  {
    my $repo = $self->get_sync_db_pkg_reponame($pkgname);
    $repo = 'foreign' if not defined($repo);
    &refpush($href,$repo,$pkgname);
  }
  return $href;
}








# accepts boolean, section name, pkgname, and optionally version and inequality
# the boolean determines if a single result is returned (true) or multiple (false)
sub sync_db_pkg_info_search
{
  my ($self,@args) = @_;
  return $self->pkgs_info_search([$self->get_sync_db_pkg_list],@args);
}

# accepts pkgname and optionally pkgver and inequality
sub get_sync_db_provider
{
  my ($self,$pkgname,@args) = @_;
  my $pkg = $self->sync_db_pkg_info_search(1,'PROVIDES',$pkgname,@args);
  return $pkg->{'name'} if defined($pkg);
  return undef;
}

# accepts pkgname and optionally pkgver and inequality
sub get_sync_db_replacement
{
  my ($self,$pkgname,@args) = @_;
  my $pkg = $self->sync_db_pkg_info_search(1,'REPLACES',$pkgname,@args);
  return $pkg->{'name'} if defined($pkg);
  return undef;
}

# accepts group name
sub get_sync_db_group
{
  my $self = shift;
  return map {$_->{'name'}} $self->sync_db_pkg_info_search(0,'GROUPS',@_);
}

# accepts pkgname and optionally pkgver and inequality
sub is_in_sync_db
{
  my ($self,$pkgname,@args) = @_;
  foreach my $repo ($self->get_repo_list())
  {
    my $repo = $self->get_sync_db_repo($repo);
    return 1 if $self->is_in_db($repo,$pkgname,@args);
  }
  return 0;
}

# accepts pkgname and optionally pkgver and inequality
sub is_effectively_in_sync_db
{
  my ($self,$pkgname,@args) = @_;
  return $pkgname if $self->is_in_sync_db($pkgname,@args);
  return $self->get_sync_db_provider($pkgname,@args);
}

sub is_forced
{
  my ($self,$pkgname) = @_;
  my $pkg = $self->get_sync_db_pkg($pkgname);
  return 0 if not defined($pkg);
  return defined($pkg->{'get'}->('FORCE'));
}















sub get_sync_db_pkg_urls
{
  my ($self,$pkgname,$arch) = @_;
  my $reponame = $self->get_sync_db_pkg_reponame($pkgname);
  return () if not defined($reponame);
  my $filename = $self->get_sync_db_pkg($pkgname)->{'get'}->('FILENAME')->[0];
  return () if not defined($filename);
  return $self->get_repo_file_urls($reponame,$filename,$arch);
}

sub get_sync_db_pkg_file_array
{
  my ($self,$pkgname,$what) = @_;
  my $whatever = $self->get_sync_db_pkg($pkgname)->{'get'}->($what);
  return @$whatever if defined($whatever);
  return ()
}

sub get_sync_db_pkg_file_scalar
{
  my ($self,$pkgname,$what) = @_;
  my $whatever = $self->get_sync_db_pkg($pkgname)->{'get'}->($what);
  return $whatever->[0] if defined($whatever);
  return undef;
}

sub get_sync_db_pkg_file_bool
{
  my ($self,$pkgname,$what) = @_;
  my $whatever = $self->get_sync_db_pkg($pkgname)->{'get'}->($what);
  return defined($whatever);
}

sub get_sync_db_pkg_deps
{
  my ($self,$pkgname) = @_;
  return $self->get_sync_db_pkg_file_array($pkgname,'DEPENDS');
}

sub get_sync_db_pkg_filename
{
  my ($self,$pkgname) = @_;
  return $self->get_sync_db_pkg_file_scalar($pkgname,'FILENAME');
}

sub get_sync_db_pkg_pkgbase
{
  my ($self,$pkgname) = @_;
  return $self->get_sync_db_pkg_file_scalar($pkgname,'BASE');
}

sub get_sync_db_pkg_md5sum
{
  my ($self,$pkgname) = @_;
  return $self->get_sync_db_pkg_file_scalar($pkgname,'MD5SUM');
}

sub get_sync_db_pkg_csize
{
  my ($self,$pkgname) = @_;
  return $self->get_sync_db_pkg_file_scalar($pkgname,'CSIZE');
}

sub get_sync_db_pkg_isize
{
  my ($self,$pkgname) = @_;
  return $self->get_sync_db_pkg_file_scalar($pkgname,'ISIZE');
}

sub append_sync_db_pkg_csizes
{
  my ($self,@pkgnames) = @_;
  return map {$_ . '[' . $self->get_sync_db_pkg_csize($_) . ']'} @pkgnames;
}

sub append_sync_db_pkg_isizes
{
  my ($self,@pkgnames) = @_;
  return map {$_ . '[' . $self->get_sync_db_pkg_isize($_) . ']'} @pkgnames;
}


sub get_sync_db_pkg_ver
{
  my ($self,$pkgname) = @_;
  foreach my $repo ($self->get_repo_list())
  {
    my $repo = $self->get_sync_db_repo($repo);
    return $repo->{$pkgname}->{'version'} if defined($repo->{$pkgname});
  }
  return undef;
}

















sub get_foreign_pkgnames
{
  my ($self) = @_;
  return grep {not $self->is_in_sync_db($_)} $self->get_local_db_pkgname_list;
}







########## PKG SUBS ##########


sub maybe_load_db_pkg_files
{
  my ($pkg) = @_;
  if (not defined($pkg->{'.lastupdate'}) and defined($pkg->{'dpath'}))
  {
    my $dpath = $pkg->{'dpath'};
    if (opendir(my $dh,$dpath))
    {
      foreach my $f (readdir($dh))
      {
        next if substr($f,0,1) eq '.';
        next if not grep {$f eq $_} qw/desc depends/;#qw/install changelog files/;
        my $fpath = join_paths($dpath,$f);
        my $db_file = &parse_db_file($fpath);;
        $pkg = {
          %$pkg,
          %$db_file,
        };
      }
      $pkg->{'.lastupdate'} = time();
      closedir($dh);
      #$pkg->{'parent'}->{ $pkg->{'name'} } = $pkg;
    }
  }
  return $pkg
}

sub get_db_pkg_filename_by_section
{
  my ($section) = @_;
  if ($section eq 'CONFLICTS'
  or  $section eq 'DEPENDS'
  or  $section eq 'OPTDEPENDS'
  or  $section eq 'PROVIDES')
  {
    return 'depends';
  }
  return 'files' if $section eq 'FILES';
  return 'desc';
}

sub get_db_pkg_file_section_lines
{
  my ($dpath,$section) = @_;
  my $filename = &get_db_pkg_filename_by_section($section);
  my $fpath = &join_paths($dpath,$filename);
  my $content = &get_db_file_section($fpath,$section);
  return defined($content) ? [split("\n",$content)] : undef;
}

sub check_db_pkg_file_section
{
  my ($dpath,$section,$what) = @_;
  my $filename = &get_db_pkg_filename_by_section($section);
  my $fpath = &join_paths($dpath,$filename);
  return &db_file_section_contains($fpath,$section,$what);
}

sub get_pkg_groups
{
  my ($pkg) = @_;
  my $groups = $pkg->{'get'}->('GROUPS');
  return @$groups if defined($groups);
  return ();
}


########## Memoized Functions ##########

sub db_file_get
{
  my ($self,$subpath,$section) = @_;
  if ($self->{MEMOIZE})
  {
    my $k = join '/', $subpath,$section;
    if (exists $self->{MEMOIZED}->{db_file_get}->{$k})
    {
      return $self->{MEMOIZED}->{db_file_get}->{$k}
    }
    my $r = &get_db_pkg_file_section_lines($self->get_db_path($subpath),$section);
    $self->{MEMOIZED}->{db_file_get}->{$k} = $r;
    return $r
  }
  else
  {
    return &get_db_pkg_file_section_lines($self->get_db_path($subpath),$section);
  }
}

sub db_file_check
{
  my ($self,$subpath,$section,$what) = @_;
  if ($self->{MEMOIZE})
  {
    my $k = join '/', $subpath,$section,$what;
    if (exists $self->{MEMOIZED}->{db_file_check}->{$k})
    {
      print "found $k\n";
      return $self->{MEMOIZED}->{db_file_check}->{$k}
    }
    my $r = &check_db_pkg_file_section($self->get_db_path($subpath),$section,$what);
    $r = 0 if not defined($r);
    $self->{MEMOIZED}->{db_file_check}->{$k} = $r;
    return $r
  }
  else
  {
    return &check_db_pkg_file_section($self->get_db_path($subpath),$section,$what);
  }
}



# sub preload_sync_db_data
# {
#   my ($self)
#   {
#     my @reponames = defined($reponame) ? ($reponame) : $self->get_repo_list();
#
#     foreach $reponame (@reponames)
#     {
#       my $repo = $self->get_sync_db_repo($reponame);
#       foreach my $pkg (values(%$repo))
#       {
#         my $dpath = $pkg->{dpath};
#         my $filename = &get_db_pkg_filename_by_section($section);
#         my $fpath = &join_paths($dpath,$filename);
#         my $content = &get_db_file_section($fpath,$section);
#         return defined($content) ? [split("\n",$content)] : undef;
#       }
#     }
#   }
# }








########## DB FILES ##########

sub parse_db_file
{
  my ($fpath) = @_;
  return undef if not defined($fpath);
  my $contents = {};
  my $section = '';
  if (open(my $fh,'<',$fpath))
  {
    while(defined(my $line = <$fh>))
    {
      chomp $line;
      #if ($line =~ m/^\s*%([^%]+)%\s*$/)
      if (substr($line,0,1) eq '%' and substr($line,-1) eq '%')
      {
        $section = $1;
        $contents->{$section} = [];
      }
      #elsif ($line =~ m/^\s*(\S.*?)\s*$/)
      elsif ($section)
      {
        push(@{ $contents->{$section} }, $1);
      }
    }
    close($fh);
    return $contents;
  }
  else
  {
    return undef;
  }
}

sub format_db_file
{
  my ($contents) = @_;
  my $text = '';
  foreach my $k (sort keys %$contents)
  {
    $text .= "%$k%\n";
    $text .= "$_\n" foreach @{ $contents->{$k} };
    $text .= "\n";
  }
  return $text;
}

sub save_db_file
{
  my ($fpath,$contents) = @_;
  my $text = &format_db_file($contents);
  if (open(my $fh, '>', $fpath))
  {
    print $fh $text;
    close $fh
  }
  else
  {
    print STDERR "error: failed to open $fpath ($!)\n";
    exit(1);
  }
}

sub db_file_section_contains
{
  my ($fpath,$section,$what) = @_;
  return undef if not defined($what);
  my $content = &get_db_file_section($fpath,$section);
  return undef if not $content;
  my $l = length($what) + 1;
  # This is faster than a regular expressions according to simple benchmarks.
  return (
    $content eq $what or
    substr($content,0,$l) eq "$what\n" or
    substr($content,-$l) eq "\n$what" or
    index($content,"\n$what\n") > -1 or
    (
      substr($fpath,-7) eq 'depends' and
      (
        substr($content,0,$l) eq "$what>" or
        substr($content,0,$l) eq "$what=" or
        substr($content,0,$l) eq "$what<" or
        index($content,"\n$what>") > -1 or
        index($content,"\n$what=") > -1 or
        index($content,"\n$what<") > -1
      )
    )
  );
}

sub get_db_file_section
{
  my ($fpath,$section) = @_;
  return undef if not defined($section);
  my $content = &slurp_file($fpath);
  return undef if not defined($content);
  my $i = index($content,'%'.$section."%\n");
  if ($i > -1)
  {
    $i += length($section) + 3;
    my $j = index($content,"\n%",$i);
    my $lines;
    if ($j > $i)
    {
      return substr($content,$i,$j-$i);
    }
    else
    {
      return substr($content,$i);
    }
  }
  return undef;
}

sub get_lastupdate
{
  my ($fpath) = @_;
  if (open(my $fh, '<', $fpath))
  {
    my $lastupdate = <$fh>;
    close $fh;
    chomp $lastupdate;
    return $lastupdate;
  }
  return 0;
}

sub set_lastupdate
{
  my ($fpath,$t) = @_;
  $t = time if not $t;
  if (open(my $fh, '>', $fpath))
  {
    print $fh $t;
    close $fh;
    return $t
  }
  return 0;
}





########## PKGNAME AND PKGVER ##########

sub compare_versions
{
  my ($a,$b) = @_;
  return -1 if not defined($a) and defined($b);
  return 1 if defined($a) and not defined($b);
  return 0 if not (defined($a) or defined($b));

  my ($rel_a,$rel_b) = (0,0);
  if ($a =~ m/(.*)-(.+)$/) {($a,$rel_a)=($1,$2);}
  if ($b =~ m/(.*)-(.+)$/) {($b,$rel_b)=($1,$2);}

  my @a = split(/\./,$a);
  my @b = split(/\./,$b);
  my $len_a = scalar @a;
  my $len_b = scalar @b;
  my $n = ($len_a < $len_b) ? $len_a : $len_b;
  for (my $i=0; $i<$n; $i++)
  {
    my ($int_a,$alph_a,$dash_a) = ($a[$i] =~ m/^(\d*)(.*?)(_.*)?$/);
    my ($int_b,$alph_b,$dash_b) = ($b[$i] =~ m/^(\d*)(.*?)(_.*)?$/);
    $int_a = 0 if $int_a eq '';
    $int_b = 0 if $int_b eq '';
    if ($int_a != $int_b)
    {
      return $int_a <=> $int_b;
    }
    if (length($alph_a) and not length($alph_b))
    {
      return -1;
    }
    if (not length($alph_a) and length($alph_b))
    {
      return 1;
    }
    if ($alph_a ne $alph_b)
    {
      return $alph_a cmp $alph_b;
    }
    return 1 if defined($dash_a) and not defined($dash_b);
    return -1 if not defined($dash_a) and defined($dash_b);
    if (defined($dash_a) and defined($dash_b))
    {
      my $dash_cmp = $dash_a cmp $dash_b;
      return $dash_cmp if $dash_cmp != 0;
    }
  }
  return $len_a <=> $len_b if not $len_a == $len_b;
  return $rel_a <=> $rel_b;
}

sub get_name_and_version
{
  my ($pkg) = @_;
  if ($pkg =~ m/^(.+)-([^-]+-\d+(?:\.\d+)?)$/)
  {
    return ($1,$2);
  }
  else
  {
    return ($pkg,undef);
  }
}

sub get_name_and_ver_spec
{
  my ($pkg) = @_;
  if ($pkg =~ m/^(.+?)(>=|>|=|<=|<)(.*)$/)
  {
    return ($1,$3,$2);
  }
  else
  {
    return ($pkg,undef,undef);
  }
}

sub check_version_inequality
{
  my ($v1,$v2,$i) = @_;
  # strip off the pkgrel
  $v1 =~ s/-.+$// if $v1;;
  $v2 =~ s/-.+$// if $v2;
  my $c = &compare_versions($v1,$v2);
  $i = '=' if not $i;
  return &check_cmpver_inequality($c,$i);
}

sub check_cmpver_inequality
{
  my ($c,$i) = @_;
  if (
    ($c == 1 and substr($i,0,1) eq '<')
  or
    ($c == 0 and substr($i,-1) eq '=')
  or
    ($c == -1 and substr($i,0,1) eq '>')
  )
  {
    return 1;
  }
  return 0;
}

sub check_name_and_ver_spec
{
  my ($name,$ver,$ineq,$spec) = @_;
  my ($n,$v,$i) = &get_name_and_ver_spec($spec);
  return 0 if $name ne $n;
  return 1 if not defined($ver);# or not $i;
  return &check_version_inequality($ver,$v,$ineq);
}

sub get_name_and_repo
{
  my ($pkgname) = @_;
  my $i = index($pkgname,'/');
  if ($i > -1)
  {
    return (substr($pkgname,$i+1),substr($pkgname,0,$i));
  }
  else
  {
    return ($pkgname,undef);
  }
}





########## URL FORMATTING ##########

sub replace_url_architecture
{
  my ($arch,@urls) = @_;
  return @urls if not defined($arch);
  # Replace known architectures with the one given, skipping the one specified.
  foreach my $a (grep {$_ ne $arch} @ARCHITECTURES)
  {
    @urls = map {$_=~s/(^|\/)\Q$a\E(\/|$)/$1$arch$2/g; $_} @urls;
  }
  return @urls;
}

sub replace_url_repo
{
  my ($reponame,@urls) = @_;
  return @urls if not defined($reponame);
  return map {$_=~s/(^|\/)\$repo(\/|$)/$1$reponame$2/; $_} @urls;
}

1;

__END__

=head1 NAME

Xyne::Arch::Pacman - A pacman class.

=head1 SYNOPSIS

  use Xyne::Arch::Pacman;

=head1 DESCRIPTION

A partial implementation of pacman primarily for use in
scripts. The pacman binary is still used for more functionality.

=head2 EXPORT

None by default.

=head1 AUTHOR

Xyne <lt>xyne at archlinux dot ca<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009,2010  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
