# Copyright (C) 2009,2010  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


# This needs to be rewritten to work with the new OptionHandler module and to
# use JSON to pass hashes between PkgDs instead of the current ugly hack.

package Xyne::Arch::PkgD;
use strict;
use warnings;

use File::Basename qw/basename/;
use IO::Socket;
#use Thread::Queue;

use Xyne::Common::Functions qw/get_unique join_paths str/;
use Xyne::Common::JSON qw/json_to_obj obj_to_json/;
use Xyne::Common::Log;
use Xyne::Common::OptionHandler;
use Xyne::Common::Threads qw/parallelize/;

require Exporter;

our @EXPORT_OK = qw/get_files_from_pkgds request_file_via_pkgd/;

our $VERSION = '0.22';

our $DEFAULT_PORT = 14687;
our $DEFAULT_CONFIG_PATH = '/etc/pkgd.conf';

our @OPTIONS = (
  {
    'name' => 'Address',
    'aliases' => ['-a','--address'],
    'type' => 'str',
    'description' => 'Bind PkgD to this address, e.g. "-a 192.168.0.1". Leave it blank to allow access to PkgD through all interfaces.',
    'parameter_name' => '<address>'
  },

  {
    'name' => 'CacheDir',
    'aliases' => ['--cachedir'],
    'type' => 'list',
    'default' => ['/var/cache/pacman/pkg'],
    'description' => 'The path to the package cache. Multiple may be specified and they will be checked in that order.',
    'parameter_name' => '<path>'
  },

  {
    'name' => '--config',
    'type' => 'str',
    'default' => $DEFAULT_CONFIG_PATH,
    'description' => 'The path to the PkgD configuration file.',
    'parameter_name' => '<path>'
  },

  {
    'name' => 'Interval',
    'aliases' => ['-i','--interval'],
    'type' => 'int',
    'default' => 300,
    'description' => 'The PkgD master notification interval (in seconds), i.e. how often PkgD should inform master PkgDs that it is available. The default is five minutes (300 s).',
  },

  {
    'name' => 'Log',
    'aliases' => ['-l','--log'],
    'type' => 'str',
    'description' => 'The PkgD log file.',
    'recommend' => '/var/log/pkgd.log',
  },

  {
    'name' => 'Master',
    'aliases' => ['-m','--master'],
    'type' => 'list',
    'description' => 'An address and port of a master PkgD. Multiple may be specified.',
    'parameter_name' => '<address:port>'
  },

  {
    'name' => 'MaxDepth',
    'aliases' => ['--max-depth'],
    'type' => 'int',
    'default' => 3,
    'description' => 'The maximum number of nodes in each query path, e.g. each immediate slave is the first node, each slave of a slave is a second node, etc.',
  },

  {
    'name' => 'Port',
    'aliases' => ['-p','--port'],
    'type' => 'int',
    'default' => $DEFAULT_PORT,
    'description' => 'The port on which to listen, e.g. "-p 44567". The default port is '.$DEFAULT_PORT,
    'parameter_name' => '<port>'
  },

  {
    'name' => 'Quiet',
    'aliases' => ['-q','--quiet'],
    'type' => 'switch',
    'description' => 'Suppress console output (this has no effect on logging).'
  },
  
  
  
  

  {
    'name' => '--get-pid',
    'type' => 'switch',
    'description' => 'Return the pid of a running PkgD. This is mostly for handling PkgD daemons.'
  },

  {
    'name' => '--help',
    'aliases' => ['-h'],
    'type' => 'switch',
    'description' => 'Display the help message and exit.',
  },

  {
    'name' => '--man',
    'type' => 'switch',
    'description' => 'Display the man page and exit.',
  },

  {
    'name' => '--man-nroff',
    'type' => 'switch',
    'description' => 'Display the man page in nroff format and exit.',
  },

  {
    'name' => '--get-config-template',
    'type' => 'switch',
    'description' => 'Display a configuration file template and exit.',
  },
);







our $MANPAGE = {
  'header' => 'PKGD 1 "January 2010"',
  'sections' => [
    ['NAME', 'pkgd - share your pacman packages over your lan'],
    ['SYNOPSIS', 'pkgd [options]'],
    ['DESCRIPTION', "PkgD  (Package Daemon) is intended to facilitate package sharing over a LAN to avoid unnecessary downloads from the external network. By running a network of pkgds you can create the equivalent of a shared cache."],
    ['CONFIGURING PKGD', "Select at least one computer on which to run a \"master\" pkgd. This will be the server that you  connect  to  when attempting  to  retrieve  packages. A master pkgd does not require any special configuration to make it a master. Master status is determined by how other \"slave\" pkgds are configured.\n\nOnce the master pkgd is set up, you should start a \"slave\" pkgd on the  other  computers  in  your  network.  You should  configure each slave to point to the master by adding a \"Master = <address:port>\" line to the slave's configuration file, e.g. \"Master = 192.168.0.1:$DEFAULT_PORT\".\n\nWhen the master pkgd is queried for a package, it will first check its own cache. If the  package  is  not  found then  it  will  check each slave pkgd that it knows about for the package until either the package is found or it has queried all of the slaves. If the package cannot be found, the master pkgd will  return  a  \"404  Not  Found\"error.\n\nAlternative  network  topologies are supported, including cyclical master-slave relationships. Pkgd should detect and prevent infinite-loop queries so you should experiment with other setups to find one which works for you."],
    ['ARCHITECTURES', "It may be necessary to take system architecture into account when running PkgD. Previously, packages in the community repo had the architecture stripped from the package filename which created overlapping namespaces between different architectures. This is no longer done but there may still be some packages which have not yet been updated to the current naming scheme. In the event that a package for the wrong architecture is downloaded, Pacman should detect a mismatched checksum and refuse to install the package."],
    ['PORTS AND SECURITY', "You must open a port for PkgD when running it on a computer with a firewall or other computers in your LAN will not be able to connect to it (yes, I realize that if you have a firewall on Arch Linux, then you probably don't need to be told this, but it's better to be too clear than not clear enough). However, you do not need to, nor should you open ports in your router's firewall or any other firewall which stands between your LAN and an external network. PkgD is only intended to be used on a LAN to avoid downloads from the external network.\n\nAlthough the author has tried to avoid potential security exploits while writing the code, this has not been a primary focus nor is the author a security expert. As the license states, PkgD comes with NO WARRANTY whatsoever. If you wish to deploy it in an unsafe environment, you are advised to review the code yourself."],
    ['PACMAN', "To use PkgD with Pacman, simply specify the PkgD master servers the same way that you specify other http servers in the Pacman configuration file, e.g.\n\n    [core]\n\n    Server = http://192.168.0.1:$DEFAULT_PORT/\$repo\n\nPlace the master PkgD server(s) at the top of each repo section or at the top of your mirrorlist to make sure that Pacman checks them before downloading from an external server.\n\nNote that pacman will report an error when a PkgD server cannot find a package. It will then move on to the next server in the list. This is due to the way pacman handle's 404 errors. Despite the errors, pacman will continue to retrieve packages as expected. Note that these errors can be avoided by using Powerpill."],
    ['OPTIONS', "PkgD recognizes the following options:\n"],
    ['AUTHORS', 'Xyne']
  ]
};



our $CONFIG_TEMPLATE = {
#  'header' => "",
  'sections' => [
    ['options', ""],
  ]
};



our $HELP_MESSAGE = {
  'header' => "USAGE\n\n    powerpill [powerpill options] [pacman options] [args]\n\n    See \"--pacman-help\" for help with Pacman's options.\n\n\nPOWERPILL OPTIONS\n\n",
  'footer' => "",
};





########## CONSTRUCTOR ##########

sub new {
  my ($proto,$arg) = @_;
  my $class = ref($proto) || $proto;
  my $self  = {};
  bless ($self, $class);
  
  $self->{CONF} = Xyne::Common::OptionHandler->new();
  $self->{CONF}->declare_options(@OPTIONS);
  $self->{CONF}->{MANPAGE} = $MANPAGE;
  $self->{CONF}->{CONFIG_TEMPLATE} = $CONFIG_TEMPLATE;
  $self->{CONF}->{HELP_MESSAGE} = $HELP_MESSAGE;
  
  $self->{SLAVES} = [];
  
  return $self;
}




sub main
{
  my ($self,@args) = @_;
  $self->{CONF}->parse_cli_input(@args);
    
  if ($self->{CONF}->get_value('--help'))
  {
    print $self->{CONF}->get_cli_help_message;
    exit;
  }
    
  elsif ($self->{CONF}->get_value('--man'))
  {
    print $self->{CONF}->get_cli_simple_man_page;
    exit;
  }
    
  elsif ($self->{CONF}->get_value('--man-nroff'))
  {
    print $self->{CONF}->get_cli_nroff_man_page;
    exit;
  }
    
  elsif ($self->{CONF}->get_value('--get-config-template'))
  {
    print $self->{CONF}->get_config_template;
    exit;
  }

  my $config = $self->{CONF}->get_value('--config');
  $self->{CONF}->reload({'config'=>$config});
    
  if ($self->{CONF}->get_value('--get-pid'))
  {
    print $self->get_pid;
    exit;
  }

  $self->maybe_notify_masters;
  $self->{CHILD_PIDS} = Thread::Queue->new();
  my $reaper_thr = threads->create('reap_children',$self->{CHILD_PIDS});
  $reaper_thr->detach;
  $self->listen;
}


sub message
{
  my ($self,$type,@msgs) = @_;
  if (not $self->{CONF}->get_value('Quiet'))
  {
    my $h = ($type eq 'ERROR') ? *STDERR : *STDOUT;
    @msgs = map {lc($type).': '.$_} @msgs if $type ne 'STANDARD';
    print $h $_ foreach @msgs;
  }
  
  my $lpath = $self->{CONF}->get_value('Log');
  if (defined($lpath))
  {
    $self->{LOG} = Xyne::Common::Log->new($lpath) if not $self->{LOG};
    $self->{LOG}->log(@msgs);
  }  
}





########## SLAVE HANDLING ##########

sub get_slaves
{
  my ($self) = @_;
  return @{$self->{SLAVES}};
}


sub set_slaves
{
  my ($self,@slaves) = @_;
  $self->{SLAVES} = [grep {defined($_)} @slaves];
}


sub maybe_register_slave
{
  my ($self,$ip,$port) = @_;
  my $slave = $ip.':'.$port;
  my @slaves = $self->get_slaves;
  if (not grep {$_ eq $slave} @slaves)
  {
    $self->set_slaves(@slaves,$slave);
    $self->message('STANDARD',"registered slave $slave\n");
  }
  else
  {
    $self->message('STANDARD',"slave $slave has already been registered\n");
  }
}









########## REQUEST HANDLING ##########

sub listen
{
  my ($self) = @_;
  my $port = $self->{CONF}->get_value('Port');
  my %args = (
    LocalPort => $port,
    Listen => 10,
    Reuse => 1,
    Proto => 'tcp'
  );
  my $addr = $self->{CONF}->get_value('Bind');
  $args{'LocalAddr'} = $addr if defined($addr);
  if (my $sock = new IO::Socket::INET(%args))
  {
    $self->message('STANDARD',"PkgD now listening on port $port...\n");

    while(my $client = $sock->accept())
    {
      $self->handle_request($client);
    }
    close($sock);
  }
  else
  {
    $self->message('ERROR', "failed to create socket: $!\n");
    exit(1);
  }
}



sub handle_request
{
  my ($self,$client) = @_;
  my $client_ip = $client->peerhost;
  my $request = &read_socket($client);
  $self->message('STANDARD',"request received from $client_ip: $request\n\n");
  if ($request =~ m/^\{/)
  {
    my $obj = &json_to_obj($request);
    return if not ref($obj) eq 'HASH' or not defined($obj->{'operation'});
    my $operation = $obj->{'operation'};
    
    if ($operation eq 'get_pid')
    {
      if ($client_ip eq '127.0.0.1')
      {
        $self->message('STANDARD',"returning PID $$ to $client_ip\n");
        print $client "$$\r\n\r\n";
      }
      else
      {
        $self->message('STANDARD',"request denied\n");
      }
      close $client;
      return
    }
    elsif ($operation eq 'notify_master')
    {
      my $port = $obj->{'port'};
      $self->maybe_register_slave($client_ip,$port);
      close $client;
      return
    }
    elsif ($operation eq 'query_file')
    {
      my $obj = $self->query_file($obj);
      my $response = &obj_to_pkgd_json($obj);
      $self->message('STANDARD',"returning query results to$client_ip: $response\n");
      print $client "$response\r\n\r\n";
      close $client;
      return
    }
    elsif ($operation eq 'request_file')
    {
      my $pid = fork();
      if ($pid == 0)
      {
        $self->handle_pkgd_file_request($obj,$client);
        exit;
      }
      else
      {
        $self->{CHILD_PIDS}->enqueue($pid);
      }
      return
    }
    else
    {
      $self->message('STANDARD',"ignoring unrecognized operation: $operation\n");
    }
  }
  elsif ($request =~ m/^GET \/([^\/]+)\/(.+?) HTTP\/\d\.\d\r$/m)
  {
    my ($repo,$filename) = ($1,$2);
    my $pid = fork();
    if ($pid == 0)
    {
      $self->handle_http_file_request($client,$repo,$filename,$request);
      exit;
    }
    else
    {
      $self->{CHILD_PIDS}->enqueue($pid);
    }
    return
  }
  else
  {
    $self->message('STANDARD',"dropped unrecognized request from $client_ip\n");
    return
  }
}



sub handle_http_file_request
{
  my ($self,$client,$repo,$filename,$request) = @_;
  my $client_ip = $client->peerhost;
  my $obj = $self->query_file({
    'operation' => 'query_file',
    'query' => {
      $filename => {'repo' => $repo}
    }
  },1);
  if (defined($obj->{'found'}->{$filename}->{'address'}))
  {
    my $address = $obj->{'found'}->{$filename}->{'address'};
    if ($address eq '%SELF%')
    {
      my $fpath = $obj->{'found'}->{$filename}->{'path'};
      $self->return_http_file($client,$fpath,$request);
      return;
    }
    else
    {
      $self->message('STANDARD',"redirecting $client_ip to $address\n");
      print $client "HTTP/1.1 303 See Other\r\n";
      print $client "Location: http://$address/$repo/$filename\r\n\r\n";
      close $client;
      return;
    }
  }
  else
  {
    $self->message('STANDARD',"returning 404 to $client_ip\n");
    print $client "HTTP/1.1 404 Not Found\r\n\r\n";
    close $client;
  }
}




sub return_http_file
{
  my ($self,$client,$file,$request) = @_;
  my $client_ip = $client->peerhost;
  $self->message('STANDARD',"sending $file to $client_ip\n");
  my $size = (stat($file))[7];
  my $range_start;
  my $range_end;
  if ($request =~ m/Range: bytes=(\d*)-(\d*)/)
  {
    $range_start = $1;
    $range_end = $2;
  }
  $range_start = 0 if not defined($range_start);
  $range_end = $size - 1 if not $range_end;
  my $remaining = $range_end - $range_start;
  my $header = "HTTP/1.1 200 OK\r\nAccept-Ranges: bytes\r\nContent-Length: $size\r\n";
  $header .= "Content-Range: bytes $range_start-$range_end/$size\r\n" if $remaining != $size;
  $header .= "Connection: close\r\nContent-Type: application/x-gzip\r\n\r\n";
  print $client $header;
  if (open(my $fh,'<',$file))
  {
    binmode $fh;
    binmode $client;
    seek($fh,$range_start,0) if $range_start > 0;
    my $data;
    my $step = 1024;
    while(defined($client) and my $r = read( $fh, $data, $step ))
    {
      print $client $data;
      $client->flush;
      $remaining -= $r;
      $step = $remaining + 1 if $remaining < $step;
    }
    close $fh;
  }
  close($client);
}






##########  PKGD ##########

sub get_pid
{
  my ($self) = @_;
  my $json = &obj_to_json({'operation'=>'get_pid'});
  return &simple_send_request($json, '127.0.0.1', $self->{CONF}->get_value('Port'));
}



sub maybe_notify_masters
{
  my ($self) = @_;
  my @masters = $self->{CONF}->get_value('Master');
  if (@masters)
  {
    my %masters;
    foreach my $master (@masters)
    {
      my ($url,$port) = &get_ip_and_port($master);
      $masters{$url} = $port;
    }
    my $args;
    $args->{'pid'} = $$;
    $args->{'port'} = $self->{CONF}->get_value('Port');
    $args->{'masters'} = \%masters;
    $args->{'interval'} = $self->{CONF}->get_value('Interval');
    my $thr = threads->create('notify_masters',$args);
  }
}



sub check_cachedirs_for_file
{
  my ($self,$filename) = @_;
  return undef if not defined($filename);
  # prevent relative paths from getting out of the directory
  $filename = basename($filename);
  foreach my $cachedir ($self->{CONF}->get_value('CacheDir'))
  {
    my $fpath = &join_paths($cachedir,$filename);
    return $fpath if (-r $fpath);
  }
  return undef;
}



sub query_file
{
  my ($self,$obj,$get_path) = @_;
  return $obj if ref($obj->{'query'}) ne 'HASH';
  
  my $found = {};
  foreach my $filename (keys(%{$obj->{'query'}}))
  {
    my $fpath = $self->check_cachedirs_for_file($filename);
    if (defined($fpath))
    {
      $found->{$filename} = {
        'address' => '%SELF%'
      };
      $found->{$filename}->{'path'} = $fpath if $get_path;
      delete($obj->{'query'}->{$filename});
    }
  }
  
  my $slaves_obj;
  if (%{$obj->{'query'}})
  {
    $slaves_obj = &file_query_pkgds($obj, $self->get_slaves);
  }
  
  # Update the object after the query to avoid passing unneeded information.
  $obj = &update_file_query_obj($obj,$found);
  $obj = &update_file_query_obj($obj,$slaves_obj->{'found'}) if defined($slaves_obj->{'found'});
  return $obj;
}



sub handle_pkgd_file_request
{
  my ($self,$obj,$client) = @_;
  return if not defined($obj->{'filename'});
  my $client_ip = $client->peerhost;
  
  my $fpath = $self->check_cachedirs_for_file($obj->{'filename'});
  if (not defined($fpath))
  {
    $self->message('STANDARD',"$obj->{'filename'} not found in cache\n");
    $obj->{'found'} = 0;
    print $client &obj_to_pkgd_json($obj);
    close $client;
    return;
  }
  else
  {
    $obj->{'found'} = 1;
    $obj->{'size'} = (stat($fpath))[7];
    print $client &obj_to_json($obj);
    print $client "\r\n\r\n";
  }
  if (open(my $fh,'<',$fpath))
  {
    $self->message('STANDARD',"returning $fpath to $client_ip\n");
    binmode $fh;
    binmode $client;
    my $data;
    my $step = 4096;
    while(defined($client) and my $r = read( $fh, $data, $step ))
    {
      print $client $data;
    }
    close $fh;
  }
  else
  {
    $self->message('WARNING',"failed to open $fpath\n");
  }
  close($client);
}







#####################################
########## CLASS FUNCTIONS ##########
#####################################

sub get_ip_and_port
{
  my ($server) = @_;
  if ($server =~ m/^(.*?):(\d+)$/)
  {
    return ($1,$2);
  }
  return ($server, $DEFAULT_PORT);
}



sub read_socket
{
  my ($sock) = @_;
  return '' if not $sock;
  my $data = '';
  my $byte;
  while(read( $sock, $byte, 1 ))
  {
    $data .= $byte;
    last if (substr($data,-4) eq "\r\n\r\n");
  }
  $data = substr($data,0,-4) if substr($data,-4) eq "\r\n\r\n";
  return $data;
}

sub send_request
{
  my ($request,$url,$port) = @_;
  my $sock = new IO::Socket::INET(
    PeerAddr => $url,
    PeerPort => $port,
    Proto => 'tcp')
  || return '';

  $request .= "\r\n\r\n" if substr($request,-4) ne "\r\n\r\n";
  print $sock $request;
  $sock->flush;
  return $sock; 
}

sub simple_send_request
{
  my ($request,$url,$port) = @_;
  my $sock = &send_request($request,$url,$port);
  my $response = &read_socket($sock);
  close($sock); 
  return $response;
}

# Keep this centralized for future protocol changes.
sub obj_to_pkgd_json
{
  return &obj_to_json(@_);
}






sub notify_masters
{
  my ($args) = @_;
  my $parent_pid = $args->{'pid'};
  my $port = $args->{'port'};
  my $interval = $args->{'interval'};
  $interval = 900 if not defined($interval);
  my %masters = %{ $args->{'masters'} };
  exit if not $parent_pid or not $port or not %masters;
  sleep 3;
  while(kill 0,$parent_pid)
  {
    foreach my $ip (keys %masters)
    {
      &send_request(&obj_to_pkgd_json({'operation'=>'notify_master','port'=>$port}),$ip,$masters{$ip});
    }
    sleep $interval;
  }
}

sub reap_children
{
  my ($queue) = @_;
  while (defined(my $pid = $queue->dequeue))
  {
    waitpid $pid,0;
  }
}










sub get_file_query_obj_from_list
{
  my @filenames = @_;
  my $obj = {
    'operation' => 'query_file',
    'query' => {}
  };
  $obj->{'query'}->{$_} = {} foreach @filenames;
  return $obj;
}



sub update_file_query_obj
{
  my ($obj,$found) = @_;
  return $obj if not ref($obj) eq 'HASH' or not ref($found) eq 'HASH';
  $obj->{'found'} = {} if not ref($obj->{'found'}) eq 'HASH';
  foreach my $filename (keys(%$found))
  {
    $obj->{'found'}->{$filename} = $found->{$filename} if not defined($obj->{'found'}->{$filename});
  }
  return $obj;
}



sub file_query_pkgds
{
  my ($obj,@pkgds) = @_;
  return $obj if not ref($obj->{'query'}) eq 'HASH' or not %{$obj->{'query'}};
  $obj->{'checked'} = [] if not ref($obj->{'checked'});
  my $found = {};
  
  foreach my $pkgd (@pkgds)
  {
    my ($ip,$port) = &get_ip_and_port($pkgd);
    next PKGD if grep {$_ eq $pkgd} @{ $obj->{'checked'} };
    push(@{ $obj->{'checked'} },$pkgd);
    
    my $json = &obj_to_pkgd_json($obj);
    my $response = &simple_send_request($json,$ip,$port);
    $response =~ s/%SELF%/$pkgd/g;
    
    my $slaves_found = &json_to_obj($response)->{'found'};
    next if ref($slaves_found) ne 'HASH';
    foreach my $filename (keys(%$slaves_found))
    {
      # skip anything which was not originally queried
      next if not exists($obj->{'query'}->{$filename});
      if (defined($slaves_found->{$filename}->{'address'}))
      {
        $found->{$filename} = {
          'address' => $slaves_found->{$filename}->{'address'}
        };
        delete($obj->{'query'}->{$filename});
      }
    }
    last if not %{$obj->{'query'}};
  }
  $obj = &update_file_query_obj($obj,$found);
  return $obj;
}



sub get_files_from_pkgds
{
  my ($cachedir,$files,$pkgds) = @_;
  return (undef) if not (-d $cachedir and -w $cachedir) or not ref($files) eq 'ARRAY' or not ref($pkgds) eq 'ARRAY';
  my @pkgds = @$pkgds;
  my $obj = &get_file_query_obj_from_list(@$files);
  $obj = &file_query_pkgds($obj,@pkgds);
  
  return if ref($obj->{'found'}) ne 'HASH';
  
  my @args = ();
  foreach my $filename (keys(%{ $obj->{'found'} }))
  {
    my $address = $obj->{'found'}->{$filename}->{'address'};
    push(@args,[$address,$filename,$cachedir]) if defined($address);
  }
  return &parallelize(\&request_file_via_pkgd,@args);
}



sub request_file_via_pkgd
{
  return undef if grep {not defined($_)} @_;
  my ($address,$filename,$cachedir) = @_;
  my $obj = {
    'operation' => 'request_file',
    'filename' => $filename
  };
  my $sock = &send_request(&obj_to_json($obj),&get_ip_and_port($address));
  my $response = &read_socket($sock);
  $obj = &json_to_obj($response);
  if ($obj->{'found'})
  {
    my $fpath = join_paths($cachedir,basename($filename));
    my $size = $obj->{'size'};
    if (open(my $fh,'>',$fpath))
    {
      my $data;
      my $step = 4096;
      while(read( $sock, $data, $step ))
      {
        print $fh $data;
      }
      close $fh;
      return $fpath;
    }
  }
  return undef;
}







1;

__END__

=head1 NAME

Xyne::Arch::PkgD - A class for sharing pacman packages over a LAN.

=head1 SYNOPSIS

  use Xyne::Arch::PkgD;

=head1 DESCRIPTION

A class to provide client and server functionality for sharing
pacman packages over a LAN to avoid unnecessary downloads from
the external network.

=head2 EXPORT

None by default.

=head1 AUTHOR

Xyne <lt>xyne at archlinux dot ca<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009,2010  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
