# Copyright (C) 2009,2010  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


package Xyne::Arch::Powerpill;
use strict;
use warnings;

use Cwd qw/abs_path/;
use File::Basename qw/basename/;
use IPC::Open2;

use Xyne::Arch::Pacman qw/get_name_and_version compare_versions/;
use Xyne::Arch::PkgD qw/get_files_from_pkgds/;
use Xyne::Arch::PowerpillConf;
use Xyne::Arch::Reflector;
use Xyne::Common::Functions qw/colorize confirm format_bytes get_term_width join_paths str get_unique wrap_words get_file_md5 get_file_sha/;
use Xyne::Common::Metalink;
use Xyne::Common::Tar qw/extract_archive/;
use Xyne::Common::Threads qw/parallelize/;

our $VERSION = '0.33';

our @ISA = qw/Xyne::Arch::Pacman/;







########## CONSTRUCTOR ##########

sub new {
  my ($proto,$args) = @_;
  my $class = ref($proto) || $proto;
  my $self  = {};
  bless ($self, $class);


  $self->reset_db;

  if ($args->{'autoload'})
  {
    $args->{'pacman_autoload'} = 1;
    $args->{'powerpill_autoload'} = 1;
  }


  if (defined($args->{PacmanConf}))
    { $self->{PacmanConf} = $args->{PacmanConf}; }
  else
    { $self->{PacmanConf} = Xyne::Arch::PacmanConf->new(); }
  $self->{PacmanConf}->load_config($args->{'pacman_config_path'}) if $args->{'pacman_autoload'};


  if (defined($args->{PowerpillConf}))
    { $self->{PowerpillConf} = $args->{PowerpillConf}; }
  else
  {
    $self->{PowerpillConf} = Xyne::Arch::PowerpillConf->new({
      'PacmanConf' => $self->{PacmanConf}
    });
  }
  $self->{PowerpillConf}->load_config($args->{'powerpill_config_path'}) if $args->{'powerpill_autoload'};


  return $self;
}







########## DEBUGGING ##########
sub get_report
{
  my ($self,$report) = @_;
  $report = &Xyne::Common::Functions::get_report($report, 'Powerpill', ['Database', $self->{DB}]);
  $report = $self->{PowerpillConf}->get_report($report);
  $report = $self->{PacmanConf}->get_report($report);
  return $report;
}







########## CONVENIENCE METHODS  ##########

sub get_config_template
{
  my ($self) = @_;
  return $self->{PowerpillConf}->get_config_template();
}

sub get_help_message
{
  my ($self) = @_;
  return $self->{PowerpillConf}->get_cli_help_message();
}

sub get_man_page
{
  my ($self) = @_;
  return $self->{PowerpillConf}->get_cli_simple_man_page();
}

sub get_man_page_nroff
{
  my ($self) = @_;
  return $self->{PowerpillConf}->get_cli_nroff_man_page();
}

sub get_value
{
  my ($self,$opt) = @_;
  return $self->{PowerpillConf}->get_value($opt);
}

sub get_pacvalue
{
  my ($self,$opt) = @_;
  return $self->{PacmanConf}->get_value($opt);
}

sub get_repo_list
{
  my ($self) = @_;
  return $self->{PacmanConf}->get_repo_list()
}

sub load_config
{
  my ($self) = @_;
  return $self->{PowerpillConf}->load_config();
}

sub parse_cli_input
{
  my ($self,@args) = @_;
  return $self->{PowerpillConf}->parse_cli_input(@args);
}







########## MAIN ##########
sub main
{
  my ($self,@args) = @_;
  $SIG{INT} = sub{
    $self->abort(@_);
  };
  my ($clopts,$popts,$clargs) = $self->parse_cli_input(@args);

  if (not @args or $self->get_value('--help'))
  {
    print $self->get_help_message;
    exit;
  }

  elsif ($self->get_value('--man'))
  {
    print $self->get_man_page;
    exit;
  }

  elsif ($self->get_value('--man-nroff'))
  {
    print $self->get_man_page_nroff;
    exit;
  }

  elsif ($self->get_value('--get-config-template'))
  {
    print $self->get_config_template;
    exit;
  }

  elsif ($self->get_value('--pacman-help'))
  {
    my $cmd = $self->get_value('PacmanBin');
    $cmd .= ' --help';
    my @args = $self->{PacmanConf}->get_cli_input();
    $cmd .= ' '.join(' ',@args) if @args;
    exec $cmd;
    exit;
  }



  # Load the configuration files and reload the command-line options to preserve
  # correct overriding order.
  $self->{PacmanConf}->unset_all_values();
  $self->{PacmanConf}->load_config($self->get_value('PacmanConf'));
  $self->{PacmanConf}->reload_cli();

  $self->{PowerpillConf}->unset_all_values();
  $self->{PowerpillConf}->load_config($self->get_value('--config'));
  $self->{PowerpillConf}->reload_cli();

  $self->{MEMOIZE} = $self->get_value('Memoize');



  if ($self->get_pacvalue('-Q')
  and $self->get_pacvalue('-u'))
  {
    print $self->query_upgrades;
    exit;
  }
  elsif (not $self->get_pacvalue('-S'))
  {
    $self->switch_to_pacman();
    exit;
  }
  # Check if the database needs to be refreshed before parsing the download.
  my $ys = $self->{PacmanConf}->get_cli_opt_count('-y');
  if ($ys > 0)
  {
    if ($self->get_value('Rebase'))
    {
      $self->maybe_lock_and_rebase(($ys>1));
    }
    else
    {
      my $cmd = $self->get_value('PacmanBin') . ' -S' . ('y' x $ys);
      $cmd = '/usr/bin/sudo '.$cmd if $ENV{'USER'} ne 'root' and -x '/usr/bin/sudo';
      system $cmd;
    }
    $self->{PacmanConf}->clear_cli_opt('-y');
    $self->reset_db;
  }



  if (not $self->operation_involves_possible_download)
  {
    $self->switch_to_pacman() if not $ys;
    exit;
  }




  my $cachedir = $self->get_writable_cachedir;
  if (not $cachedir)
  {
    my $dirlist = "    " . join("\n    ", $self->get_pacvalue('CacheDir'));
    print STDERR "error: no write permissions to any of the following cache directories:\n$dirlist\n";
    exit(1);
  }
  my $sync_queue = $self->maybe_get_sync_queue_interactively();
  my @pkgnames = ();
  my $unresolved = 0;
  foreach my $pkgname (keys(%$sync_queue))
  {
    if ($sync_queue->{$pkgname}->{unresolvable})
    {
      $self->pretty_warning('unresolved dependencies: '.$sync_queue->{$pkgname}->{unresolvable}."\n");
      $unresolved ++;
    }
    else
    {
      push(@pkgnames,$pkgname);
    }
  }
  if ($unresolved
  and not (
    $self->get_pacvalue('--noconfirm')
    or $self->pretty_confirm("Do you want to ignore the above packages for this operation?",'y')
  ))
  {
    exit();
  }


  if ($self->get_value('--get-metalink')
  or $self->get_value('--save-metalink'))
  {
    $self->maybe_reflect_pkgnames(@pkgnames);
    my $metalink = $self->get_metalink_from_pkgnames({},@pkgnames);

    my $fpath = $self->get_value('--save-metalink');
    if ($fpath)
    {
      open(my $fh,'>',$fpath) or die "error: failed to open $fpath ($!)\n";
      print $fh $metalink;
      close($fh);
      exit;
    }
    else
    {
      print $metalink;
    }
    exit;
  }

  if (@pkgnames)
  {
    # Check for writable cachedir and remove existing packages from queue.
    # The user can specify an empty cache to ignore existing files.
    @pkgnames = $self->list_uncached_pkgs(@pkgnames);
  }


  @pkgnames = $self->maybe_download_from_pkgds($cachedir,@pkgnames);


  if (@pkgnames)
  {
    exit unless $self->get_pacvalue('--noconfirm') or $self->approve_download(@pkgnames);
    $self->maybe_reflect_pkgnames(@pkgnames);
    my $metalink = $self->get_metalink_from_pkgnames({},@pkgnames);

    if ($self->get_pacvalue('--noconfirm'))
    {
      my $i = 0;
      my $n =  $self->get_value('RetryLimit');
      while(@pkgnames and $i<$n)
      {
        $self->download_metalink_with_aria2($metalink,$cachedir);
        @pkgnames = $self->list_uncached_pkgs(@pkgnames);
        $i ++;
      }
    }
    else
    {
      while(@pkgnames)
      {
        $self->download_metalink_with_aria2($metalink,$cachedir);
        @pkgnames = $self->list_uncached_pkgs(@pkgnames);

        if (@pkgnames)
        {
          my $n = scalar(@pkgnames);
          last if not $self->pretty_confirm("$n unsuccessful download(s). Would you like to retry?", 'y');
        }
      }
    }
  }

  $self->remove_aria2_control_files($cachedir);
  $self->maybe_switch_to_pacman;
}













########## CACHE FILES ##########

sub get_writable_cachedir
{
  my ($self) = @_;
  my $cachedir;
  foreach my $cdir ($self->get_pacvalue('CacheDir'))
  {
    $cdir = abs_path($cdir) if -e $cdir;
    $cachedir = $cdir if not defined($cachedir) and -w $cdir;
  }
  return $cachedir;
}

sub list_uncached_pkgs
{
  my ($self,@pkgnames) = @_;
  my %filenames = ();
  $filenames{$self->get_sync_db_pkg_filename($_)} = $_ foreach @pkgnames;
  foreach my $cdir ($self->get_pacvalue('CacheDir'))
  {
    $cdir = abs_path($cdir) if -e $cdir;
    if (-r $cdir and opendir(my $dh,$cdir))
    {
      foreach my $i (readdir($dh))
      {
        next if substr($i,0,1) eq '.';
        if (exists($filenames{$i}))
        {
          my $pkgname = $filenames{$i};
          my $fpath = &join_paths($cdir,$i);
          if ($self->check_package_integrity($pkgname,$fpath))
          {
            delete($filenames{$i});
          }
          # aria2 should overwrite the file so don't both unlinking it
          #else
          #{
          #  unlink($filename);
          #}
        }
      }
      close $dh;
    }
  }
  return values(%filenames);
}

sub check_package_integrity
{
  my ($self,$pkgname,$fpath) = @_;
  my $expected = $self->get_sync_db_pkg_md5sum($pkgname);
  my $inferred = &get_file_md5($fpath);
  return (defined($expected) and defined($inferred) and $expected eq $inferred) ? $fpath : undef;
}

sub check_package_integrity_by_dir
{
  my ($self,$pkgname,$dpath) = @_;
  my $filename = $self->get_sync_db_pkg_filename($pkgname);
  my $fpath = &join_paths($dpath,$filename);
  return $self->check_package_integrity($pkgname,$fpath);
}

sub remove_aria2_control_files
{
  my ($self,$dpath) = @_;
  return if not defined($dpath);
  if (-w $dpath and opendir(my $dh,$dpath))
  {
    unlink $_ foreach grep {substr($_,-6) eq '.aria2'} readdir($dh);
    close $dh;
  }
}







########## PACMAN ##########

sub switch_to_pacman
{
  my ($self) = @_;
  my $cmd = $self->get_value('PacmanBin');
  my @args = $self->{PacmanConf}->get_cli_input();
  $cmd .= ' '.join(' ',@args) if @args;
  if ($self->operation_requires_root()
  and $ENV{'USER'} ne 'root'
  and -x '/usr/bin/sudo')
  {
    $cmd = '/usr/bin/sudo ' . $cmd;
  }
  exec $cmd;
  #print "cmd: $cmd\n";
  exit;
}



sub maybe_switch_to_pacman
{
  my ($self) = @_;
  $self->switch_to_pacman if not $self->get_pacvalue('-w');
}






########## METALINK ##########

sub get_metalink_from_pkgnames
{
  my ($self,$args,@pkgnames) = @_;
  my $metalink = Xyne::Common::Metalink -> new();
  my $split_size = $self->get_value('SplitSize');
  $split_size *= 1000000;
  my $dir = defined($args->{'dir'}) ? $args->{'dir'} : '';
  foreach my $pkgname (@pkgnames)
  {
    my $pkg = $self->get_sync_db_pkg($pkgname);
    my $csize = $pkg->{'get'}->('CSIZE')->[0];
    my $filename = $pkg->{'get'}->('FILENAME')->[0];
    $filename = &join_paths($dir,$filename);

    my $metalink_args = {
      name         => $filename,
      identity     => $pkgname,
      version      => $pkg->{'get'}->('VERSION')->[0],
      size         => $csize,
      verification => {
        md5 => $pkg->{'get'}->('MD5SUM')->[0]
      }
    };

    $metalink_args->{maxconnections} = 1 if $split_size > $csize;

    my @urls = $self->get_reflected_sync_db_pkg_urls($pkgname);
    $metalink_args->{'urls'} = [@urls];
    $metalink->add_file($metalink_args);
  }
  return $metalink->get_metalink;
}




########## REFLECTOR ##########


sub maybe_reflect
{
  my ($self,@target_reponames) = @_;
  my $args = $self->get_value('Reflect');
  if (@target_reponames and $args)
  {
    # Only reflect if it's possible to append mirrors to the repos of the
    # requested download.
    my $reflect = 0;
    my @reponames = Xyne::Arch::Reflector->get_repos();
    DR: foreach my $t (@target_reponames)
    {
      foreach my $r (@reponames)
      {
        if ($t eq $r)
        {
          $reflect = 1;
          last DR;
        }
      }
    }
    return $self->reflect if $reflect;
  }
  return 0
}



sub maybe_reflect_pkgnames
{
  my ($self,@pkgnames) = @_;
  my @download_repos = $self->get_sync_db_pkg_reponames(@pkgnames);
  return $self->maybe_reflect(@download_repos);
}



sub reflect
{
  my ($self) = @_;
  my $args = $self->get_value('Reflect');
  if ($args)
  {
    my $reflector = Xyne::Arch::Reflector->new();
    $reflector->parse_args(split(/\s+/,$args));

    $self->pretty_message("Reflecting... ");
    my @servers = $reflector->get_servers();
    my $n = scalar(@servers);
    $self->display("retrieved $n mirrors.\n");
    $self->{REFLECTOR} = {
      'servers' => \@servers,
      'time' => time()
    };
    return scalar(@servers);
  }
  return 0;
}



sub get_additional_repo_servers
{
  my ($self,$reponame) = @_;
  return () if not defined($reponame) or ref($self->{REFLECTOR}->{'servers'}) ne 'ARRAY';
  return () if not grep {$reponame eq $_} Xyne::Arch::Reflector->get_repos;
  my %existing;
  $existing{$_} = 1 foreach $self->{PacmanConf}->get_repo_servers($reponame);
  return grep {not defined($existing{$_})} @{ $self->{REFLECTOR}->{'servers'} };

}

# This is to keep pacman.conf and reflector mirrors separate so that the
# database and abs archives are only downloaded from the former.

sub get_reflected_repo_server_urls
{
  my ($self,$reponame,$arch) = @_;
  return () if not defined($reponame);
  my @urls = $self->{PacmanConf}->get_repo_servers($reponame);
  push(@urls, $self->get_additional_repo_servers($reponame));
  @urls = &replace_url_architecture($arch,@urls) if defined($arch);
  return &Xyne::Arch::Pacman::replace_url_repo($reponame,@urls);
}

sub get_reflected_repo_file_urls
{
  my ($self,$reponame,$filename,$arch) = @_;
  return () if not defined($reponame) or not defined($filename);
  return map {$_.'/'.$filename} $self->get_reflected_repo_server_urls($reponame,$arch);
}

sub get_reflected_sync_db_pkg_urls
{
  my ($self,$pkgname,$arch) = @_;
  my $reponame = $self->get_sync_db_pkg_reponame($pkgname);
  return () if not defined($reponame);
  my $filename = $self->get_sync_db_pkg($pkgname)->{'get'}->('FILENAME')->[0];
  return () if not defined($filename);
  return $self->get_reflected_repo_file_urls($reponame,$filename,$arch);
}






########## ARCHIVES ##########

sub get_reflected_repo_archive
{
  my ($self,$reponame,$archive,$arch) = @_;
  return $self->{ARCHIVES}->{$reponame}->{$archive} if defined($self->{ARCHIVES}->{$reponame}->{$archive});
  return undef if not defined($reponame) or not defined($archive);
  my $filename = $reponame.'.'.$archive.$Xyne::Arch::Pacman::ARCHIVE_EXT;
  my $a = Xyne::Common::RemoteFile->new({
    'filename' => $filename,
    'urls' => [$self->get_reflected_repo_file_urls($reponame,$filename,$arch)]
  });
  $self->{ARCHIVES} = {} if not defined($self->{ARCHIVES});
  $self->{ARCHIVES}->{$reponame} = {} if not defined($self->{ARCHIVES}->{$reponame});
  $self->{ARCHIVES}->{$reponame}->{$archive} = $a;
  return $a;
}

sub get_repo_dbs_content
{
  my ($self,$reponames,$arch) = @_;
  my @reponames = @$reponames;
  my $s = sub {
    my ($self,$reponame) = @_;
    my $a = $self->get_repo_archive($reponame,'db',$arch);
    return [$reponame,$a->get_content]
  };
  my @args = map {[$self,$_]} @reponames;
  return &parallelize($s,@args);
}




########## REBASE ##########

sub rebase
{
  my ($self,$args,$reponames,$arch) = @_;
  $self->pretty_message("Rebasing...\n");
  my @reponames = (ref($reponames) eq 'ARRAY') ? @$reponames : ();
  $args = {} if not ref($args) eq 'HASH';
  $args->{'purge'} = 1;
  $args->{'umask'} = 022;
  @reponames = $self->get_repo_list if not @reponames;
  my %dbs;
  foreach my $rv ($self->get_repo_dbs_content(\@reponames,$arch))
  {
    $dbs{$rv->[0]} = $rv->[1] if defined($rv->[0]) and defined($rv->[1]);
  }

  my $results = {};
  foreach my $reponame (@reponames)
  {
    if (defined($dbs{$reponame}))
    {
      my $dpath = $self->get_db_path('sync',$reponame);
      $results->{$reponame} = &extract_archive($dbs{$reponame},$dpath,$args);
      $self->set_sync_db_repo_lastupdate($reponame);
    }
  }
  return $results;
}

sub maybe_rebase
{
  my ($self,$args,$reponames,$arch) = @_;
  my @reponames = (ref($reponames) eq 'ARRAY') ? @$reponames : ();
  @reponames = $self->get_repo_list if not @reponames;
  my @old = ();
  foreach my $reponame (@reponames)
  {
    my $rt = $self->get_repo_db_remote_mtime($reponame,$arch);
    my $lt = $self->get_sync_db_repo_lastupdate($reponame);
    push(@old,$reponame) if $rt and $rt > $lt;
  }
  return $self->rebase($args,\@old,$arch) if @old;
  return {};
}

sub maybe_lock_and_rebase
{
  my ($self,$definitely,$reponames,$arch) = @_;
  $self->lock_db;
  my $r;
  if ($definitely)
  {
    $r = $self->rebase({'keep'=>1},$reponames,$arch);
  }
  else
  {
    $r = $self->maybe_rebase({'keep'=>1},$reponames,$arch);
  }
  $self->display_rebase_results($r);
  $self->unlock_db;
}



########## PKGD ##########

sub maybe_download_from_pkgds
{
  my ($self,$cachedir,@pkgnames) = @_;
  return () if not @pkgnames;
  my @pkgds = $self->get_value('PkgD');
  if (@pkgds)
  {
    $self->pretty_message("checking PkgDs...\n");
    my @filenames = map {$self->get_sync_db_pkg_filename($_)} @pkgnames;
    &Xyne::Arch::PkgD::get_files_from_pkgds($cachedir,\@filenames,\@pkgds);
    my @remaining = ();
    foreach my $pkgname (@pkgnames)
    {
      my $fpath = $self->check_package_integrity_by_dir($pkgname,$cachedir);
      if ($fpath)
      {
        $self->pretty_message("successfully downloaded $fpath\n");
        $self->maybe_log("successfully downloaded $fpath\n");
      }
      else
      {
        push (@remaining,$pkgname);
      }
    }
    @pkgnames = @remaining;
  }
  return @pkgnames;
}



########## ARIA2 ##########


sub download_metalink_with_aria2
{
  my ($self,$metalink,$dir) = @_;
  my $aria2bin = $self->get_value('Aria2Bin');
  if (not defined($aria2bin) or not -x $aria2bin)
  {
    print STDERR "error: $aria2bin is not an executable path\n";
    exit(1);
  }
  my $aria2args = $self->{PowerpillConf}->get_aria2_arg_str();
  $aria2args .= "--dir=$dir" if defined($dir);

  #my $cmd = "$aria2bin $aria2args <<'METALINK'\n$metalink\nMETALINK\n";
  my $cmd = "$aria2bin $aria2args";

  # I'm using a 2-way pipe to /bin/bash here because Perl complains about argument
  # length if I try to open a pipe directly with the command above (fair enough,
  # the metalinks can be huge).
  my ($pin,$pout,$pid);
  #if ($pid=open2($pout,$pin,'/bin/bash'))
  if ($pid=open2($pout,$pin,$cmd))
  {
    local $| = 1;

    #print $pin $cmd;
    print $pin $metalink;
    close $pin;

    $self->{FILTER_ARIA2} = 1;
    $self->display("\n");
    while(defined(my $line = <$pout>))
    {
      chomp $line;
      $self->filter_aria2_output($line);
    }
    close $pout;
    $self->display("\n");
  }
  else
  {
    print STDERR "error: failed to pipe metalink to aria2c: $!\n";
    exit(1);
  }
}

sub filter_aria2_output
{
  my ($self,$line) = @_;
  if ($line =~ m/Download Results:/)
  {
    $self->{FILTER_ARIA2} = 0;
    $line = "\r\n$line\n";
  }
#   elsif ($line =~ m/NOTICE - Verification finished successfully. file=(.*)$/)
#   {
#     my $fpath = $1;
#     $line = "\r".$self->maybe_color('VERIFIED','bold green').' '.$fpath."\n";
#     $self->maybe_log("verified $fpath\n");
#   }
  elsif ($line =~ m/NOTICE - Download complete: (.*)$/)
  {
    my $fpath = $1;
    my $prefix = $self->maybe_color('-->','bold blue');
    $fpath = basename($fpath);
    $line = "\r$prefix successfully downloaded $fpath\n";
    $self->maybe_log("successfully downloaded $fpath\n");
  }
  elsif (substr($line,0,1) eq '[')
  {
    $line = "\r$line";
  }
  elsif ($self->{FILTER_ARIA2})
  {
    return;
  }
  else
  {
    $line .= "\n";
  }
  # clear previous characters
  $line = "\r" . (' ' x &get_term_width) . $line if substr($line,0,1) eq "\r";
  $self->display($line) if defined($line);
}


########## MESSAGES ##########

sub display
{
  my ($self,@msgs) = @_;
  return if $self->get_value('Silent');
  print $_ foreach @msgs;
}

sub pretty_message
{
  my ($self,@msgs) = @_;
  my $prefix = $self->maybe_color('-->','bold blue');
  $self->display(map {$prefix.' '.$_} @msgs);
}

sub pretty_warning
{
  my ($self,@msgs) = @_;
  my $pre = $self->maybe_color('WARNING:', 'bold red');
  $self->display( map {$pre . ' ' . $_} @msgs );
}

sub pretty_error
{
  my ($self,@msgs) = @_;
  my $pre = $self->maybe_color('ERROR:', 'bold red');
  $self->display( map {$pre . ' ' . $_} @msgs );
}

sub maybe_color
{
  my ($self,$str,$color) = @_;
  return $str if $self->get_value('NoColor');
  return &colorize($str,$color);
}

sub pretty_confirm
{
  my ($self,$question,$default) = @_;
  my $marker = $self->maybe_color('::','bold blue');
  return &confirm($marker.' '.$question,$default);
}

sub maybe_log
{
  my ($self,@msgs) = @_;
  my $lpath = $self->get_value('Log');
  if (defined($lpath))
  {
    $self->{LOG} = Xyne::Common::Log->new($lpath) if not $self->{LOG};
    $self->{LOG}->log(@msgs);
  }
}

sub get_reponame_str
{
  my ($self,$reponame) = @_;
  return $self->maybe_color('[','bold blue').$reponame.$self->maybe_color(']','bold blue');
}

sub get_pkgsize_str
{
  my ($self,$size) = @_;
  $size = defined($size) ? &format_bytes($size,'MiB') : 'N/A';
  return $self->maybe_color(" [$size]",'bold black')
}



sub select_group_packages
{
  my ($self,$group,@pkgnames) = @_;
  my @selected = ();
  my $w = &get_term_width();
  my $marker = $self->maybe_color('::','bold blue');
  print "$marker group $group (including ignored packages):\n";
  my $groups = $self->group_pkgnames_by_repo(@pkgnames);
  foreach my $reponame ($self->get_repo_list)
  {
    if (ref($groups->{$reponame}) eq 'ARRAY')
    {
      print '   '.$self->get_reponame_str($reponame)."\n";
      print &wrap_words($w,'     ',@{$groups->{$reponame}});
    }
  }
  if ($self->pretty_confirm('Install whole content?','y'))
  {
    return @pkgnames;
  }
  else
  {
    foreach my $pkgname (@pkgnames)
    {
      push(@selected,$pkgname) if $self->pretty_confirm("Install $pkgname from group $group?",'y');
    }
  }
  return @selected;
}

sub approve_download
{
  my ($self,@pkgnames) = @_;
  my $w = &get_term_width();
  my $marker = $self->maybe_color('::','bold blue');
  my $csize = 0;
  my $isize = 0;
  print "\n$marker The following packages will be downloaded:\n";
  my $groups = $self->group_pkgnames_by_repo(@pkgnames);
  foreach my $reponame ($self->get_repo_list)
  {
    if (ref($groups->{$reponame}) eq 'ARRAY')
    {
      print '   '.$self->get_reponame_str($reponame)."\n";
      my @ps = ();
      foreach my $p (@{ $groups->{$reponame} })
      {
        my $size = $self->get_sync_db_pkg_csize($p);
        $csize += $size;
        $isize += $self->get_sync_db_pkg_isize($p);
        $p .= '-' . $self->get_sync_db_pkg_ver($p);
        if ($self->get_pacvalue('ShowSize'))
        {
          $p .= $self->get_pkgsize_str($size);
        }
        push(@ps,$p);
      }
      print &wrap_words($w,'     ',@ps);
    }
  }
  my $n = scalar(@pkgnames);

  $csize = &format_bytes($csize,'MiB');
  $isize = &format_bytes($isize,'MiB');

  print "\n";
  printf "$marker Number of Packages:   %13d\n", $n;
  printf "$marker Total Download Size:  %13s\n", $csize;
  printf "$marker Total Installed Size: %13s\n", $isize;
  print "\n";


  return $self->pretty_confirm('Proceed with download?','y');
}

sub display_rebase_results
{
  my ($self,$results) = @_;
  return if ref($results) ne 'HASH';
  #my $w = &get_term_width();
  foreach my $reponame ($self->get_repo_list)
  {
    next if ref($results->{$reponame}) ne 'HASH';
    my $repodir = $self->get_db_path('sync',$reponame);
    my $l = length($repodir) + 1;
    my $pkgs = {};
    my $reponame_str = $self->get_reponame_str($reponame);
    foreach my $fpath (keys(%{ $results->{$reponame} }))
    {
      next if not defined($results->{$reponame}->{$fpath});
      my $pkgdir = substr($fpath,$l);
      if (index($pkgdir,'/') < 0)
      {
        my ($pkgname,$pkgver) = &get_name_and_version($pkgdir);
        next if not defined($pkgver);
        my $n = $results->{$reponame}->{$fpath};
        if ($n == 1)
        {
          $pkgs->{$pkgname}={} if not defined($pkgs->{$pkgname});
          $pkgs->{$pkgname}->{'new'} = $pkgver;
        }
        elsif ($n == -2)
        {
          $pkgs->{$pkgname}={} if not defined($pkgs->{$pkgname});
          $pkgs->{$pkgname}->{'old'} = $pkgver;
        }
      }
    }
    my @pkgnames = keys(%{ $pkgs });
    $self->display("\n") if (@pkgnames);
    my $p = 0;
    my $u = 0;
    my $m = 0;
    my $p_sign = $self->maybe_color('+','bold green');
    my $u_sign = $self->maybe_color('^','bold cyan');
    my $m_sign = $self->maybe_color('-','bold red');
    foreach my $pkgname (sort @pkgnames)
    {
      my $new_ver = $pkgs->{$pkgname}->{'new'};
      my $old_ver = $pkgs->{$pkgname}->{'old'};
      my $pre = '';
      my $post = '';
      if (defined($new_ver) and defined($old_ver))
      {
        $u++;
        $pre = $u_sign;
        $post = "$old_ver -> $new_ver";
      }
      elsif (defined($new_ver))
      {
        $p++;
        $pre = $p_sign;
        $post = $new_ver;
      }
      elsif (defined($old_ver))
      {
        $m++;
        $pre = $m_sign;
        $post = $old_ver;
      }
      # 6: 1 for pre, 3 spaces, "[" and "]"
      #my $x = 6 + length($pkgname) + length($post) + length($reponame);
      #my $r = $w - $x;
      #my $spacer = ($r > 0) ? ' ' x $r : '';
      $post = $self->maybe_color($post,'bold black');
      #$self->display("$pre $pkgname $post ${spacer}$reponame_str\n");
      $self->display("$pre $reponame/$pkgname $post\n");
    }
    $self->pretty_message("updated $reponame_str $p_sign$p $u_sign$u $m_sign$m\n");
    $self->maybe_log("updated [$reponame] +$p ^$u -$m\n");
  }
}







########## CUSTOMIZED PACMAN OUTPUT ##########

sub query_upgrades
{
  my ($self,$clargs) = @_;
  my $output = '';

  my @pkgnames = @$clargs if defined($clargs);
  @pkgnames = $self->get_local_db_pkgname_list() if not @pkgnames;
  my $groups = $self->group_pkgnames_by_repo(@pkgnames);

  foreach my $reponame ($self->get_repo_list)
  {
    my $not_empty = 0;
    if (ref($groups->{$reponame}) eq 'ARRAY')
    {
      foreach my $pkgname (sort @{$groups->{$reponame}})
      {
        my $local_ver = $self->get_local_db_pkg_ver($pkgname);
        my $sync_ver = $self->get_sync_db_pkg_ver($pkgname);
        my $cmpver = &compare_versions($sync_ver,$local_ver);
        if ($cmpver > 0 or ($cmpver != 0 and $self->is_forced($pkgname)))
        {
          $not_empty = 1;
          $output .= $reponame.'/'.$pkgname;
          #$output .= $self->maybe_color($reponame,'bold green').'/'.$pkgname;
          if (not $self->get_pacvalue('-q'))
          {
            #$output .= ' ' . $self->maybe_color("$local_ver -> $sync_ver", 'bold green');
            $output .= " $local_ver -> $sync_ver";
          }
          $output .= "\n";
        }
      }
      $output .= "\n" if $not_empty;
    }
  }
  return $output;
}




1;

__END__

=head1 NAME

Xyne::Arch::Powerpill - A pacman wrapper and download-accelerator class.

=head1 SYNOPSIS

  use Xyne::Arch::Powerpill;

=head1 DESCRIPTION

Powerpill is intended to be a fully-functional wrapper for pacman
that uses aria2 to accelerate downloads while leaving all other
operations to pacman itself.

=head2 EXPORT

None by default.

=head1 AUTHOR

Xyne <lt>xyne at archlinux dot ca<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009,2010  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.