# Copyright (C) 2009,2010  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


package Xyne::Arch::PowerpillConf;
use strict;
use warnings;

use Xyne::Common::OptionHandler;
use Xyne::Common::Functions qw/get_config_dir strip_trailing_slash/;

use Xyne::Arch::PacmanConf;

our @ISA = ("Xyne::Common::OptionHandler");

our $VERSION = '0.04';

our $MAX_SERVERS = 45;

our @ARIA2_DEFAULT_ARGS = (
  'allow-overwrite=true',
  'continue',
  'file-allocation=none',
  'ftp-pasv=true',
  "max-concurrent-downloads=$MAX_SERVERS",
  'max-tries=2',
  'metalink-enable-unique-protocol=false',
  "metalink-servers=$MAX_SERVERS",
  'split=10',
  'summary-interval=0',
  'timeout=5',
  'file-allocation=none'
);

our @ARIA2_RESERVED_ARGS = (
  ['metalink-file','-'],
  ['auto-file-renaming','false'],
  ['dir','%DIR%']
);

our @OPTIONS = (
  {
    'name' => 'Aria2Args',
    'aliases' => ['--aria2-args'],
    'type' => 'list',
    'description' => "Additional arguments to aria2c. These will be appended to the internal list of default arguments, replacing some of them when appropriate. The format used is the same as in aria2c configuration files, e.g. \"lowest-speed-limit=10K\"\nOptions of immediate interest are \"lowest-speed-limit\", which can be used to filter slow mirrors while the download is in progress, and \"max-download-limit\", to limit download speed.\nOptions which are required by Powerpill will be ignored. Beyond that, Powerpill does not check the sanity of the arguments given.\nConsult aria2c's documentation for a full list of available options and information about what they do.",
    'default' => \@ARIA2_DEFAULT_ARGS
  },

  {
    'name' => 'Aria2Bin',
    'aliases' => ['--aria2-bin'],
    'type' => 'str',
    'default' => '/usr/bin/aria2c',
    'description' => 'The path to the aria2c binary.',
    'parameter_name' => '<path>'
  },

  {
    'name' => 'Log',
    'aliases' => ['--log'],
    'type' => 'str',
    'recommend' => '/var/log/powerpill.log',
    'description' => 'Powerpill output will be logged to this path if it is given.',
    'parameter_name' => '<path>'
  },

  {
    'name' => 'Memoize',
    'aliases' => ['--memoize'],
    'type' => 'switch',
    'description' => 'Cache database file lookups in memory to enable faster subsequent lookups. This will greatly speed up longer operations in most cases at the cost of a larger memory footprint.',
  },

  {
    'name' => 'NoColor',
    'aliases' => ['--no-color'],
    'type' => 'switch',
    'description' => 'Suppress colored output.'
  },

  {
    'name' => 'PacmanBin',
    'aliases' => ['--pacman-bin'],
    'type' => 'str',
    'default' => '/usr/bin/pacman',
    'description' => 'The path to the pacman binary or an equivalent, such as pacman-color.',
    'parameter_name' => '<path>'
  },

  {
    'name' => 'PacmanConfig',
    'aliases' => ['--pacman-config'],
    'type' => 'str',
    'description' => 'Specify a different pacman configuration file to use.',
    'parameter_name' => '<path>',
    'default' => '/etc/pacman.conf'
  },

  {
    'name' => 'PkgD',
    'aliases' => ['--pkgd'],
    'type' => 'list',
    'description' => 'Specify PkgD servers to query for packages before using external mirrors.',
    'parameter_name' => '[address:port]'
  },

  {
    'name' => 'Rebase',
    'aliases' => ['--rebase'],
    'type' => 'switch',
    'description' => 'Use the Rebase module to update the sync database. See the Rebase documentation for more information.',
  },

  {
    'name' => 'Reflect',
    'aliases' => ['--reflect'],
    'type' => 'str',
    'description' => "Enable retrieval of mirrors from the MirrorStatus webpage during a sync operation. This uses the Reflector module internally and accepts the same arguments that Reflector does. See the Reflector documentation for more information.\nThere is absolutely no reason to pass any of the ranking options with this command as the order of the servers is completely irrelevant.",
    'default' => "-l $MAX_SERVERS",
    'parameter_name' => '"[Reflector arguments]"'
  },

  {
    'name' => 'RetryLimit',
    'aliases' => ['--retry-limit'],
    'type' => 'int',
    'default' => 3,
    'description' => 'The number of times that a failed download should be retried when run automatically.'
  },

  {
    'name' => 'SplitSize',
    'aliases' => ['--split-size'],
    'type' => 'int',
    'default' => 5,
    'description' => 'The minimum size, in MB (not MiB), that a file must be before it is split for segmented downloading.'
  },










  {
    'name' => '--config',
    'type' => 'str',
    'description' => 'Specify a different configuration file to use.',
    'parameter_name' => '<path>'
  },

  {
    'name' => '--get-config-template',
    'type' => 'switch',
    'description' => 'Print powerpill\'s configuration file template.'
  },

  {
    'name' => '--get-metalink',
    'type' => 'switch',
    'description' => 'Output a metalink for the requested operation and exit.'
  },

  {
    'name' => '--help',
    'aliases' => ['-h'],
    'type' => 'switch',
    'description' => 'Print powerpill\'s help message and exit.'
  },

  {
    'name' => '--man',
    'type' => 'switch',
    'description' => 'Print powerpill\'s man page and exit.'
  },

  {
    'name' => '--man-nroff',
    'type' => 'switch',
    'description' => 'Print powerpill\'s man page in nroff format.'
  },

  {
    'name' => '--pacman-help',
    'type' => 'switch',
    'description' => 'Print pacman\'s help message and exit.'
  },

  {
    'name' => '--save-metalink',
    'type' => 'str',
    'parameter_name' => '<path>',
    'description' => 'Save the metalink to a file.'
  }
);


our $MANPAGE = {
  'header' => 'POWERPILL 1 "January 2010"',
  'sections' => [
    ['NAME', 'powerpill - a wrapper and download accelerator for pacman'],
    ['SYNOPSIS', 'powerpill [options] [packages]'],
    ['DESCRIPTION', "Powerpill is a wrapper for the pacman package manager used by Arch Linux and other GNU/Linux distributions. It determines the target packages of a requested synchronization operation and then uses the mirrorlist to create a comprehensive metalink. This metalink is then piped to the download manager aria2 for package retrieval. Significant reductions in download times are often possible due to the combined effects of simultaneous and segmented downloads.\n\nPowerpill also internally supports database updates via Rebase, package retrieval via PkgD, and mirrorlist retrieval via Reflector."],
    ['OPTIONS', "Powerpill supports the following in addition to Pacman's arguments:\n"],
    ['AUTHORS', 'Xyne']
  ]
};



our $CONFIG_TEMPLATE = {
#  'header' => "Each section begins with a name in square brackets, e.g. \"[options]\". There are currently only two reserved sections, namely [options] and [PkgD]. Any other section will be interpreted as a repo and servers specified within that section will be appended to the server list for that repo.",
  'sections' => [
    ['options', ""],
#    ['PkgD', "This section can be used to specify PkgD servers that should be queried before regular servers."]
  ]
};



our $HELP_MESSAGE = {
  'header' => "USAGE\n\n    powerpill [powerpill options] [pacman options] [args]\n\n    See \"--pacman-help\" for help with Pacman's options.\n\n\nPOWERPILL OPTIONS\n\n",
  'footer' => "",
};




our @DEFAULT_PATHS = (
  &get_config_dir().'/powerpill/powerpill.conf',
  '/etc/powerpill.conf'
);







########## CONSTRUCTOR ##########

sub new {
  my ($proto,$args) = @_;
  my $class = ref($proto) || $proto;
  my $self  = {};
  bless ($self, $class);

  $self->declare_options(@OPTIONS);
  $self->{CLI} = {};
  $self->{MANPAGE} = $MANPAGE;
  $self->{CONFIG_TEMPLATE} = $CONFIG_TEMPLATE;
  $self->{HELP_MESSAGE} = $HELP_MESSAGE;

  if (defined($args->{CALLBACKS}))
    { $self->{CALLBACKS} = $args->{CALLBACKS}; }
  else
    { $self->{CALLBACKS} = $Xyne::Arch::OptionHandler::CALLBACKS; }


  if (defined($args->{PacmanConf}))
    { $self->{PacmanConf} = $args->{PacmanConf}; }
  else
    { $self->{PacmanConf} = Xyne::Arch::PacmanConf->new(); }

  $self->{PacmanConf}->load_config($args->{'pacman_config_path'}) if $args->{'pacman_autoload'};
  $self->load_config($args->{'config_path'}) if $args->{'autoload'};

  return $self;
}




sub load_config
{
  my ($self,$fpath) = @_;
  $fpath = $self->get_value('--config') if not defined($fpath);
  if (not defined($fpath))
  {
    foreach my $p (@DEFAULT_PATHS)
    {
      if (-r $p)
      {
        $fpath = $p;
        last;
      }

    }
  }
  return &Xyne::Common::OptionHandler::load_config($self,$fpath);
}




########## CLI ##########
sub parse_cli_input
{
  my ($self,@args) = @_;
  my ($clopts,$popts,$clargs);
  ($clopts,$clargs) = &Xyne::Common::OptionHandler::parse_cli_input($self,@args);
  ($popts,$clargs) = $self->{PacmanConf}->parse_cli_input(@$clargs);
  return ($clopts,$popts,$clargs);
}


# sub get_repo_servers
# {
#   my ($self,$repo) = @_;
#   return () if not defined($repo);
#   return grep {$_} map {$1 if $_=~m/^\s*Server\s*=\s*(\S+)/} ($self->{PacmanConf}->get_config_section($repo),$self->get_config_section($repo));
# }



sub get_aria2_arg_str
{
  my ($self,$dir) = @_;
  my $args = '';
  ARG: foreach my $arg ($self->get_value('Aria2Args'))
  {
    foreach my $reserved (@ARIA2_RESERVED_ARGS)
    {
      my $n = $reserved->[0];
      next ARG if $arg =~ m/(?:^|\s)\Q$n\E(?:=|\s|$)/;
    }
    $args .= "--$arg ";
  }
  foreach my $r (@ARIA2_RESERVED_ARGS)
  {
    my ($n,$v) = @$r;
    if ($n eq 'dir')
    {
      next if not defined($dir);
      $args .= "--dir=$dir ";
    }
    else
    {
      $args .= "--$n=$v ";
    }
  }
  return $args;
}



1;

__END__

=head1 NAME

Xyne::Arch::PowerpillConf - Powerpill configuration file class.

=head1 SYNOPSIS

  use Xyne::Arch::PowerpillConf;

=head1 DESCRIPTION

A class to load and parse powerpill configuration files and
store powerpill settings, including arguments passed to it from
powerpill objects.

=head2 EXPORT

None by default.

=head1 AUTHOR

Xyne <lt>xyne at archlinux dot ca<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009,2010  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

=cut
