# Copyright (C) 2009,2010 Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


package Xyne::Arch::Rebase;
use strict;
use warnings;

use Cwd;
use Xyne::Common::Functions qw/join_paths/;
use Xyne::Common::OptionHandler;
use Xyne::Common::Threads qw/parallelize/;
use Xyne::Arch::Powerpill;

our $VERSION = '0.23';

our @ISA = qw/Xyne::Arch::Powerpill/;

our %OPTION_MAP = (
  'NoColor' => '--no-color',
  'Log' => '--log',
  'Silent' => '--silent'
);

our @OPTIONS = (
  {
    'name' => '--arch',
    'aliases' => ['-a'],
    'type' => 'str',
    'description' => 'Retrieve databases for this architecture.',
    'parameter_name' => '<architecture>'
  },
  {
    'name' => '--by-lastupdate',
    'type' => 'switch',
    'description' => 'Only download databases which appear to be newer than the time reported in the ".lastupdate" file.'
  },

  {
    'name' => '--dbpath',
    'aliases' => ['-b'],
    'type' => 'str',
    'description' => 'Change the database path.',
    'parameter_name' => '<path>'
  },

  {
    'name' => '--dir',
    'aliases' => ['-d'],
    'type' => 'str',
    'description' => 'Change the output directory. This applies to the "--get-archives" option.',
    'parameter_name' => '<path>'
  },

  {
    'name' => '--get-archives',
    'type' => 'switch',
    'description' => 'Download the database archive files.'
  },

  {
    'name' => '--help',
    'aliases' => ['-h'],
    'type' => 'switch',
    'description' => 'Display the help message and exit.'
  },

  {
    'name' => '--log',
    'type' => 'str',
    'description' => 'Enable logging.',
    'parameter_name' => '<path>',
    'recommend' => '/var/log/rebase.log'
  },

  {
    'name' => '--man',
    'type' => 'switch',
    'description' => 'Display the man page and exit.'
  },

  {
    'name' => '--man-nroff',
    'type' => 'switch',
    'description' => 'Display the man page in nroff format and exit.'
  },

  {
    'name' => '--no-color',
    'type' => 'switch',
    'description' => 'Suppress colored output.'
  },

  {
    'name' => '--overwrite',
    'type' => 'switch',
    'description' => 'Overwrite existing database files. This is what Pacman does by default.'
  },

  {
    'name' => '--pacman-config',
    'type' => 'str',
    'description' => 'Use a different pacman configuration file.',
    'parameter_name' => '<path>'
  },

  {
    'name' => '--silent',
    'type' => 'switch',
    'description' => 'Suppress all output.'
  }
);


my $DESCRIPTION = <<DESC;
Rebase is a pacman database synchronization tool. It is able to retrieve the database archives for repos listed in a pacman configuration file and selectively extract the files. Rebase uses parallel downloads to speed up retrieval and pipes to keep everything in memory during a sync operation. It only writes files to the disk when necessary. This is in contrast to pacman which deletes and recreates all files in each database with every update, leading to unnecessary IO operations.

The output during a sync operation will show you exactly which files have been added and will alert you about any errors such as failing to retrieve a database from a given URL. By default, rebase parses /etc/pacman.conf to get a list of servers for each repo and it respects the order just as pacman does.

Rebase is also able to download databases for other architectures which may be useful when installing a system remotely or when configuring a chroot environment.

Finally, rebase only updates the sync databases. It never touches the local database so it should be completely safe to use. Even if something were to happen to your sync database, rebase should be able to fix it. In the worst case scenario, rebase has a "--reset" option will will use pacman to force a database refresh, overwriting all files with a fresh database.
DESC

our $MANPAGE = {
  'header' => 'REBASE 1 "January 2010"',
  'sections' => [
    ['NAME', 'rebase - a pacman database synchronization tool'],
    ['SYNOPSIS', 'rebase [options] [repos]'],
    ['DESCRIPTION', $DESCRIPTION ],
    ['USAGE', "rebase [options] [repos]\n\nIf no repos are specified, all repos are used."],
    ['OPTIONS', ""],
    ['AUTHORS', 'Xyne']
  ]
};


our $HELP_MESSAGE = {
  'header' => "USAGE\n\n    rebase [options] [repos]\n\n\nOPTIONS\n\n",
  'footer' => "",
};


sub new {
  my ($proto,$args) = @_;
  my $class = ref($proto) || $proto;
  my $self  = {};
  bless ($self, $class);

  $self->{CONF} = Xyne::Common::OptionHandler->new();
  $self->{CONF}->declare_options(@OPTIONS);
  $self->{CONF}->{MANPAGE} = $MANPAGE;
  $self->{CONF}->{HELP_MESSAGE} = $HELP_MESSAGE;

  if (defined($args->{CALLBACKS}))
    { $self->{CALLBACKS} = $args->{CALLBACKS}; }
  else
    { $self->{CALLBACKS} = $Xyne::Arch::OptionHandler::CALLBACKS; }

  if (defined($args->{PacmanConf}))
    { $self->{PacmanConf} = $args->{PacmanConf}; }
  else
    { $self->{PacmanConf} = Xyne::Arch::PacmanConf->new(); }
  $self->{PacmanConf}->load_config($args->{'pacman_config_path'}) if $args->{'pacman_autoload'};

  return $self;
}






sub get_value
{
  my ($self,$option) = @_;
  $option = $OPTION_MAP{$option} if defined($OPTION_MAP{$option});
  return $self->{CONF}->get_value($option);
}


sub main
{
  my ($self,@args) = @_;
  my ($clopts,$repos) = $self->{CONF}->parse_cli_input(@args);

  if ($self->get_value('--help'))
  {
    print $self->{CONF}->get_cli_help_message;
    exit;
  }

  elsif ($self->get_value('--man'))
  {
    print $self->{CONF}->get_cli_simple_man_page;
    exit;
  }

  elsif ($self->get_value('--man-nroff'))
  {
    print $self->{CONF}->get_cli_nroff_man_page;
    exit;
  }

  $self->{PacmanConf}->load_config($self->{CONF}->get_value('--pacman-config'));
  my $dbpath = $self->get_value('--dbpath');
  $self->{PacmanConf}->set_value('DBPath',$dbpath) if defined($dbpath);
  my $arch = $self->get_value('--arch');

  my @repos = @$repos;
  @repos = $self->get_repo_list if not @repos;

  if ($self->get_value('--get-archives'))
  {
    my $dir = $self->get_value('--dir');
    $dir = getcwd if not $dir;

    my $s = sub {
      my ($self,$reponame) = @_;
      my $a = $self->get_repo_archive($reponame,'db',$arch);
      my $fpath = &join_paths($dir,$a->{'filename'});
      return $a->store_content($fpath);
    };
    my @args = map {[$self,$_]} @repos;
    &parallelize($s,@args);
  }
  else
  {
    my $by_lastupdate = $self->get_value('--by-lastupdate');
    $self->maybe_lock_and_rebase((not $by_lastupdate),$repos,$arch);
  }
}





1;

__END__

=head1 NAME

Xyne::Arch::Rebase - A class to retrieve and update pacman package databases.

=head1 SYNOPSIS

  use Xyne::Arch::Rebase;

=head1 DESCRIPTION

This class is capable of retrieving archived pacman package databases from
mirror servers and extracting them to update the pacman sync database.
Extraction may be partial and limited to only those files which have changed.

=head2 EXPORT

None by default.

=head1 AUTHOR

Xyne <lt>xyne at archlinux dot ca<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009,2010  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.