# Copyright (C) 2009,2010  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


package Xyne::Arch::Reflector;
use strict;
use warnings;

use LWP::Simple qw/get/;
use POSIX qw/uname/;
use Time::Local;
use Xyne::Common::Functions qw/strip_trailing_slash/;
use Xyne::Common::OptionHandler;

our $VERSION = '0.21';

my @REPOS = qw/core extra community testing community-testing/;
my $MIRRORSTATUS_URL = 'https://www.archlinux.de/?page=MirrorStatusReflector';
#my $MIRRORCHECK_URL = 'https://users.archlinux.de/~gerbra/mirrorcheck.html';


our @OPTIONS = (
  {
    'name' => '--all',
    'aliases' => ['-a'],
    'type' => 'switch',
    'description' => 'Retrieve all mirrors. This is the default behavior and it is overridden by other options.'
  },

  {
    'name' => '--arch',
    'type' => 'str',
    'parameter_name' => '<arch>',
    'description' => 'Retrieve mirrors for a different architecture. By default the architecture of the host system is used.'
  },

  {
    'name' => '--fastest',
    'aliases' => ['-f'],
    'type' => 'int',
    'parameter_name' => 'n',
    'description' => "Retrieve the n fastest mirrors.\nThis necessarily implies \"--rank\" and the caveats that go with it."
  },

  {
    'name' => '--grep',
    'type' => 'str',
    'parameter_name' => '<regex>',
    'description' => 'Retrieve mirrors which match <regex>.'
  },

  {
    'name' => '--hours',
    'aliases' => ['-h'],
    'type' => 'int',
    'parameter_name' => 'n',
    'description' => 'Retrieve all mirrors which have synchronized within the last n hours.'
  },

  {
    'name' => '--help',
    'type' => 'switch',
    'description' => 'Display the help message and exit.'
  },

  {
    'name' => '--latest',
    'aliases' => ['-l'],
    'type' => 'int',
    'parameter_name' => 'n',
    'description' => 'Retrieve the n most up-to-date mirrors.'
  },

  {
    'name' => '--man',
    'type' => 'switch',
    'description' => 'Display the man page and exit.'
  },

  {
    'name' => '--man-nroff',
    'type' => 'switch',
    'description' => 'Display the man page in nroff format and exit.'
  },

  {
    'name' => '--output',
    'aliases' => ['-o'],
    'type' => 'str',
    'parameter_name' => '<path>',
    'description' => 'Save the mirrorlist to the given path.'
  },

  {
    'name' => '--rank',
    'aliases' => ['-r'],
    'type' => 'switch',
    'description' => 'Rank the returned mirrorlist using rankmirrors. This option should not be used on its own as it will take ages to rank the full unfiltered mirrorlist and there is no reason to do that.'
  }
);


our $MANPAGE = {
  'header' => 'REFLECTOR 1 "January 2010"',
  'sections' => [
    ['NAME', 'reflector - retrieve and filter a list of current Arch Linux package repository servers'],
    ['SYNOPSIS', 'reflector [options]'],
    ['DESCRIPTION', "Reflector retrieves a list of Arch Linux package repository servers from $MIRRORSTATUS_URL and filters them according to user input. See the options section for more information about available filters."],
    ['OPTIONS', "Reflector recognized the following options:\n"],
    ['AUTHORS', 'Xyne']
  ]
};


sub new {
  my $proto = shift;
  my $class = ref($proto) || $proto;
  my $self  = {};
  bless ($self, $class);
  $self->{ARCH} = [uname()]->[4];
  $self->{CONF} = Xyne::Common::OptionHandler->new();
  $self->{CONF}->declare_options(@OPTIONS);
  $self->{CONF}->{HELP_MESSAGE}->{'header'} = "USAGE\n\n    rankmirrors [options]\n\n\nOPTIONS\n\n";
  $self->{CONF}->{MANPAGE} = $MANPAGE;
  return $self;
}





sub main
{
  my ($self,@args) = @_;
  $self->parse_args(@args);
  my $conf = $self->{CONF};

  if ($conf->get_value('--help'))
  {
    print $conf->get_cli_help_message();
    exit;
  }

  elsif ($conf->get_value('--man'))
  {
    print $conf->get_cli_simple_man_page();
    exit;
  }

  elsif ($conf->get_value('--man-nroff'))
  {
    print $conf->get_cli_nroff_man_page();
    exit;
  }



  my ($second,$minute,$hour,$day,$month,$year)=localtime(time);
  my $timestamp = sprintf("%4d-%02d-%02d %02d:%02d ",$year+1900,$month+1,$day,$hour,$minute);

  my $mirrorlist = "# $timestamp\n";
  $mirrorlist .= "# generated by reflector\n";
  $mirrorlist .= "# ranked by rankmirrors\n" if $conf->get_value('--rank') or $conf->get_value('--fastest');
  $mirrorlist .= $self->get_mirrorlist();

  my $file = $conf->get_value('--output');
  if (defined($file))
  {
    open (my $fh,'>',$file) or die "invalid output file specified: $!";
    print $fh $mirrorlist;
    close $fh;
  }
  else
  {
    print $mirrorlist;
  }
}



sub parse_args
{
  my ($self,@args) = @_;
  $self->{CONF}->parse_cli_input(@args);
  $self->set_arch($self->{CONF}->get_value('--arch'))
}


# This is for use with other modules.
sub get_repos
{
  return @REPOS
}





sub get_mirrorstatus
{
  my ($self) = @_;
  my $txt = get($MIRRORSTATUS_URL);
  my $mirrorstatus = {};
  return $mirrorstatus if not defined($txt);
  my $t = time();
  foreach my $line (split /\s*\n\s*/, $txt)
  {
    if ($line =~ m/^(\d{4})-(\d\d)-(\d\d) (\d\d):(\d\d) (.+)$/)
    {
      my ($Y,$m,$d,$H,$M,$url) = ($1,$2,$3,$4,$5,$6);
      my $u = timegm(0,$M,$H,$d,$m-1,$Y);
      my $dt = ($t-$u)/3600; # time since last update, in hours
      $mirrorstatus->{$url} = $dt;
    }
  }
  return $mirrorstatus;
}




sub get_mirrorlist
{
  my ($self,@servers) = @_;
  @servers = $self->get_servers if not @servers;
  my $mirrorlist = '';
  my $arch = $self->{ARCH};
  $mirrorlist .= "Server = $_\n" foreach @servers;
  return $mirrorlist;
}





sub get_servers
{
  my ($self) = @_;
  my $arch = $self->{ARCH};
  my $mirrorstatus = $self->get_mirrorstatus;
  my @servers = keys %{ $mirrorstatus };


  my $g = $self->{CONF}->get_value('--grep');
  @servers = grep {m/$g/i} @servers if defined($g);


  my $h = $self->{CONF}->get_value('--hours');
  @servers = grep {$mirrorstatus->{$_} <= $h} @servers if defined($h);


  my $l = $self->{CONF}->get_value('--latest');
  @servers = (sort {$mirrorstatus->{$a} <=> $mirrorstatus->{$b}} @servers)[0..($l-1)] if defined($l);

  @servers = map {&strip_trailing_slash($_)} @servers;

  my $r = $self->{CONF}->get_value('--rank');
  my $f = $self->{CONF}->get_value('--fastest');

  # yeah, this is ugly
  if ($r or defined($f))
  {
    my $rmver = `rankmirrors --version 2>/dev/null`;
    if (length($rmver) > 0)
    {
      my $mirrorlist = join "\n", map {"Server = $_/\$repo/os/$arch"} @servers;
      my $n = defined($f) ? "-n $f" : '';
      open (my $rh,'-|',"rankmirrors $n - <<'MIRRORLIST'\n$mirrorlist\nMIRRORLIST\n");
      @servers = ();
      $|++;
      while (defined(my $line = <$rh>))
      {
        push @servers, $1 if $line =~ m/Server = (.*?)\/\$repo\/os\/$arch$/g;
      }
      close $rh;
    }
    else
    {
      print STDERR "error: failed to run rankmirrors... make sure that Python is installed\n";
      exit(1);
    }
  }
  return map {"$_/\$repo/os/$arch"} @servers;
}





sub set_arch
{
  my ($self,$arch) = @_;
  $self->{ARCH} = $arch if defined($arch);
}

1;

__END__

=head1 NAME

Xyne::Arch::Reflector - A class to retrieve and filter the latest Arch Linux mirror list.

=head1 SYNOPSIS

  use Xyne::Arch::Reflector;

=head1 DESCRIPTION

A class to retrieve and filter servers from the Arch Linux mirror status page.

=head2 EXPORT

None by default.

=head1 AUTHOR

Xyne <lt>xyne at archlinux dot ca<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009,2010  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
